/*
 *	errmes.c
 *		Copyright (C)2001-2003 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "errmes.h"

static void veprintf(const char *format, va_list argptr);

static const char *err_progname = NULL;

const char *err_getprogname(void)
{
	return err_progname;
}

void err_setprogname(const char *ptr_progname)
{
	const char *p;

	if (ptr_progname == NULL)
	{
		assert(0);
		return;
	}

	p = strrchr(ptr_progname, '/');
	err_progname = (p == NULL) ? ptr_progname : p + 1;
}

void err_exit(int status)
{
	fflush(stdout);

	if (err_progname != NULL)
		fprintf(stderr, "%s: ", err_getprogname());

	fprintf(stderr, "%s\n", strerror(errno));

	exit(status);
}

void err_printf(const char *format, ...)
{
	va_list argptr;

	fflush(stdout);
	if (format != NULL)
	{
		va_start(argptr, format);
		veprintf(format, argptr);
		va_end(argptr);
	}

	exit(EXIT_FAILURE);
}

void wrn_printf(const char *format, ...)
{
	va_list argptr;

	if (format != NULL)
	{
		va_start(argptr, format);
		veprintf(format, argptr);
		va_end(argptr);
	}
}

static void veprintf(const char *format, va_list argptr)
{
	if (format == NULL)
		return;

	if (err_progname != NULL)
		fprintf(stderr, "%s: ", err_getprogname());

	vfprintf(stderr, format, argptr);

	if (*format != '\0' && format[strlen(format) - 1] == ':')
		fprintf(stderr, " %s", strerror(errno));
	fprintf(stderr, "\n");
}
