/*
 *	ff4analyzer.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *  ff4a_analyzer
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "utility.h"
#include "ff4file.h"
#include "ff4decode.h"
#include "ff4monster.h"
#include "ff4character.h"
#include "ff4item.h"
#include "ff4magic.h"

#define FILENAME_FF4_E4				"e4"
#define FILENAME_MON_PARAM			"ff4monparam.txt"
#define FILENAME_MON_ELEM			"ff4monelem.txt"
#define FILENAME_MON_STATE			"ff4monstate.txt"
#define FILENAME_MON_FLAG			"ff4monflag.txt"
#define FILENAME_MON_ITEM			"ff4monitem.txt"
#define FILENAME_MON_ACT			"ff4monact.txt"
#define FILENAME_MON_MES			"ff4monmes.txt"
#define FILENAME_MON_PARTY			"ff4monparty.txt"
#define FILENAME_MON_ENC			"ff4monenc.txt"
#define FILENAME_MON_MAP			"ff4monmap.txt"
#define FILENAME_CHR_PARAM			"ff4chrparam.txt"
#define FILENAME_CHR_EQUIP			"ff4chrequip.txt"
#define FILENAME_CHR_LVUP			"ff4chrlvup.txt"
#define FILENAME_ITEM_PARAM			"ff4itemparam.txt"
#define FILENAME_ITEM_ELEM			"ff4itemelem.txt"
#define FILENAME_ITEM_STATE			"ff4itemstate.txt"
#define FILENAME_ITEM_FLAG			"ff4itemflag.txt"
#define FILENAME_ITEM_EFFECT		"ff4itemeffect.txt"
#define FILENAME_MAG_PARAM			"ff4magparam.txt"
#define FILENAME_MAG_ELEM			"ff4magelem.txt"
#define FILENAME_MAG_STATE			"ff4magstate.txt"
#define FILENAME_MAG_FLAG			"ff4magflag.txt"
#define FILENAME_TEXT_KANA			"ff4kana.txt"
#define FILENAME_TEXT_KANJI			"ff4kanji.txt"

#define STR_INVALID_OPTION			"Invalid option"
#define STR_INVALID_ROMIMAGE_FILE	"Invalid ROM-Image file: %s"
#define STR_INVALID_CDROM_FILES		"Invalid CD-ROM files"
#define STR_CANNOT_CREATE_FILE		"Cannot create file: %s"
#define STR_FILENAME_TOO_LONG		"Filename too long"

#define STR_READ_FROM_FILE			"Reading from %s\n"
#define STR_WRITE_TO_FILE			"Writing to %s\n"

/* ɽ */
static const char *str_copyright =
	"ff4analyzer   - Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>\n"
	"ff4a_analyzer - Modified by Inu";

static const char *tbl_gametitle[] =
{
	"FINAL FANTASY 4",
			/* 0x46, 0x49, 0x4e, 0x41, 0x4c, 0x20, 0x46, 0x41, 0x4e, 0x54, 0x41, 0x53, 0x59, 0x20, 0x34 */
			/* פǤ⸡ФǤ褦ˡʬоȤȤ */
	NULL,
			/* unused */
	"FF4ADVANCE\0\0",
			/* 0x46, 0x46, 0x34, 0x41, 0x44, 0x56, 0x41, 0x4E, 0x43, 0x45, 0x00, 0x00 */
};

void usage(void);

/* E4 Ǥξ硢Ϥѥ ff4*  ff4e4* ѹ */
#define ff4_sprintf(x, dstpath, filename)         \
	snprintf(x, sizeof(x), "%s%.3s%s%s",          \
			dstpath,                              \
			filename,                             \
			(flag_e4 > 0) ? FILENAME_FF4_E4 : "", \
			filename + 3)                         \

#define do_writedata(filename, function) do {         \
	char path[sizeof(str_dstpath) + 128] = "";        \
	ff4_sprintf(path, str_dstpath, filename);         \
	if (!flag_dry)                                    \
	{                                                 \
		fprintf(stderr, STR_WRITE_TO_FILE, path);     \
		if (!function)                                \
			err_printf(STR_CANNOT_CREATE_FILE, path); \
	}                                                 \
} while (0)

int main(int argc, char *argv[])
{
	int optind = 0;
	int flag_dry = 0;
	int flag_e4 = 0;
	char str_srcpath[1024] = "";
	char str_dstpath[1024] = "";
	mondata_t  *buf_mondata  = NULL;
	monact_t   *buf_monact   = NULL;
	monparty_t *buf_monparty = NULL;
	monenc_t   *buf_monenc   = NULL;
	monmap_t   *buf_monmap   = NULL;
	chrdata_t  *buf_chrdata  = NULL;
	itemdata_t *buf_itemdata = NULL;
	magdata_t  *buf_magdata  = NULL;

	errno = 0;
	err_setprogname(argv[0]);

	/* ޥɥ饤Ϥ */
	for (optind = 1; optind < argc; optind++)
	{
		const char *p = argv[optind];

		if (p == NULL)
			continue;

		if (*p != '-')
			break;
		else if (*(p + 1) == '-')
		{
			optind++;
			break;
		}

		while (*++p != '\0')
		{
			switch (*p)
			{
			case 'n':
				flag_dry = 1;
				break;
			case 'h':	/* ˡɽ */
			default:
				usage();
				break;
			}
		}
	}
	argc -= optind;
	argv += optind;

	/* ϥѥꤹ */
	if (argc < 1)
		usage();

	if (strlen(argv[0]) > sizeof(str_srcpath))
		err_printf(STR_FILENAME_TOO_LONG);
	transfilename(str_srcpath, argv[0], sizeof(str_srcpath));

	/* ϥѥꤹ */
	if (argc >= 2)
	{
		if (strlen(argv[1]) > sizeof(str_dstpath))
			err_printf(STR_FILENAME_TOO_LONG);
		transpath(str_dstpath, argv[1], sizeof(str_dstpath));
	}

	/* ե򥪡ץ󤹤 */
	if (!file_open(str_srcpath, tbl_gametitle[FILEMODE_GBA], FILEMODE_GBA))
		err_printf(STR_INVALID_ROMIMAGE_FILE, str_srcpath);

	/* ROM Сγǧ */
	if (ff4_get_version() == 1)
	{
		wrn_printf("Detecting Final Fantasy IV Advance, V1.1 (E4)");
		flag_e4 = 1;
	}

	/* ƼǡϡϤ */
	buf_mondata  = mon_readdata();
#if 0
	buf_monact   = mon_readact();
#endif
	buf_monparty = mon_readparty();
	buf_monenc   = mon_readenc();
	buf_monmap   = mon_readmap();
	buf_chrdata  = chr_readdata();
	buf_itemdata = item_readdata();
	buf_magdata  = mag_readdata();

	if (buf_mondata  == NULL ||
#if 0
		buf_monact   == NULL ||
#endif
		buf_monparty == NULL ||
		buf_monenc   == NULL ||
		buf_monmap   == NULL ||
		buf_chrdata  == NULL ||
		buf_itemdata == NULL ||
		buf_magdata  == NULL)
	{
		err_printf(STR_INVALID_ROMIMAGE_FILE, str_srcpath);
	}

	/* ƼǡϤ */
	do_writedata(FILENAME_MON_PARAM,	mon_writeparam(path, buf_mondata));
	do_writedata(FILENAME_MON_ELEM,		mon_writeelem(path, buf_mondata));
	do_writedata(FILENAME_MON_STATE,	mon_writestate(path, buf_mondata));
	do_writedata(FILENAME_MON_FLAG,		mon_writeflag(path, buf_mondata));
	do_writedata(FILENAME_MON_ITEM,		mon_writeitem(path, buf_mondata));
#if 0
	do_writedata(FILENAME_MON_ACT,		mon_writeact(path, buf_monact, buf_mondata));
#endif
	do_writedata(FILENAME_MON_MES,		mon_writemes(path));
	do_writedata(FILENAME_MON_PARTY,	mon_writeparty(path, buf_monparty));
	do_writedata(FILENAME_MON_ENC,		mon_writeenc(path, buf_monenc, buf_monparty));
	do_writedata(FILENAME_MON_MAP,		mon_writemap(path, buf_monmap));
	do_writedata(FILENAME_CHR_PARAM,	chr_writeparam(path, buf_chrdata));
	do_writedata(FILENAME_CHR_EQUIP,	chr_writeequip(path, buf_chrdata));
	do_writedata(FILENAME_CHR_LVUP,		chr_writelvup(path, buf_chrdata));
	do_writedata(FILENAME_ITEM_PARAM,	item_writeparam(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_ELEM,	item_writeelem(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_STATE,	item_writestate(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_FLAG,	item_writeflag(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_EFFECT,	item_writeeffect(path, buf_itemdata));
	do_writedata(FILENAME_MAG_PARAM,	mag_writeparam(path, buf_magdata));
	do_writedata(FILENAME_MAG_ELEM,		mag_writeelem(path, buf_magdata));
	do_writedata(FILENAME_MAG_STATE,	mag_writestate(path, buf_magdata));
	do_writedata(FILENAME_MAG_FLAG,		mag_writeflag(path, buf_magdata));
	do_writedata(FILENAME_TEXT_KANA,	dec_writedata(path, 0, -1, FLAG_KANA));
	do_writedata(FILENAME_TEXT_KANJI,	dec_writedata(path, 0, -1, FLAG_KANJI));

	/* ե򥯥 */
	file_close();

	/* Хåե */
#if 0
	mon_freeact(buf_monact);
#endif
	free(buf_mondata);
	free(buf_monact);
	free(buf_monparty);
	free(buf_monenc);
	free(buf_monmap);
	free(buf_chrdata);
	free(buf_itemdata);
	free(buf_magdata);
	buf_mondata  = NULL;
	buf_monact   = NULL;
	buf_monparty = NULL;
	buf_monenc   = NULL;
	buf_monmap   = NULL;
	buf_chrdata  = NULL;
	buf_itemdata = NULL;
	buf_magdata  = NULL;

	return EXIT_SUCCESS;
}

void usage(void)
{
	fprintf(stderr, "%s\n\n", str_copyright);
	fprintf(stderr,	"usage: ff4analyzer: [-hn] src-pathname [dst-pathname]\n"
					"    -h                - display this help and exit\n"
					"    -n                - dry run\n"
					"    src-pathname      - ROM image file\n"
					"    dst-pathname      - write files into directory\n");

	exit(EXIT_SUCCESS);
}
