/*
 *	ff4character.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4file.h"
#include "ff4define.h"
#include "ff4decode.h"
#include "ff4item.h"
#include "ff4character.h"

#define FILEPOS_CHRDATA			0x0f2eb0		/* SNES: 0x07a900 */
#define FILEPOS_CHRCRITICAL		0x0f5850
#define FILEPOS_CHRLVUP			0x0e7bf0		/* SNES: 0x07b528 */
#define NUM_CHR					15
#define NUM_CHRLVUP				(NUM_CHR - 2)	/* = 13 */
#define NUM_CHRNAME				13
#define NUM_CHRJOBNAME			15
#define SIZEOF_CHRDATA			32
#define SIZEOF_CHRCRITICAL		2
#define SIZEOF_CHRLVUP			16 				/* SNES: 5 */
#define MSGID_CHRNAME			244
#define MSGID_CHRJOBNAME		3227

static void chr_setname(void);
static void chr_setjobname(void);

static char tbl_chrname[NUM_CHRNAME][256];			/* 饯̾Ǽơ֥ */
static char tbl_chrjobname[NUM_CHRJOBNAME][256];	/* ֤̾Ǽơ֥ */
static int flag_setchrname    = 0;					/* tbl_chrname˥ǡǼƤ뤫ɤ */
static int flag_setchrjobname = 0;					/* tbl_chrjobname˥ǡǼƤ뤫ɤ */

const char *chr_getname(int index)
{
	/* å */
	if (index < 0 || NUM_CHR <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setchrname)
		chr_setname();

	return tbl_chrname[index];
}

const char *chr_getjobname(int index)
{
	/* å */
	if (index < 0 || NUM_CHRJOBNAME <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setchrjobname)
		chr_setjobname();

	return tbl_chrjobname[index];
}

static void chr_setname(void)
{
	int i;

	/* 饯̾ǡ */
	for (i = 0; i < NUM_CHRNAME; i++)
		*tbl_chrname[i] = '\0';

	/* 饯̾ǡɤ߹ */
	for (i = 0; i < NUM_CHRNAME; i++)
		dec_msg(tbl_chrname[i], sizeof(tbl_chrname[i]), MSGID_CHRNAME + i, FLAG_KANA);

	flag_setchrname = 1;
}

static void chr_setjobname(void)
{
	int i;

	/* 饯Υ̾ǡ */
	for (i = 0; i < NUM_CHRJOBNAME; i++)
		*tbl_chrjobname[i] = '\0';

	/* 饯Υ̾ǡɤ߹ */
	for (i = 0; i < NUM_CHRJOBNAME; i++)
		dec_msg(tbl_chrjobname[i], sizeof(tbl_chrjobname[i]), MSGID_CHRJOBNAME + i, FLAG_KANA);

	flag_setchrjobname = 1;
}

chrdata_t *chr_readdata(void)
{
	int i, j;
	long filepos;
	chrdata_t *buf_chrdata = NULL;
	const int tbl_index_chrname[NUM_CHR] = {
		0, 1,  2,  3,
		4, 5,  6,  7,
		8, 9, 10, 11,
		0, 3, 12,
	};
	const int tbl_index_chrlvup[NUM_CHRLVUP] = {
		0,  1, 3,  5,
		6,  2, 7,  8,
		9, 12, 4, 10,
		11,
	};

	/* 饯ǡǼХåեݤ */
	if ((buf_chrdata = malloc(sizeof(chrdata_t) * NUM_CHR)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 饯ǡɤ߹ */
	for (i = 0; i < NUM_CHR; i++)
	{
		int index = 0;
		unsigned char buf[SIZEOF_CHRDATA];

		file_read(buf, FILEPOS_CHRDATA + SIZEOF_CHRDATA * i, SIZEOF_CHRDATA);

		/* ͷϥǡ */
		buf_chrdata[i].chr_job         = i;
		buf_chrdata[i].chr_name        = tbl_index_chrname[i];
		buf_chrdata[i].initial_level   = buf[0];
		buf_chrdata[i].flag_right_hand = (buf[1] == 0 || buf[1] == 2);
		buf_chrdata[i].flag_left_hand  = (buf[1] == 1 || buf[1] == 2);

		/* ƥǡ */
		for (j = 0; j < 5; j++)
		{
			if (j < 2)
			{
				buf_chrdata[i].equip[j].item_no  = (buf[17 + j * 4] << 8) | buf[16 + j * 4];
				buf_chrdata[i].equip[j].item_num = (buf[19 + j * 4] << 8) | buf[18 + j * 4];
			}
			else
			{
				buf_chrdata[i].equip[j].item_no  = (buf[21 + j * 2] << 8) | buf[20 + j * 2];
				buf_chrdata[i].equip[j].item_num = (buf_chrdata[i].equip[j].item_no != 0x00) ? 1 : 0;
			}
		}

		/* ΰ򥯥ꥢ */
		index = buf_chrdata[i].initial_level - 1;
		for (j = 0; j < index; j++)
		{
			buf_chrdata[i].level[j].experience_point         = 0;
			buf_chrdata[i].level[j].hit_point_min            = 0;
			buf_chrdata[i].level[j].hit_point_max            = 0;
			buf_chrdata[i].level[j].magic_point_min          = 0;
			buf_chrdata[i].level[j].magic_point_max          = 0;
			buf_chrdata[i].level[j].increase_hit_point_min   = 0;
			buf_chrdata[i].level[j].increase_hit_point_min   = 0;
			buf_chrdata[i].level[j].increase_magic_point_max = 0;
			buf_chrdata[i].level[j].increase_magic_point_max = 0;
			buf_chrdata[i].level[j].strength                 = 0;
			buf_chrdata[i].level[j].speed                    = 0;
			buf_chrdata[i].level[j].stamina                  = 0;
			buf_chrdata[i].level[j].wisdom                   = 0;
			buf_chrdata[i].level[j].will_power               = 0;
		}

		/* ٥ˤи */
		buf_chrdata[i].initial_experience_point              = bits_make32(buf[12], buf[13], buf[14], buf[15]);

		/* ٥ˤѥ᡼ */
		buf_chrdata[i].level[index].hit_point_min            = (buf[ 3] << 8) | buf[ 2];
		buf_chrdata[i].level[index].magic_point_min          = (buf[ 5] << 8) | buf[ 4];
		buf_chrdata[i].level[index].strength                 = buf[ 6];
		buf_chrdata[i].level[index].speed                    = buf[ 7];
		buf_chrdata[i].level[index].stamina                  = buf[ 8];
		buf_chrdata[i].level[index].wisdom                   = buf[ 9];
		buf_chrdata[i].level[index].will_power               = buf[10];

		buf_chrdata[i].level[index].hit_point_max            = buf_chrdata[i].level[index].hit_point_min;
		buf_chrdata[i].level[index].magic_point_max          = buf_chrdata[i].level[index].magic_point_min;
	}

	/* 饯Υƥ̥ǡɤ߹ */
	for (i = 0; i < NUM_CHR; i++)
	{
		unsigned char buf[SIZEOF_CHRCRITICAL];

		file_read(buf, FILEPOS_CHRCRITICAL + SIZEOF_CHRCRITICAL * i, SIZEOF_CHRCRITICAL);

		buf_chrdata[i].critical_hit_rate     = buf[0];
		buf_chrdata[i].critical_effect_power = buf[1];
	}

	/* 饯Ĺǡɤ߹ */
	filepos = FILEPOS_CHRLVUP;
	for (i = 0; i < NUM_CHRLVUP; i++)
	{
		const int chr_id = tbl_index_chrlvup[i];
		const int index = buf_chrdata[chr_id].initial_level - 1;

		/* Ĺǡ */
		for (j = index; j < NUM_CHRLV; j++)
		{
			unsigned char buf[SIZEOF_CHRLVUP];

			filepos += file_read(buf, filepos, SIZEOF_CHRLVUP);

			buf_chrdata[chr_id].level[j].increase_hit_point_min   = buf[0];
			buf_chrdata[chr_id].level[j].increase_hit_point_max   = buf[1];
			buf_chrdata[chr_id].level[j].increase_magic_point_min = buf[2];
			buf_chrdata[chr_id].level[j].increase_magic_point_max = buf[3];

			buf_chrdata[chr_id].level[j].experience_point = bits_make32(buf[4], buf[5], buf[6], buf[7]);

			if (j < 70)
			{
				buf_chrdata[chr_id].level[j].strength   = buf[ 8];
				buf_chrdata[chr_id].level[j].speed      = buf[ 9];
				buf_chrdata[chr_id].level[j].stamina    = buf[10];
				buf_chrdata[chr_id].level[j].wisdom     = buf[11];
				buf_chrdata[chr_id].level[j].will_power = buf[12];
			}
			else
			{
				const int grow_index = j - 70;

				/* ٥71ʾǤĹѥ */
				if (grow_index < NUM_CHRGROWPTN)
				{
					buf_chrdata[chr_id].grow_pattern[grow_index].increase_strength   = ff4_signed_int(buf[ 8]);
					buf_chrdata[chr_id].grow_pattern[grow_index].increase_speed      = ff4_signed_int(buf[ 9]);
					buf_chrdata[chr_id].grow_pattern[grow_index].increase_stamina    = ff4_signed_int(buf[10]);
					buf_chrdata[chr_id].grow_pattern[grow_index].increase_wisdom     = ff4_signed_int(buf[11]);
					buf_chrdata[chr_id].grow_pattern[grow_index].increase_will_power = ff4_signed_int(buf[12]);
				}

				buf_chrdata[chr_id].level[j].strength   = 0;
				buf_chrdata[chr_id].level[j].speed      = 0;
				buf_chrdata[chr_id].level[j].stamina    = 0;
				buf_chrdata[chr_id].level[j].wisdom     = 0;
				buf_chrdata[chr_id].level[j].will_power = 0;
			}

			/* ߷ץҥåȥݥȡޥåݥ */
			if (j > index)
			{
				buf_chrdata[chr_id].level[j].hit_point_min    = buf_chrdata[chr_id].level[j - 1].hit_point_min +
																buf_chrdata[chr_id].level[j].increase_hit_point_min;
				buf_chrdata[chr_id].level[j].hit_point_max    = buf_chrdata[chr_id].level[j - 1].hit_point_max +
																buf_chrdata[chr_id].level[j].increase_hit_point_max;
				buf_chrdata[chr_id].level[j].magic_point_min  = buf_chrdata[chr_id].level[j - 1].magic_point_min +
																buf_chrdata[chr_id].level[j].increase_magic_point_min;
				buf_chrdata[chr_id].level[j].magic_point_max  = buf_chrdata[chr_id].level[j - 1].magic_point_max +
																buf_chrdata[chr_id].level[j].increase_magic_point_max;

				if (buf_chrdata[chr_id].level[j].hit_point_min > 9999)
					buf_chrdata[chr_id].level[j].hit_point_min = 9999;
				if (buf_chrdata[chr_id].level[j].hit_point_max > 9999)
					buf_chrdata[chr_id].level[j].hit_point_max = 9999;
				if (buf_chrdata[chr_id].level[j].magic_point_min > 9999)
					buf_chrdata[chr_id].level[j].magic_point_min = 9999;
				if (buf_chrdata[chr_id].level[j].magic_point_max > 9999)
					buf_chrdata[chr_id].level[j].magic_point_max = 9999;
			}
		}
	}

	return buf_chrdata;
}

int chr_writeparam(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_hands[2][2] =
	{
		{"Դ",	""},
		{"",	"ξ"},
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯ǡ(ͷ)Ϥե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯ǡ(ͷ)Ϥ */
	fprintf(fp, "No.      ̾           "
				"        "
				"  LV HP MP   EXP "
				"     ®    "
				"CH   CHΨ  "
				"\n");
	for (i = 0; i < NUM_CHR; i++)
	{
		const int index = ptr_chrdata[i].initial_level - 1;

		fprintf(fp, "%03d(%02X): %-12s : %-12s,  %3d, %5d, %5d, %8ld, %3d, %3d, %3d, %3d, %3d, %5d, %5d, %-6s\n", i, i,
				chr_getname(ptr_chrdata[i].chr_name),
				chr_getjobname(ptr_chrdata[i].chr_job),
				ptr_chrdata[i].initial_level,
				ptr_chrdata[i].level[index].hit_point_min,
				ptr_chrdata[i].level[index].magic_point_min,
				ptr_chrdata[i].initial_experience_point,
				ptr_chrdata[i].level[index].strength,
				ptr_chrdata[i].level[index].speed,
				ptr_chrdata[i].level[index].stamina,
				ptr_chrdata[i].level[index].wisdom,
				ptr_chrdata[i].level[index].will_power,
				ptr_chrdata[i].critical_effect_power,
				ptr_chrdata[i].critical_hit_rate,
				str_hands[(ptr_chrdata[i].flag_right_hand != 0) ? 1 : 0][(ptr_chrdata[i].flag_left_hand != 0) ? 1 : 0]);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writeequip(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯ƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯ƥǡϤ */
	fprintf(fp, "No.      ̾           "
				"              "
				"              "
				"Ƭ                "
				"                "
				"\n");
	for (i = 0; i < NUM_CHR; i++)
	{
		const char *ptr_name[5];

		for (j = 0; j < 5; j++)
			ptr_name[j] = (ptr_chrdata[i].equip[j].item_no != 0x00) ? item_getname(ptr_chrdata[i].equip[j].item_no) : str_nothing;

		fprintf(fp, "%03d(%02X): %-12s : %-16s, %-16s, %-16s, %-16s, %-16s\n", i, i,
				chr_getname(ptr_chrdata[i].chr_name),
				ptr_name[0],
				ptr_name[1],
				ptr_name[2],
				ptr_name[3],
				ptr_name[4]);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writelvup(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯ĹǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯ĹǡϤ */
	for (i = 0; i < NUM_CHRLVUP; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X): %s(%s)]\n", i, i, chr_getname(ptr_chrdata[i].chr_name), chr_getjobname(ptr_chrdata[i].chr_job));

		/* ٥71ʾǤĹѥ */
		fprintf(fp, "No.      ®   \n");
		for (j = 0; j < NUM_CHRGROWPTN; j++)
		{
			fprintf(fp, "%02d:  %3d, %3d, %3d, %3d, %3d\n", j + 1,
					ptr_chrdata[i].grow_pattern[j].increase_strength,
					ptr_chrdata[i].grow_pattern[j].increase_speed,
					ptr_chrdata[i].grow_pattern[j].increase_stamina,
					ptr_chrdata[i].grow_pattern[j].increase_wisdom,
					ptr_chrdata[i].grow_pattern[j].increase_will_power);
		}

		/* Ĺǡ */
		fprintf(fp, "LV        HP (ϰ)     徺HP    MP (ϰ)   徺MP    и "
					"     ®   \n");
		for (j = ptr_chrdata[i].initial_level - 1; j < 70; j++)
		{
			fprintf(fp, "%2d:  %4d (%4d - %4d), %3d - %3d, %3d (%3d - %3d), %2d - %2d, %8ld, %3d, %3d, %3d, %3d, %3d\n", j + 1,
					(ptr_chrdata[i].level[j].hit_point_min + ptr_chrdata[i].level[j].hit_point_max) / 2,
					ptr_chrdata[i].level[j].hit_point_min,
					ptr_chrdata[i].level[j].hit_point_max,
					ptr_chrdata[i].level[j].increase_hit_point_min,
					ptr_chrdata[i].level[j].increase_hit_point_max,
					(ptr_chrdata[i].level[j].magic_point_min + ptr_chrdata[i].level[j].magic_point_max) / 2,
					ptr_chrdata[i].level[j].magic_point_min,
					ptr_chrdata[i].level[j].magic_point_max,
					ptr_chrdata[i].level[j].increase_magic_point_min,
					ptr_chrdata[i].level[j].increase_magic_point_max,
					ptr_chrdata[i].level[j].experience_point,
					ptr_chrdata[i].level[j].strength,
					ptr_chrdata[i].level[j].speed,
					ptr_chrdata[i].level[j].stamina,
					ptr_chrdata[i].level[j].wisdom,
					ptr_chrdata[i].level[j].will_power);
		}
		for (; j < NUM_CHRLV; j++)
		{
			fprintf(fp, "%2d:  %4d (%4d - %4d), %3d - %3d, %3d (%3d - %3d), %2d - %2d, %8ld\n", j + 1,
					(ptr_chrdata[i].level[j].hit_point_min + ptr_chrdata[i].level[j].hit_point_max) / 2,
					ptr_chrdata[i].level[j].hit_point_min,
					ptr_chrdata[i].level[j].hit_point_max,
					ptr_chrdata[i].level[j].increase_hit_point_min,
					ptr_chrdata[i].level[j].increase_hit_point_max,
					(ptr_chrdata[i].level[j].magic_point_min + ptr_chrdata[i].level[j].magic_point_max) / 2,
					ptr_chrdata[i].level[j].magic_point_min,
					ptr_chrdata[i].level[j].magic_point_max,
					ptr_chrdata[i].level[j].increase_magic_point_min,
					ptr_chrdata[i].level[j].increase_magic_point_max,
					ptr_chrdata[i].level[j].experience_point);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
