/*
 *	ff4define.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4define.h"

#define elem_maskbit(elem_no)		(1U << (elem_no))
#define state_maskbit(state_no)		(1U << (state_no))
#define montype_maskbit(montype_no)	(1U << (montype_no))

#define STR_NOTHING		""
#define STR_FLAG_TRUE	""
#define STR_FLAG_FALSE	""

static const char *str_element[/* NUM_ELEMENT */] =
{
	"",	/*  */
	"ɹ",	/* 䵤 */
	"",	/*  */
	"",	/* Ǥ */
	"",	/* ʤ */
	"",	/* ƻ */
	"",	/* ۼ */
	"",	/* ()*/
};

static const char *str_state[/* NUM_STATE */] =
{
	"",	/*  */
	"",	/* Ű */
	"",	/*  */
	"",	/*  */
	"",	/*  */
	"",	/*  */
	"",	/* в */
	"",	/* Ʈǽ */

	"",	/* в */
	"",	/* в(ߡ?) */
	"",	/*  */
	"",	/*  */
	"̲",	/* ̲ */
	"",	/*  */
	"",	/* ӥƥ */
	"",	/*  */

	"",	/*  */
	"??",	/*  */
	"??",	/*  */
	"??",	/*  */
	"??",	/*  */
	"",	/* ȥå */
	"??",	/*  */
	"??",	/*  */

	"",	/* λ */
	"??",	/*  */
	"ʬ",	/* ʬ(1) */
	"ʬ",	/* ʬ(2) */
	"",	/* Хꥢ */
	"ȿ",	/* ե쥯 */
	"??",	/*  */
	"??",	/* (Υæ? ̵Ũ?) */
};

static const char *str_montype[/* NUM_MONTYPE */] =
{
	"ɥ饴",
	"",
	"ꥶ",
	"",
	"",
	"ץ",
	"ˡȤ",
	"ǥå",
};

const char *str_getflag(int flag)
{
	return (flag != 0x00) ? STR_FLAG_TRUE : STR_FLAG_FALSE;
}

const char *str_getelem(elem_t elem, int index)
{
	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* å */
	if (index < 0 || NUM_ELEMENT <= index)
	{
		assert(0);
		return "";
	}

	return ((elem & elem_maskbit(index)) != 0x00) ? str_element[index] : "";
}

void str_printelem(FILE *stream, elem_t elem, elem_t elem_mask)
{
	int i;

	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* å */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_ELEMENT; i++)
	{
		if ((elem_mask & elem_maskbit(i)) != 0x00)
			fprintf(stream, "%-2s", ((elem & elem_maskbit(i)) != 0x00) ? str_element[i] : STR_NOTHING);
	}
}

const char *str_getstate(state_t state, int index)
{
	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* å */
	if (index < 0 || NUM_STATE <= index)
	{
		assert(0);
		return "";
	}

	return ((state & state_maskbit(index)) != 0x00) ? str_state[index] : "";
}

void str_printstate(FILE *stream, state_t state, state_t state_mask)
{
	int i, count = 0;

	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* å */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_STATE; i++)
	{
		if ((i % 8) == 0)
		{
			if (count > 0 && state_mask > state_maskbit(i))
				fprintf(stream, " ");
			count = 0;
		}

		if ((state_mask & state_maskbit(i)) != 0x00)
		{
			fprintf(stream, "%-2s", ((state & state_maskbit(i)) != 0x00) ? str_state[i] : STR_NOTHING);
			count++;
		}
	}
}

const char *str_getmontype(montype_t montype, int index)
{
	assert(sizeof(str_montype) / sizeof(str_montype[0]) == NUM_MONTYPE);

	/* å */
	if (index < 0 || NUM_MONTYPE <= index)
	{
		assert(0);
		return "";
	}

	return ((montype & montype_maskbit(index)) != 0x00) ? str_montype[index] : "";
}

void str_printmontype(FILE *stream, montype_t montype, montype_t montype_mask)
{
	int i, count = 0;

	assert(sizeof(str_montype) / sizeof(str_montype[0]) == NUM_MONTYPE);

	/* å */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_MONTYPE; i++)
	{
		if ((montype_mask & montype_maskbit(i)) != 0x00)
		{
			if ((montype & montype_maskbit(i)) != 0x00)
			{
				if (count > 0)
					fprintf(stream, " ");
				count++;
			}

			fprintf(stream, "%s", str_getmontype(montype, i));
		}
	}
}
