/*
 *	ff4item.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4file.h"
#include "ff4decode.h"
#include "ff4item.h"

#define FILEPOS_ITEMDATA		0x0ecfa0		/* SNES: 0x079100 */
#define FILEPOS_ITEMEQUIP		0x0e5dcc
#define FILEPOS_ITEMEFFECT		0x0ec578
#define NUM_ITEM				325 			/* SNES: 256 */
#define NUM_ITEMEQUIP			230
#define SIZEOF_ITEMDATA			20
#define SIZEOF_ITEMEFFECT		8
#define SIZEOF_ITEMEQUIP		12
#define MSGID_ITEMNAME			3997

static void item_setname(void);

static char tbl_itemname[NUM_ITEM][256];	/* ƥ̾Ǽơ֥ */
static int flag_setitemname = 0;			/* tbl_itemname˥ǡǼƤ뤫ɤ */

const char *item_getname(int index)
{
	/* å */
	if (index < 0 || NUM_ITEM <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setitemname)
		item_setname();

	return tbl_itemname[index];
}

static void item_setname(void)
{
	int i;

	/* ƥ̾ǡ */
	for (i = 0; i < NUM_ITEM; i++)
		*tbl_itemname[i] = '\0';

	/* ƥ̾ǡɤ߹ */
	for (i = 0; i < NUM_ITEM; i++)
		dec_msg(tbl_itemname[i],sizeof(tbl_itemname[i]), MSGID_ITEMNAME + i, FLAG_KANA);

	flag_setitemname = 1;
}

itemdata_t *item_readdata(void)
{
	int i;
	itemdata_t *buf_itemdata = NULL;

	/* ƥγƼǡǼХåեݤ */
	if ((buf_itemdata = malloc(sizeof(itemdata_t) * NUM_ITEM)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ƥγƼǡɤ߹ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMDATA];

		file_read(buf, FILEPOS_ITEMDATA + SIZEOF_ITEMDATA * i, SIZEOF_ITEMDATA);

		/* ͷϥǡ */
		buf_itemdata[i].type_no   = buf[0];
		buf_itemdata[i].price     = bits_make32(buf[4], buf[5], buf[6], buf[7]);
		buf_itemdata[i].effect_no = (buf[17] << 8) | buf[16];
		buf_itemdata[i].flag_use  = buf[8] & 0x07;
		buf_itemdata[i].flag_bank = ((buf[9] & 0x80) == 0x00);

		/* ƥμȽ̤ */
		if (i == 0x46 || (buf_itemdata[i].type_no >= 0x01 && buf_itemdata[i].type_no <= 0x12))
			buf_itemdata[i].type = ITEM_TYPE_WEAPON;
		else if (i == 0x60 || (buf_itemdata[i].type_no >= 0x13 && buf_itemdata[i].type_no <= 0x16))
			buf_itemdata[i].type = ITEM_TYPE_PROTECTOR;
		else if (buf_itemdata[i].effect_no != 0x00 && buf_itemdata[i].type_no == 0x00)
			buf_itemdata[i].type = ITEM_TYPE_TOOL;
		else
			buf_itemdata[i].type = ITEM_TYPE_OTHER;

		switch (buf_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			/* ե饰ϥǡ */
			buf_itemdata[i].param.weapon.flag_good_hand = ((buf[8] & 0x40) != 0x00);	/*  */
			buf_itemdata[i].param.weapon.flag_weak_hand = ((buf[8] & 0x10) != 0x00);	/*  */
			buf_itemdata[i].param.weapon.flag_2handed   = ((buf[8] & 0x20) != 0x00);
			break;

		case ITEM_TYPE_PROTECTOR:
		case ITEM_TYPE_OTHER:
			break;

		case ITEM_TYPE_TOOL:
			/* ϰϥǡ */
			buf_itemdata[i].param.tool.flag_range_self  = 0;
			buf_itemdata[i].param.tool.direction        = 0;
			buf_itemdata[i].param.tool.change_direction = 0;
			buf_itemdata[i].param.tool.range            = 0;
			buf_itemdata[i].param.tool.change_range     = 0;

			if (buf[18] == 0x08)
			{
				buf_itemdata[i].param.tool.direction = 1;
				buf_itemdata[i].param.tool.range     = 1;
			}
			else if (buf[18] == 0x20)
			{
				buf_itemdata[i].param.tool.flag_range_self = 1;
			}
			else
			{
				buf_itemdata[i].param.tool.direction        = ((buf[18] & 0x80) != 0x00) ? 1 : 0;
				buf_itemdata[i].param.tool.change_direction = ((buf[18] & 0x04) != 0x00) ? 1 : 0;
				buf_itemdata[i].param.tool.range            = ((buf[18] & 0x02) != 0x00) ? 1 : 0;
				buf_itemdata[i].param.tool.change_range     = ((buf[18] & 0x03) == 0x03) ? 1 : 0;
			}
			break;

		default:
			assert(0);
		}
	}

	/* ƥγƼ°Ѳե饰ϥǡɤ߹ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMEFFECT];

		file_read(buf, FILEPOS_ITEMEFFECT + SIZEOF_ITEMEFFECT * i, SIZEOF_ITEMEFFECT);

		switch (buf_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			/* °Ѳϥǡ */
			buf_itemdata[i].param.weapon.elem_attack    = buf[0];
			buf_itemdata[i].param.weapon.state_attack   = bits_make32(buf[2], buf[3], buf[4], buf[5]);
			buf_itemdata[i].param.weapon.montype_attack = buf[6];

			/* ե饰ϥǡ */
			buf_itemdata[i].param.weapon.flag_magnetic  = ((buf[7] & 0x01) != 0x00);
			buf_itemdata[i].param.weapon.flag_throw     = ((buf[7] & 0x02) != 0x00);
			buf_itemdata[i].param.weapon.flag_backrow   = ((buf[7] & 0x04) != 0x00);
			break;

		case ITEM_TYPE_PROTECTOR:
			/* °Ѳϥǡ */
			buf_itemdata[i].param.protector.elem_defense    = buf[0];
			buf_itemdata[i].param.protector.state_defense   = bits_make32(buf[2], buf[3], buf[4], buf[5]);
			buf_itemdata[i].param.protector.montype_defense = buf[6];

			/* ե饰ϥǡ */
			buf_itemdata[i].param.protector.flag_magnetic   = ((buf[7] & 0x01) != 0x00);
			break;

		case ITEM_TYPE_TOOL:
			/* °Ѳϥǡ */
			buf_itemdata[i].param.tool.elem_effect  = buf[0];
			buf_itemdata[i].param.tool.state_effect = bits_make32(buf[2], buf[3], buf[4], buf[5]);
			break;
		}
	}

	/* ƥγƼǡɤ߹ */
	for (i = 0; i < NUM_ITEMEQUIP; i++)
	{
		const int item_no = (i <= 0xaf) ? i :
							(i >= 0xe5) ? i + 0x144 - 0xe5 :
							(i >= 0xc1) ? i + 0x116 - 0xc1 :
							(i >= 0xb0) ? i + 0x0fc - 0xb0 : 0;
		unsigned char buf[SIZEOF_ITEMEQUIP];

		file_read(buf, FILEPOS_ITEMEQUIP + SIZEOF_ITEMEQUIP * i, SIZEOF_ITEMEQUIP);

		switch (buf_itemdata[item_no].type)
		{
		case ITEM_TYPE_WEAPON:
			/* ͷϥǡ */
			buf_itemdata[item_no].param.weapon.attack_power        = buf[0];
			buf_itemdata[item_no].param.weapon.hit_rate            = buf[2] & 0x7f;

			/* °Ѳϥǡ */
			buf_itemdata[item_no].param.weapon.bits_equip_chr      = (buf[11] << 8) | buf[10];

			/* ѥ᡼Ѳǡ */
			buf_itemdata[item_no].param.weapon.increase_strength   = ff4_signed_int(buf[5]);
			buf_itemdata[item_no].param.weapon.increase_speed      = ff4_signed_int(buf[6]);
			buf_itemdata[item_no].param.weapon.increase_stamina    = ff4_signed_int(buf[7]);
			buf_itemdata[item_no].param.weapon.increase_wisdom     = ff4_signed_int(buf[8]);
			buf_itemdata[item_no].param.weapon.increase_will_power = ff4_signed_int(buf[9]);
			break;

		case ITEM_TYPE_PROTECTOR:
			/* ͷϥǡ */
			buf_itemdata[item_no].param.protector.defense_power       = buf[0];
			buf_itemdata[item_no].param.protector.evade_rate          = buf[2] & 0x7f;
			buf_itemdata[item_no].param.protector.magic_defense_power = buf[3];
			buf_itemdata[item_no].param.protector.magic_evade_rate    = buf[4] & 0x7f;

			/* °Ѳϥǡ */
			buf_itemdata[item_no].param.protector.bits_equip_chr      = (buf[11] << 8) | buf[10];

			/* ѥ᡼Ѳǡ */
			buf_itemdata[item_no].param.protector.increase_strength   = ff4_signed_int(buf[5]);
			buf_itemdata[item_no].param.protector.increase_speed      = ff4_signed_int(buf[6]);
			buf_itemdata[item_no].param.protector.increase_stamina    = ff4_signed_int(buf[7]);
			buf_itemdata[item_no].param.protector.increase_wisdom     = ff4_signed_int(buf[8]);
			buf_itemdata[item_no].param.protector.increase_will_power = ff4_signed_int(buf[9]);
			break;
		}
	}

	return buf_itemdata;
}

int item_writeparam(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥοͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥοͷϥǡϤ() */
	fprintf(fp, "No.       ̾               "
				"   "
				" ̿Ψ               "
				"     ®   \n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		if (ptr_itemdata[i].type == ITEM_TYPE_WEAPON)
		{
			fprintf(fp, "%03d(%03X): %-16s : %6ld, %5d, %5d,               %3d, %3d, %3d, %3d, %3d\n", i, i,
					item_getname(i),
					ptr_itemdata[i].price,
					ptr_itemdata[i].param.weapon.attack_power,
					ptr_itemdata[i].param.weapon.hit_rate,
					ptr_itemdata[i].param.weapon.increase_strength,
					ptr_itemdata[i].param.weapon.increase_speed,
					ptr_itemdata[i].param.weapon.increase_stamina,
					ptr_itemdata[i].param.weapon.increase_wisdom,
					ptr_itemdata[i].param.weapon.increase_will_power);
		}
	}

	/* ƥοͷϥǡϤ(ɶ) */
	fprintf(fp, "No.       ̾               "
				"   "
				"ɸ Ψ  Ψ "
				"     ®   \n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		if (ptr_itemdata[i].type == ITEM_TYPE_PROTECTOR)
		{
			fprintf(fp, "%03d(%03X): %-16s : %6ld, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d, %3d\n", i, i,
					item_getname(i),
					ptr_itemdata[i].price,
					ptr_itemdata[i].param.protector.defense_power,
					ptr_itemdata[i].param.protector.evade_rate,
					ptr_itemdata[i].param.protector.magic_defense_power,
					ptr_itemdata[i].param.protector.magic_evade_rate,
					ptr_itemdata[i].param.protector.increase_strength,
					ptr_itemdata[i].param.protector.increase_speed,
					ptr_itemdata[i].param.protector.increase_stamina,
					ptr_itemdata[i].param.protector.increase_wisdom,
					ptr_itemdata[i].param.protector.increase_will_power);
		}
	}

	/* ƥοͷϥǡϤ(ƻ) */
	fprintf(fp, "No.       ̾               "
				"  \n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		if (ptr_itemdata[i].type == ITEM_TYPE_TOOL)
		{
			fprintf(fp, "%03d(%03X): %-16s : %6ld\n", i, i,
					item_getname(i),
					ptr_itemdata[i].price);
		}
	}

	/* ƥοͷϥǡϤ(¾) */
	fprintf(fp, "No.       ̾               "
				"  \n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		if (ptr_itemdata[i].type == ITEM_TYPE_OTHER)
		{
			fprintf(fp, "%03d(%03X): %-16s : %6ld\n", i, i,
					item_getname(i),
					ptr_itemdata[i].price);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeelem(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥ°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥ°ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"°\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			fprintf(fp, " ");

			/* ° */
			str_printelem(fp, ptr_itemdata[i].param.weapon.elem_attack, ELEM_ALL);
			break;

		case ITEM_TYPE_PROTECTOR:
			fprintf(fp, " ");

			/* °ؤ */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_defense, ELEM_ALL);
			break;

		case ITEM_TYPE_TOOL:
			fprintf(fp, " ");

			/* ° */
			str_printelem(fp, ptr_itemdata[i].param.tool.elem_effect, ELEM_ALL);
			break;
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writestate(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥξѲϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"Ѳ1        2                3    4\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			fprintf(fp, " ");

			/* ̾ﹶξѲ */
			str_printstate(fp, ptr_itemdata[i].param.weapon.state_attack, STATE_ALL & ~STATE_OTHER);
			break;

		case ITEM_TYPE_PROTECTOR:
			fprintf(fp, " ");

			/* Ѳؤ */
			str_printstate(fp, ptr_itemdata[i].param.protector.state_defense, STATE_ALL & ~STATE_OTHER);
			break;

		case ITEM_TYPE_TOOL:
			fprintf(fp, " ");

			/* Ѳ */
			str_printstate(fp, ptr_itemdata[i].param.tool.state_effect, STATE_ALL);
			break;
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeflag(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "..";
	const char *str_chrmain[] =
	{
		"CD",	/* (Ź) */
		"Ca",	/*  */
		"Ro",	/*  */
		"RC",	/* ǥ(Ҷ) */
		"Ci",	/*  */
		"Te",	/* ƥ */
		"Gi",	/* С */
		"Ya",	/*  */

		"Pa",	/* ѥ */
		"Po",	/* ݥ */
		"Ed",	/* å */
		"Fu",	/* ա */
		"CP",	/* (ѥǥ) */
		"RA",	/* ǥ() */
		"Go",	/* ١ */
		"??",	/*  */
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥΥե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥΥե饰ϥǡϤ() */
	fprintf(fp, "No.       ̾               "
				" ꤲ  ξ  "
				"ǽ饯                               "
				"ͭ²\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		if (ptr_itemdata[i].type == ITEM_TYPE_WEAPON || ptr_itemdata[i].type == ITEM_TYPE_PROTECTOR)
		{
			fprintf(fp, "%03d(%03X): %-16s :", i, i, item_getname(i));

			switch (ptr_itemdata[i].type)
			{
			case ITEM_TYPE_WEAPON:
				/* Ƽե饰 */
				fprintf(fp, "  %2s   %2s   %2s   %2s  ",
						str_getflag(ptr_itemdata[i].param.weapon.flag_magnetic),
						str_getflag(ptr_itemdata[i].param.weapon.flag_throw),
						str_getflag(ptr_itemdata[i].param.weapon.flag_backrow),
						str_getflag(ptr_itemdata[i].param.weapon.flag_2handed));

				/* ǽ饯 */
				for (j = 0; j < 16; j++)
				{
					fprintf(fp, " %-2s",
							((ptr_itemdata[i].param.weapon.bits_equip_chr & (1U << j)) != 0x00) ?
							str_chrmain[j] : str_nothing);
				}

				/* ͭ󥹥 */
				if (ptr_itemdata[i].param.weapon.montype_attack != 0x00)
				{
					fprintf(fp, "  ");
					str_printmontype(fp, ptr_itemdata[i].param.weapon.montype_attack, MONTYPE_ALL);
				}
				break;

			case ITEM_TYPE_PROTECTOR:
				/* Ƽե饰 */
				fprintf(fp, "  %2s                 ",
						str_getflag(ptr_itemdata[i].param.protector.flag_magnetic));

				/* ǽ饯 */
				for (j = 0; j < 16; j++)
				{
					fprintf(fp, " %-2s",
							((ptr_itemdata[i].param.protector.bits_equip_chr & (1U << j)) != 0x00) ?
							str_chrmain[j] : str_nothing);
				}

				/* ͭ󥹥 */
				if (ptr_itemdata[i].param.protector.montype_defense != 0x00)
				{
					fprintf(fp, "  ");
					str_printmontype(fp, ptr_itemdata[i].param.protector.montype_defense, MONTYPE_ALL);
				}
				break;
			}

			fputc('\n', fp);
		}
	}

	/* ƥΥե饰ϥǡϤ(ʰʳ) */
	fprintf(fp, "No.       ̾               "
				" ϰ\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		if (ptr_itemdata[i].type == ITEM_TYPE_TOOL)
		{
			fprintf(fp, "%03d(%03X): %-16s :", i, i, item_getname(i));

			/* ϰ */
			fprintf(fp, "  ");
			mag_printtarget(fp, &ptr_itemdata[i].param.tool);

			fputc('\n', fp);
		}
	}

	/* ƥΥե饰ϥǡϤ(¾) */
	fprintf(fp, "No.       ̾\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		if (ptr_itemdata[i].type == ITEM_TYPE_OTHER)
			fprintf(fp, "%03d(%03X): %-16s :\n", i, i, item_getname(i));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeeffect(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥθ̥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥθ̥ǡϤ */
	fprintf(fp, "No.      ̾               "
				"Ѹ\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : %-16s", i, i,
					item_getname(i),
					(ptr_itemdata[i].effect_no != 0x00) ?
					mag_getname(ptr_itemdata[i].effect_no) : str_nothing);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
