/*
 *	ff4magic.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4file.h"
#include "ff4decode.h"
#include "ff4magic.h"
#include "ff4item.h"

#define FILEPOS_MAGDATA			0x0e68b0		/* SNES: 0x0797a0 */
#define NUM_MAG					307				/* SNES: 256 */
#define SIZEOF_MAGDATA			16				/* SNES: 6 */
#define MSGID_MAGNAME			4853

static void mag_setname(void);

static char tbl_magname[NUM_MAG][256];	/* ˡ̾Ǽơ֥ */
static int flag_setmagname = 0;			/* tbl_magname˥ǡǼƤ뤫ɤ */

const char *mag_getname(int index)
{
	/* å */
	if (index < 0 || NUM_MAG <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmagname)
		mag_setname();

	return tbl_magname[index];
}

static void mag_setname(void)
{
	int i;

	/* ˡ̾ǡ */
	for (i = 0; i < NUM_MAG; i++)
		*tbl_magname[i] = '\0';

	/* ˡ̾ǡɤ߹ */
	for (i = 0; i < 174; i++)
		dec_msg(tbl_magname[i], sizeof(tbl_magname[i]), MSGID_MAGNAME + i, FLAG_KANA);

	/* ˡ̾ǡ (ƥ̾) ɤ߹ */
	for (i = 176; i < 252; i++)
		strncpy(tbl_magname[i], item_getname(i), sizeof(tbl_magname[i]));

	/* ˡ̾ǡ ("ΤL*") ɤ߹ */
	for (i = 256; i < 265; i++)
		strncpy(tbl_magname[i], item_getname(269 + (i - 256)), sizeof(tbl_magname[i]));

	/* ˡ̾ǡ ("饹ȥꥯ") ɤ߹ */
	for (i = 265; i < 266; i++)
		strncpy(tbl_magname[i], item_getname(322), sizeof(tbl_magname[i]));

	flag_setmagname = 1;
}

magdata_t *mag_readdata(void)
{
	int i;
	magdata_t *buf_magdata = NULL;

	/* ˡγƼǡǼХåեݤ */
	if ((buf_magdata = malloc(sizeof(magdata_t) * NUM_MAG)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ˡγƼǡɤ߹ */
	for (i = 0; i < NUM_MAG; i++)
	{
		unsigned char buf[SIZEOF_MAGDATA];

		file_read(buf, FILEPOS_MAGDATA + SIZEOF_MAGDATA * i, SIZEOF_MAGDATA);
		mag_setparam(&buf_magdata[i], buf, sizeof(buf));
	}

	return buf_magdata;
}

int mag_setparam(magdata_t *ptr_magdata, unsigned char *ptr_buf, size_t bufsize)
{
	/* å */
	if (ptr_magdata == NULL || ptr_buf == NULL || bufsize < SIZEOF_MAGDATA)
		return 0;

	/* ͷϥǡ */
	ptr_magdata->magic_point  = ptr_buf[0];
	ptr_magdata->hit_rate     = ptr_buf[3];
	ptr_magdata->delay_time   = ptr_buf[4];
	ptr_magdata->effect_power = 0;
	ptr_magdata->effect_rate  = 0;

	/* Ψ᡼ */
	if (ptr_buf[2] != 0x00)
		ptr_magdata->effect_rate  = ptr_buf[1];
	else
		ptr_magdata->effect_power = ptr_buf[1];

	/* ե饰ϥǡ */
	ptr_magdata->flag_boss_noeffect = ((ptr_buf[5] & 0x80) == 0x00);
	ptr_magdata->flag_once          = ((ptr_buf[5] & 0x10) != 0x00);
	ptr_magdata->flag_reflect       = 0;	/* XXX */

	/* ϰϥǡ */
	ptr_magdata->flag_range_self    = 0;
	ptr_magdata->direction          = 0;
	ptr_magdata->change_direction   = 0;
	ptr_magdata->range              = 0;
	ptr_magdata->change_range       = 0;

	if (ptr_buf[12] == 0x08)
	{
		/* 0x08 = "Ũ" */
		ptr_magdata->direction = 1;
		ptr_magdata->range     = 1;
	}
	else if (ptr_buf[12] == 0x20)
	{
		/* 0x20 = "ʬ" */
		ptr_magdata->flag_range_self = 1;
	}
	else
	{
		ptr_magdata->direction        = ((ptr_buf[12] & 0x80) != 0x00) ? 1 : 0;
		ptr_magdata->change_direction = ((ptr_buf[12] & 0x04) != 0x00) ? 1 : 0;
		ptr_magdata->range            = ((ptr_buf[12] & 0x02) != 0x00) ? 1 : 0;
		ptr_magdata->change_range     = ((ptr_buf[12] & 0x03) == 0x03) ? 1 : 0;
	}

	/* °Ѳϥǡ */
	ptr_magdata->elem_effect  = ptr_buf[6];
	ptr_magdata->state_effect = bits_make32(ptr_buf[8], ptr_buf[9], ptr_buf[10], ptr_buf[11]);

	return 1;
}

int mag_writeparam(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡοͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡοͷϥǡϤ */
	fprintf(fp, "No.      ̾               "
				"MP    Ψ ̿Ψ   \n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %5d, %5d, %2s%3d, %5d, %5d\n", i, i,
				mag_getname(i),
				ptr_magdata[i].magic_point,
				ptr_magdata[i].effect_power,
				(ptr_magdata[i].effect_rate != 0) ? "1/" : "  ",
				ptr_magdata[i].effect_rate,
				ptr_magdata[i].hit_rate,
				ptr_magdata[i].delay_time);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeelem(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡ°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡ°ϥǡϤ */
	fprintf(fp, "No.      ̾               "
				"°\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i, mag_getname(i));

		/* ° */
		str_printelem(fp, ptr_magdata[i].elem_effect, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writestate(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡξѲϥǡϤ */
	fprintf(fp, "No.      ̾               "
				"Ѳ1        2                3    4\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i, mag_getname(i));

		/* Ѳ */
		str_printstate(fp, ptr_magdata[i].state_effect, STATE_ALL /* & ~STATE_OTHER */);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeflag(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡΥե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡΥե饰ϥǡϤ */
	fprintf(fp, "No.      ̾               "
				"ܥ 1  ȿ  "
				"ϰ\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %2s   %2s   %2s   ", i, i, mag_getname(i),
				str_getflag(!ptr_magdata[i].flag_boss_noeffect),
				str_getflag(ptr_magdata[i].flag_once),
				str_getflag(ptr_magdata[i].flag_reflect));

		mag_printtarget(fp, &ptr_magdata[i]);

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

void mag_printtarget(FILE *stream, const magdata_t *ptr_magdata)
{
	const char *str_direction[2][2] =
	{
		{"̣",	"̣/Ũ"},
		{"Ũ",		"Ũ/̣"},
	};
	const char *str_range[2][2] =
	{
		{"ñ",	"ñ"},
		{"",	"ñ/"},
	};
	const char *str_range_self = "ʬ";

	/* å */
	if (stream == NULL || ptr_magdata == NULL)
	{
		assert(0);
		return;
	}

	/* ˡθϰϥǡϤ */
	if (ptr_magdata->flag_range_self)
		fprintf(stream, "%-17s", str_range_self);
	else
	{
		fprintf(stream, "%-7s %-9s",
				str_direction[ptr_magdata->direction][ptr_magdata->change_direction],
				str_range[ptr_magdata->range][ptr_magdata->change_range]);
	}
}
