/*
 *	ff4monster.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "errmes.h"
#include "ff4file.h"
#include "ff4decode.h"
#include "ff4character.h"
#include "ff4monster.h"

#define FILEPOS_MONDATA			0x0f94e8		/* SNES: 0x072860 */
#define FILEPOS_MONITEM			0x0fb484		/* SNES: 0x071f00 */
#define FILEPOS_MONELEM			0x0fc210
#define FILEPOS_MONACT			0x000000
#define FILEPOS_MONPARTY_FLAG	0x0fcb18
#define FILEPOS_MONPARTY_ADDR	0x0fe290
#define FILEPOS_MONPARTY	 	0x0fee4c 		/* SNES: 0x070000 */
#define FILEPOS_MONENC			0x0da760		/* SNES: 0x074796 */
#define FILEPOS_MONMAP1			0x0db42c		/* SNES: 0x074542 */
#define FILEPOS_MONMAP2			0x0dbcac
#define NUM_MON					288				/* SNES: 224 */
#define NUM_MONNAME				(NUM_MON - 5)	/* = 283 */
#define NUM_MONMES				186
#define NUM_MONPARTY			751				/* SNES: 512 */
#define NUM_MONENC				150				/* SNES: 96 */
#define NUM_MONMAP				3
#define SIZEOF_MONDATA			28
#define SIZEOF_MONITEM			12				/* SNES: 4 */
#define SIZEOF_MONELEM			8
#define SIZEOF_MONPARTY_FLAG	8
#define SIZEOF_MONPARTY_ADDR	4
#define SIZEOF_MONPARTY			8
#define SIZEOF_MONENC			20				/* SNES: 8 */
#define SIZEOF_MONMAP			16				/* SNES: 1 */
#define MSGID_MONNAME			4395
#define MSGID_MONMES			266

static void mon_setname(void);
static void mon_setmes(void);
static const char *mon_getmes(int index);
#if 0
static const char *mon_gettarget(int index);
static unsigned char *mon_printact(FILE *stream, unsigned char *buf_action);
#endif

static char tbl_monname[NUM_MON][256];			/* 󥹥̾Ǽơ֥ */
static char tbl_monmes[NUM_MONMES][256];		/* 󥹥ΥåǼơ֥ */
static int flag_setmonname    = 0;				/* tbl_monname˥ǡǼƤ뤫ɤ */
static int flag_setmonmes     = 0;				/* tbl_monmes˥ǡǼƤ뤫ɤ */

static const int tbl_sizeof_map[NUM_MONMAP] = {8, 4, 1};

const char *mon_getname(int index)
{
	/* å */
	if (index < 0 || NUM_MON <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmonname)
		mon_setname();

	return tbl_monname[index];
}

static const char *mon_getmes(int index)
{
	/* å */
	if (index < 0 || NUM_MONMES <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmonmes)
		mon_setmes();

	return tbl_monmes[index];
}

#if 0
static const char *mon_gettarget(int index)
{
}
#endif

static void mon_setname(void)
{
	int i;

	/* 󥹥̾ǡ */
	for (i = 0; i < NUM_MON; i++)
		*tbl_monname[i] = '\0';

	/* 󥹥̾ǡɤ߹ */
	for (i = 0; i < NUM_MONNAME; i++)
		dec_msg(tbl_monname[i], sizeof(tbl_monname[i]), MSGID_MONNAME + i, FLAG_KANA);

	flag_setmonname = 1;
}

static void mon_setmes(void)
{
	int i;
	char *ptr;

	/* 󥹥Υåǡ */
	for (i = 0; i < NUM_MONMES; i++)
		*tbl_monmes[i] = '\0';

	/* 󥹥Υåǡɤ߹ */
	for (i = 0; i < NUM_MONMES; i++)
	{
		dec_msg(tbl_monmes[i], sizeof(tbl_monmes[i]), MSGID_MONMES + i, FLAG_KANJI);

		/* Ԥ򵭹֤ */
		while ((ptr = strchr(tbl_monmes[i], '\n')) != NULL)
			*ptr = '/';
	}

	flag_setmonmes = 1;
}

mondata_t *mon_readdata(void)
{
	int i, j;
	mondata_t *buf_mondata = NULL;

	/* 󥹥γƼǡǼХåեݤ */
	if ((buf_mondata = malloc(sizeof(mondata_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥γƼǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONDATA];

		file_read(buf, FILEPOS_MONDATA + SIZEOF_MONDATA * i, SIZEOF_MONDATA);

		/* ͷϥǡ */
		buf_mondata[i].level               = buf[ 0];
		buf_mondata[i].hit_point           = bits_make32(buf[4], buf[5], buf[6], buf[7]);
		buf_mondata[i].experience_point    = (buf[ 9] << 8) | buf[ 8];
		buf_mondata[i].gil                 = (buf[11] << 8) | buf[10];
		buf_mondata[i].attack_times        = buf[12];
		buf_mondata[i].attack_power        = buf[13];
		buf_mondata[i].hit_rate            = buf[14];
		buf_mondata[i].defense_times       = buf[15];
		buf_mondata[i].defense_power       = buf[16];
		buf_mondata[i].evade_rate          = buf[17];
		buf_mondata[i].magic_defense_times = buf[18];
		buf_mondata[i].magic_defense_power = buf[19];
		buf_mondata[i].magic_evade_rate    = buf[20];
		buf_mondata[i].speed[0]            = buf[21];
		buf_mondata[i].speed[1]            = buf[22];
		buf_mondata[i].speed_rate          = buf[23];
		buf_mondata[i].magic_power         = buf[25];
		buf_mondata[i].flag_boss           = ((buf[26] & 0x01) != 0x00);
		buf_mondata[i].montype             = buf[27];

		/* ѥ᡼黻ФǤǡ */
		buf_mondata[i].magic_point        = buf_mondata[i].hit_point / 8;
		buf_mondata[i].magic_attack_times = 1 + buf_mondata[i].magic_power / 4;
	}

	/* 󥹥νƥǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONITEM];

		file_read(buf, FILEPOS_MONITEM + SIZEOF_MONITEM * i, SIZEOF_MONITEM);

		buf_mondata[i].item_get       = buf[0];
		for (j = 0; j < 4; j++)
			buf_mondata[i].item_no[j] = (buf[2 + j * 2 + 1] << 8) | buf[2 + j * 2];
	}

	/* 󥹥°Ѳϥǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONELEM];

		file_read(buf, FILEPOS_MONELEM + SIZEOF_MONELEM * i, SIZEOF_MONELEM);

		/* °ϥǡ */
		buf_mondata[i].elem_defense  = buf[0];
		buf_mondata[i].elem_weakness = buf[1];

		/* Ѳϥǡ */
		buf_mondata[i].state_attack  = (buf[3] << 8) | buf[2];
		buf_mondata[i].state_defense = (buf[5] << 8) | buf[4];
	}

	return buf_mondata;
}

monact_t *mon_readact(void)
{
	int i, j, k;
	monact_t *buf_monact = NULL;

	/* 󥹥ιưѥǡǼХåեݤ */
	if ((buf_monact = malloc(sizeof(monact_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	for (i = 0; i < NUM_MON; i++)
	{
		for (j = 0; j < (int)(sizeof(buf_monact[i].ptn) / sizeof(buf_monact[i].ptn[0])); j++)
		{
			buf_monact[i].ptn[j].action_size = 0;
			buf_monact[i].ptn[j].action = NULL;

			for (k = 0; k < 3; k++)
				buf_monact[i].ptn[j].condition_no[k] = -1;
		}
	}

	return buf_monact;
}

monparty_t *mon_readparty(void)
{
	int i, j;
	unsigned long tbl_monparty_addr[NUM_MONPARTY];
	monparty_t *buf_monparty = NULL;

	/* 󥹥ΥѡƥǡǼХåեݤ */
	if ((buf_monparty = malloc(sizeof(monparty_t) * NUM_MONPARTY)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥ΥѡƥǡǼƤ륢ɥ쥹ɤ߹ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY_ADDR];

		file_read(buf, FILEPOS_MONPARTY_ADDR + SIZEOF_MONPARTY_ADDR * i, SIZEOF_MONPARTY_ADDR);
		tbl_monparty_addr[i] = bits_make32(buf[0], buf[1], buf[2], buf[3]);
	}

	/* 󥹥ΥѡƥγƼǡɤ߹ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY];
		const unsigned long filepos_monparty = tbl_monparty_addr[i];
		int index;

		buf_monparty[i].monster_num = 0;

		for (index = 0; index < 8; index++)
		{
			/* ѡƥǡüƤ뤫ɤ */
			file_read(buf, filepos_monparty + SIZEOF_MONPARTY * index, 1);

			/* 0xff = ü */
			if (buf[0] == 0xff)
				break;

			file_read(buf + 1, filepos_monparty + SIZEOF_MONPARTY * index + 1, SIZEOF_MONPARTY - 1);

			buf_monparty[i].monster_num += 1;

			/* ͷϥǡ */
			buf_monparty[i].group_no[index]   = buf[0] & 0x03;
			buf_monparty[i].monster_no[index] = (buf[3] << 8) | buf[2];
			buf_monparty[i].position_x[index] = buf[4];
			buf_monparty[i].position_y[index] = buf[5];

			/* ե饰ϥǡ */
			buf_monparty[i].flag_eggy[index]  = ((buf[0] & 0x10) != 0x00);
		}
	}

	/* 󥹥ΥѡƥΥե饰ϥǡɤ߹ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY_FLAG];

		file_read(buf, FILEPOS_MONPARTY_FLAG + SIZEOF_MONPARTY_FLAG * i, SIZEOF_MONPARTY_FLAG);

		/* ͷϥǡ */
		buf_monparty[i].background_no    = buf[0];
		buf_monparty[i].music_no         = (buf[4] & 0xf0) >> 4;

		/* ե饰ϥǡ */
		buf_monparty[i].flag_preemptive = ((buf[2] & 0x01) != 0x00);
		buf_monparty[i].flag_surprise   = ((buf[2] & 0x02) != 0x00);
		buf_monparty[i].flag_backattack = ((buf[2] & 0x04) != 0x00);
		buf_monparty[i].flag_float      = ((buf[2] & 0x08) != 0x00);
		buf_monparty[i].flag_event      = ((buf[2] & 0x20) != 0x00);
		buf_monparty[i].flag_escape     = ((buf[3] & 0x80) == 0x00);

		/* 󥹥νи */
		buf_monparty[i].flag_exist[0]   = 0;
		buf_monparty[i].flag_exist[1]   = ((buf[4] & 0x01) != 0x00);
		buf_monparty[i].flag_exist[2]   = ((buf[4] & 0x02) != 0x00);
	}

	/* ϤθߴΤ˥󥹥ΥѡƥΥ롼ץǡǼ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		for (j = 0; j < 3; j++)
		{
			int k;

			buf_monparty[i].grp[j].num       = 0;
			buf_monparty[i].grp[j].index     = 0;
			buf_monparty[i].grp[j].flag_eggy = 0;

			for (k = 0; k < (int)sizeof(buf_monparty[i].grp[j].monster_no); k++)
				buf_monparty[i].grp[j].monster_no[j] = 0;
		}

		for (j = 0; j < buf_monparty[i].monster_num; j++)
		{
			const int index = buf_monparty[i].group_no[j];

			if (buf_monparty[i].grp[index].num == 0)
			{
				buf_monparty[i].grp[index].monster_no[0] = buf_monparty[i].monster_no[j];
			}
			else if (buf_monparty[i].grp[index].monster_no[0] != buf_monparty[i].monster_no[j])
			{
				const int num = ++buf_monparty[i].grp[index].index;

				buf_monparty[i].grp[index].monster_no[num] = buf_monparty[i].monster_no[j];
			}

			if (buf_monparty[i].flag_eggy[j])
				buf_monparty[i].grp[index].flag_eggy = 1;

			buf_monparty[i].grp[index].num += 1;
		}
	}

	return buf_monparty;
}

monenc_t *mon_readenc(void)
{
	int i, j;
	monenc_t *buf_monenc = NULL;

	/* 󥹥ơ֥ǡǼХåեݤ */
	if ((buf_monenc = malloc(sizeof(monenc_t) * NUM_MONENC)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥ơ֥ǡɤ߹ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		unsigned char buf[SIZEOF_MONENC];

		file_read(buf, FILEPOS_MONENC + SIZEOF_MONENC * i, SIZEOF_MONENC);

		for (j = 0; j < 8; j++)
			buf_monenc[i].party_no[j] = (buf[j * 2 + 1] << 8) | buf[j * 2];
	}

	return buf_monenc;
}

monmap_t *mon_readmap(void)
{
	int i, j, x, y;
	long filepos = 0;
	monmap_t *buf_monmap = NULL;

	/* 󥹥ޥåץǡǼХåեݤ */
	if ((buf_monmap = malloc(sizeof(monmap_t) * NUM_MONMAP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥ޥåץǡɤ߹ */
	for (i = 0; i < NUM_MONMAP; i++)
	{
		if (i == 0)
			filepos = FILEPOS_MONMAP1;
		else if (i == 1)
			filepos = FILEPOS_MONMAP2;
		/* else if (i == 2) */
		/* FILEPOS_MONMAP2 ³ɤ߹ */

		for (y = 0; y < tbl_sizeof_map[i]; y++)
		{
			for (x = 0; x < tbl_sizeof_map[i]; x++)
			{
				unsigned char buf[SIZEOF_MONMAP];

				filepos += file_read(buf, filepos, SIZEOF_MONMAP);
				for (j = 0; j < 8; j++)
					buf_monmap[i].area[x][y].party_no[j] = (buf[j * 2 + 1] << 8) | buf[j * 2];
			}
		}
	}

	return buf_monmap;
}

int mon_writeparam(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥οͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥οͷϥǡϤ */
	fprintf(fp, "No.       ̾                "
				" LV      HP     MP    EXP    "
				"  ̿Ψ "
				"ɸ ɸ Ψ "
				"ɲ  Ψ "
				"⹶    "
				"  ® ®\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :  %3d, %6ld, %5ld, %5ld, %5ld, "
					"%5d, %5d, %5d, "
					"%5d, %5d, %5d, "
					"%5d, %5d, %5d, "
					"%5d, %5d, "
					"%3d-%3d,%3d/16\n", i, i,
				mon_getname(i),
				ptr_mondata[i].level,
				ptr_mondata[i].hit_point,
				ptr_mondata[i].magic_point,
				ptr_mondata[i].experience_point,
				ptr_mondata[i].gil,
				ptr_mondata[i].attack_times,
				ptr_mondata[i].attack_power,
				ptr_mondata[i].hit_rate,
				ptr_mondata[i].defense_times,
				ptr_mondata[i].defense_power,
				ptr_mondata[i].evade_rate,
				ptr_mondata[i].magic_defense_times,
				ptr_mondata[i].magic_defense_power,
				ptr_mondata[i].magic_evade_rate,
				ptr_mondata[i].magic_attack_times,
				ptr_mondata[i].magic_power,
				ptr_mondata[i].speed[0],
				ptr_mondata[i].speed[1],
				ptr_mondata[i].speed_rate);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeelem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥°ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"              "
				"\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));

		/* °ؤ */
		str_printelem(fp, ptr_mondata[i].elem_defense, ELEM_ALL);
		fprintf(fp, "  ");

		/* ° */
		str_printelem(fp, ptr_mondata[i].elem_weakness, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writestate(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ξѲϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"ܥ  "
				"Ѳ1    2                 "
				"Ѳ1    2\n");
	for (i = 0; i < NUM_MON; i++)
	{
		/* ܥǤ뤫ɤ */
		fprintf(fp, "%03d(%03X): %-16s :  %2s   ", i, i, mon_getname(i),
				str_getflag(ptr_mondata[i].flag_boss));

		/* ̾ﹶξѲ */
		str_printstate(fp, ptr_mondata[i].state_attack, STATE_ALL & ~STATE_OTHER);
		fprintf(fp, "  ");

		/* Ѳؤ */
		str_printstate(fp, ptr_mondata[i].state_defense, STATE_ALL & ~STATE_OTHER);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeflag(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Υե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Υե饰ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"ܥ  ²\n");
	for (i = 0; i < NUM_MON; i++)
	{
		/* Ƽե饰 */
		fprintf(fp, "%03d(%03X): %-16s :  %2s", i, i, mon_getname(i),
				str_getflag(ptr_mondata[i].flag_boss));

		/* 󥹥 */
		if ((ptr_mondata[i].montype & MONTYPE_ALL) != 0x00)
		{
			fprintf(fp, "   ");
			str_printmontype(fp, ptr_mondata[i].montype, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeitem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥νƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥νƥǡϤ */
	fprintf(fp, "No.       ̾               "
				"Ψ   "
				"ƥ1         "
				"ƥ2         "
				"ƥ3         "
				"ƥ4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :  %3d,  ", i, i, mon_getname(i), ptr_mondata[i].item_get);
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", item_getname(ptr_mondata[i].item_no[j]));
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeact(const char *ptr_filename, const monact_t *ptr_monact, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monact == NULL || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ιưѥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ιưѥǡϤ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X): %s]\n", i, i, mon_getname(i));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemes(const char *ptr_filename)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0')
	{
		assert(0);
		return 0;
	}

	/* 󥹥ΥåǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ΥåǡϤ */
	fprintf(fp, "No.      å\n");
	for (i = 0; i < NUM_MONMES; i++)
	{
		fprintf(fp, "%03d(%02X): ", i, i);
		fputs(mon_getmes(i), fp);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeparty(const char *ptr_filename, const monparty_t *ptr_monparty)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ΥѡƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ΥѡƥǡϤ */
	fprintf(fp, "No.       "
				"ƨ  G2 G3 "
				"ߥƥ쥰        "
				" "
				"롼1             "
				"롼2             "
				"롼3\n");
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		fprintf(fp, "%03d(%03X):  %2s   %2s   %2s   %2s  ", i, i,
				str_getflag(ptr_monparty[i].flag_escape),
				str_getflag(ptr_monparty[i].flag_event),
				str_getflag(ptr_monparty[i].flag_exist[1]),
				str_getflag(ptr_monparty[i].flag_exist[2]));

		for (j = 0; j < 8; j++)
			fprintf(fp, "%2s", str_getflag(ptr_monparty[i].flag_eggy[j]));

		fprintf(fp, "  ");

		mon_printparty(fp, &ptr_monparty[i]);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeenc(const char *ptr_filename, const monenc_t *ptr_monenc, const monparty_t *ptr_monparty)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monenc == NULL || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ơ֥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ơ֥Ϥ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X)]\n", i, i);
		for (j = 0; j < 8; j++)
		{
			const int monparty_no = ptr_monenc[i].party_no[j];

			fprintf(fp, "%02d: %03d(%03X): ", j + 1, monparty_no, monparty_no);

			mon_printparty(fp, &ptr_monparty[monparty_no]);
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemap(const char *ptr_filename, const monmap_t *ptr_monmap)
{
	int i, j, x, y;
	FILE *fp = NULL;
	const char *str_name[NUM_MONMAP] =
	{
		"Ͼ",
		"",
		"",
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monmap == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ޥåץǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ޥåפϤ */
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < tbl_sizeof_map[i]; y++)
		{
			if (i != 0 || y != 0)
				fputc('\n', fp);

			fprintf(fp, "[No.%1d-%1d: %s]\n", i, y + 1, str_name[i]);
			for (x = 0; x < tbl_sizeof_map[i]; x++)
			{
				fprintf(fp, "%d: ", x + 1);
				for (j = 0; j < 8; j++)
				{
					if (j != 0)
						fprintf(fp, " ");

					fprintf(fp, "%03d", ptr_monmap[i].area[x][y].party_no[j]);
				}
				fputc('\n', fp);
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

#if 0
static unsigned char *mon_printact(FILE *stream, unsigned char *buf_action)
{
}
#endif

void mon_freeact(monact_t *ptr_monact)
{
	int i, j;

	/* å */
	if (ptr_monact == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_MON; i++)
	{
		for (j = 0; j < (int)(sizeof(ptr_monact[i].ptn) / sizeof(ptr_monact[i].ptn[0])); j++)
		{
			if (ptr_monact[i].ptn[j].action != NULL)
			{
				free(ptr_monact[i].ptn[j].action);
				ptr_monact[i].ptn[j].action = NULL;
			}
		}
	}
}

void mon_printparty(FILE *stream, const monparty_t *ptr_monparty)
{
	int i;
	const char *str_nothing = "--------------------";

	/* å */
	if (stream == NULL || ptr_monparty == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < 3; i++)
	{
		if (ptr_monparty->grp[i].num != 0)
		{
			const int flag_absent = (ptr_monparty->grp[i].flag_eggy || ptr_monparty->flag_exist[i]);

			if (i != 0)
				fprintf(stream, ", ");

			if (ptr_monparty->grp[i].index != 0)
			{
				int index;

				fprintf(stream, "%s", (flag_absent) ? "((" : "[[");

				for (index = 0; index <= ptr_monparty->grp[i].index; index++)
				{
					if (index != 0)
						fprintf(stream, ", ");
					fprintf(stream, " %-16s ", mon_getname(ptr_monparty->grp[i].monster_no[index]));
				}

				fprintf(stream, "%s", (flag_absent) ? "))" : "]]");
			}
			else
			{
				fprintf(stream, "%s%-16s%s %1d",
						(flag_absent) ? "(" : " ",
						mon_getname(ptr_monparty->grp[i].monster_no[0]),
						(flag_absent) ? ")" : " ",
						ptr_monparty->grp[i].num);
			}
		}
		else if ((i < 2 && ptr_monparty->grp[i + 1].num) || (i < 1 && ptr_monparty->grp[i + 2].num))
		{
			if (i != 0)
				fprintf(stream, ", ");
			fprintf(stream, "%-20s", str_nothing);
		}
		else
			break;
	}
}
