/*
 *	ff5analyzer.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *	ff5a_analyzer
 *		Copyright (C) 2007-2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "utility.h"
#include "file.h"
#include "ff5decode.h"
#include "ff5monster.h"
#include "ff5map.h"
#include "ff5character.h"
#include "ff5item.h"
#include "ff5magic.h"
#include "ff5special.h"
#include "ff5misc.h"

#define FILENAME_MON_PARAM			"ff5monparam.txt"
#define FILENAME_MON_ELEM			"ff5monelem.txt"
#define FILENAME_MON_STATE			"ff5monstate.txt"
#define FILENAME_MON_FLAG			"ff5monflag.txt"
#define FILENAME_MON_ITEM			"ff5monitem.txt"
#define FILENAME_MON_FREE			"ff5monfree.txt"
#define FILENAME_MON_CTRL			"ff5monctrl.txt"
#define FILENAME_MON_TECH			"ff5montech.txt"
#define FILENAME_MON_WEAK			"ff5monweak.txt"
#define FILENAME_MON_MISC			"ff5monmisc.txt"
#define FILENAME_MON_ACT			"ff5monact.txt"
#define FILENAME_MON_MES			"ff5monmes.txt"
#define FILENAME_MON_PARTY			"ff5monparty.txt"
#define FILENAME_MON_PARTYFLAG		"ff5monpartyflag.txt"
#define FILENAME_MON_ENC			"ff5monenc.txt"
#define FILENAME_MON_MAP			"ff5monmap.txt"
#define FILENAME_MON_LOCATION		"ff5monlocation.txt"
#define FILENAME_CHR_PARAM			"ff5chrparam.txt"
#define FILENAME_CHR_EQUIP			"ff5chrequip.txt"
#define FILENAME_CHR_LVUP			"ff5chrlvup.txt"
#define FILENAME_CHR_JOB			"ff5chrjob.txt"
#define FILENAME_CHR_CMD_PARAM		"ff5chrcmdparam.txt"
#define FILENAME_CHR_CMD_ELEM		"ff5chrcmdelem.txt"
#define FILENAME_CHR_CMD_CATEGORY	"ff5chrcmdcategory.txt"
#define FILENAME_CHR_CMD_STATE		"ff5chrcmdstate.txt"
#define FILENAME_CHR_CMD_FLAG		"ff5chrcmdflag.txt"
#define FILENAME_CHR_ABT			"ff5chrabt.txt"
#define FILENAME_ITEM_PARAM			"ff5itemparam.txt"
#define FILENAME_ITEM_ELEM			"ff5itemelem.txt"
#define FILENAME_ITEM_CATEGORY		"ff5itemcategory.txt"
#define FILENAME_ITEM_STATE			"ff5itemstate.txt"
#define FILENAME_ITEM_FLAG			"ff5itemflag.txt"
#define FILENAME_ITEM_EFFECT		"ff5itemeffect.txt"
#define FILENAME_ITEM_MES			"ff5itemmes.txt"
#define FILENAME_MAG_PARAM			"ff5magparam.txt"
#define FILENAME_MAG_ELEM			"ff5magelem.txt"
#define FILENAME_MAG_CATEGORY		"ff5magcategory.txt"
#define FILENAME_MAG_STATE			"ff5magstate.txt"
#define FILENAME_MAG_FLAG			"ff5magflag.txt"
#define FILENAME_SP_PARAM			"ff5spparam.txt"
#define FILENAME_SP_ELEM			"ff5spelem.txt"
#define FILENAME_SP_CATEGORY		"ff5spcategory.txt"
#define FILENAME_SP_STATE			"ff5spstate.txt"
#define FILENAME_SP_FLAG			"ff5spflag.txt"
#define FILENAME_SP_LANDSCAPE		"ff5splandscape.txt"
#define FILENAME_SP_MIX				"ff5spmix.txt"
#define FILENAME_TREASURE			"ff5treasure.txt"
#define FILENAME_STORE				"ff5store.txt"
#define FILENAME_ITEM_PARAM_BUG		"ff5itemparambug.txt"
#define FILENAME_ITEM_ELEM_BUG		"ff5itemelembug.txt"
#define FILENAME_ITEM_STATE_BUG		"ff5itemstatebug.txt"
#define FILENAME_ITEM_FLAG_BUG		"ff5itemflagbug.txt"
#define FILENAME_MESSAGES			"ff5messages.txt"
#define FILENAME_DIALOGUE			"ff5dialogue.txt"
#define FILENAME_MAP_DATA			"ff5mapparam.txt"

#define STR_INVALID_OPTION			"Invalid option"
#define STR_INVALID_ROMIMAGE_FILE	"Invalid ROM-Image file: %s"
#define STR_INVALID_CDROM_FILES		"Invalid CD-ROM files"
#define STR_CANNOT_CREATE_FILE		"Cannot create file: %s"
#define STR_FILENAME_TOO_LONG		"Filename too long"

#define STR_READ_FROM_FILE			"Reading from %s\n"
#define STR_WRITE_TO_FILE			"Writing to %s\n"

/* ɽ */
static const char *str_copyright =
	"ff5analyzer   - Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>\n"
	"ff5a_analyzer - Modified by Inu";

static const char *tbl_gametitle[] =
{
	"FINAL FANTASY 5      ",
			/* 0x46, 0x49, 0x4E, 0x41, 0x4C, 0x20, 0x46, 0x41, 0x4E, 0x54, 0x41, 0x53, 0x59, 0x20, 0x35, */
			/* 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 */
	NULL,
			/* unused */
	"FF5ADVANCE\0\0",
			/* 0x46, 0x46, 0x35, 0x41, 0x44, 0x56, 0x41, 0x4E, 0x43, 0x45, 0x00, 0x00 */
};

void usage(void);

#define do_writedata(filename, function) do {        \
	char path[sizeof(str_dstpath) + 128] = "";       \
	snprintf(path, sizeof(path), "%s%s", str_dstpath, filename); \
	if (!flag_dry)                                    \
	{                                                 \
		fprintf(stderr, STR_WRITE_TO_FILE, path);     \
		if (!function)                                \
			err_printf(STR_CANNOT_CREATE_FILE, path); \
	}                                                 \
} while (0)

int main(int argc, char *argv[])
{
	int optind = 0;
	int flag_dry = 0;
	char str_srcpath[512] = "";
	char str_dstpath[512] = "";
	mondata_t   *buf_mondata   = NULL;
	monact_t    *buf_monact    = NULL;
	monparty_t  *buf_monparty  = NULL;
	monenc_t    *buf_monenc    = NULL;
	monmap_t    *buf_monmap    = NULL;
	mapdata_t   *buf_mapdata   = NULL;
	mapevent_t  *buf_mapevent  = NULL;
	mapobject_t *buf_mapobject = NULL;
	mapmove_t   *buf_mapmove   = NULL;
	chrdata_t   *buf_chrdata   = NULL;
	chrlvup_t   *buf_chrlvup   = NULL;
	chrjob_t    *buf_chrjob    = NULL;
	chrcmd_t    *buf_chrcmd    = NULL;
	chrabt_t    *buf_chrabt    = NULL;
	itemdata_t  *buf_itemdata  = NULL;
	magdata_t   *buf_magdata   = NULL;
	spdata_t    *buf_spdata    = NULL;
	landdata_t  *buf_landdata  = NULL;
	mixdata_t   *buf_mixdata   = NULL;
	treasure_t  *buf_trdata    = NULL;
	store_t     *buf_storedata = NULL;
	errno = 0;
	err_setprogname(argv[0]);

	/* ޥɥ饤Ϥ */
	for (optind = 1; optind < argc; optind++)
	{
		const char *p = argv[optind];

		if (p == NULL)
			continue;

		if (*p != '-')
			break;
		else if (*(p + 1) == '-')	/* "--" = ץλ */
		{
			optind++;
			break;
		}

		while (*++p != '\0')
		{
			switch (*p)
			{
			case 'n':
				flag_dry = 1;
				break;
			case 'h':	/* ˡɽ */
			default:
				usage();
				break;
			}
		}
	}
	argc -= optind;
	argv += optind;

	/* ϥѥꤹ */
	if (argc < 1)
		usage();

	if (strlen(argv[0]) > sizeof(str_srcpath))
		err_printf(STR_FILENAME_TOO_LONG);
	transfilename(str_srcpath, argv[0], sizeof(str_srcpath));

	/* ϥѥꤹ */
	if (argc >= 2)
	{
		if (strlen(argv[1]) > sizeof(str_dstpath))
			err_printf(STR_FILENAME_TOO_LONG);
		transpath(str_dstpath, argv[1], sizeof(str_dstpath));
	}

	/* ե򥪡ץ󤹤 */
	fprintf(stderr, STR_READ_FROM_FILE, str_srcpath);
	if (!file_open(str_srcpath, tbl_gametitle[FILEMODE_GBA], FILEMODE_GBA))
		err_printf(STR_INVALID_ROMIMAGE_FILE, str_srcpath);

	/* ƼǡϡϤ */
	buf_mondata   = mon_readdata();
	buf_monact    = mon_readact();
	buf_monparty  = mon_readparty();
	buf_monenc    = mon_readenc();
	buf_monmap    = mon_readmap();
	buf_mapdata   = map_readdata();
#if 0
	buf_mapevent  = map_readevent(buf_mapdata);
	buf_mapobject = map_readobject(buf_mapdata);
	buf_mapmove   = map_readmove(buf_mapdata);
#endif
	buf_chrdata   = chr_readdata();
	buf_chrlvup   = chr_readlvup();
	buf_chrjob    = chr_readjob();
	buf_chrcmd    = chr_readcmd();
	buf_chrabt    = chr_readabt();
	buf_itemdata  = item_readdata();
	buf_magdata   = mag_readdata();
	buf_spdata    = sp_readdata();
	buf_landdata  = sp_readland();
	buf_mixdata   = sp_readmix();
	buf_trdata    = tr_readdata(buf_mapdata);
	buf_storedata = store_readdata();

	if (buf_mondata   == NULL ||
		buf_monact    == NULL ||
		buf_monparty  == NULL ||
		buf_monenc    == NULL ||
		buf_monmap    == NULL ||
		buf_mapdata   == NULL ||
		buf_chrdata   == NULL ||
		buf_chrlvup   == NULL ||
		buf_chrjob    == NULL ||
		buf_chrcmd    == NULL ||
		buf_chrabt    == NULL ||
		buf_itemdata  == NULL ||
		buf_magdata   == NULL ||
		buf_spdata    == NULL ||
		buf_landdata  == NULL ||
		buf_mixdata   == NULL ||
		buf_trdata    == NULL ||
		buf_storedata == NULL)
	{
		err_printf(STR_INVALID_ROMIMAGE_FILE "\n", str_srcpath);
	}

	/* ƼǡϤ */
	do_writedata(FILENAME_MON_PARAM,		mon_writeparam(path, buf_mondata));
	do_writedata(FILENAME_MON_ELEM,			mon_writeelem(path, buf_mondata));
	do_writedata(FILENAME_MON_STATE,		mon_writestate(path, buf_mondata));
	do_writedata(FILENAME_MON_FLAG,			mon_writeflag(path, buf_mondata));
	do_writedata(FILENAME_MON_ITEM,			mon_writeitem(path, buf_mondata));
	do_writedata(FILENAME_MON_FREE,			mon_writefree(path, buf_mondata));
	do_writedata(FILENAME_MON_CTRL,			mon_writectrl(path, buf_mondata));
	do_writedata(FILENAME_MON_TECH,			mon_writetech(path, buf_mondata));
	do_writedata(FILENAME_MON_WEAK,			mon_writeweak(path, buf_mondata));
	do_writedata(FILENAME_MON_MISC,			mon_writemisc(path, buf_mondata));
	do_writedata(FILENAME_MON_ACT,			mon_writeact(path, buf_monact, buf_mondata));
	do_writedata(FILENAME_MON_MES,			mon_writemes(path));
	do_writedata(FILENAME_MON_PARTY,		mon_writeparty(path, buf_monparty, buf_mondata));
	do_writedata(FILENAME_MON_ENC,			mon_writeenc(path, buf_monenc, buf_monparty, buf_mondata));
	do_writedata(FILENAME_MON_MAP,			mon_writemap(path, buf_monmap));
	do_writedata(FILENAME_MON_LOCATION,		mon_writelocation(path, buf_mapdata, buf_monenc, buf_monparty, buf_mondata));
	do_writedata(FILENAME_CHR_PARAM,		chr_writeparam(path, buf_chrdata));
	do_writedata(FILENAME_CHR_EQUIP,		chr_writeequip(path, buf_chrdata));
	do_writedata(FILENAME_CHR_LVUP,			chr_writelvup(path, buf_chrlvup));
	do_writedata(FILENAME_CHR_JOB,			chr_writejob(path, buf_chrjob));
	do_writedata(FILENAME_CHR_CMD_PARAM,	chr_writecmdparam(path, buf_chrcmd));
	do_writedata(FILENAME_CHR_CMD_ELEM,		chr_writecmdelem(path, buf_chrcmd));
	do_writedata(FILENAME_CHR_CMD_CATEGORY,	chr_writecmdcategory(path, buf_chrcmd));
	do_writedata(FILENAME_CHR_CMD_STATE,	chr_writecmdstate(path, buf_chrcmd));
	do_writedata(FILENAME_CHR_CMD_FLAG,		chr_writecmdflag(path, buf_chrcmd));
	do_writedata(FILENAME_CHR_ABT,			chr_writeabtparam(path, buf_chrabt));
	do_writedata(FILENAME_ITEM_PARAM,		item_writeparam(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_ELEM,		item_writeelem(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_CATEGORY,	item_writecategory(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_STATE,		item_writestate(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_FLAG,		item_writeflag(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_EFFECT,		item_writeeffect(path, buf_itemdata));
	do_writedata(FILENAME_MAG_PARAM,		mag_writeparam(path, buf_magdata));
	do_writedata(FILENAME_MAG_ELEM,			mag_writeelem(path, buf_magdata));
	do_writedata(FILENAME_MAG_CATEGORY,		mag_writecategory(path, buf_magdata));
	do_writedata(FILENAME_MAG_STATE,		mag_writestate(path, buf_magdata));
	do_writedata(FILENAME_MAG_FLAG,			mag_writeflag(path, buf_magdata));
	do_writedata(FILENAME_SP_PARAM,			sp_writeparam(path, buf_spdata));
	do_writedata(FILENAME_SP_ELEM,			sp_writeelem(path, buf_spdata));
	do_writedata(FILENAME_SP_CATEGORY,		sp_writecategory(path, buf_spdata));
	do_writedata(FILENAME_SP_STATE,			sp_writestate(path, buf_spdata));
	do_writedata(FILENAME_SP_FLAG,			sp_writeflag(path, buf_spdata));
	do_writedata(FILENAME_SP_LANDSCAPE,		sp_writeland(path, buf_landdata));
	do_writedata(FILENAME_SP_MIX,			sp_writemix(path, buf_mixdata));
	do_writedata(FILENAME_TREASURE,			tr_writedata(path, buf_trdata, buf_monenc, buf_monparty, buf_mondata));
	do_writedata(FILENAME_STORE,			store_writedata(path, buf_storedata, buf_itemdata));
	do_writedata(FILENAME_ITEM_PARAM_BUG,	item_writeparambug(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_ELEM_BUG,	item_writeelembug(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_STATE_BUG,	item_writestatbug(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_FLAG_BUG,	item_writeflagbug(path, buf_itemdata));
	do_writedata(FILENAME_DIALOGUE,			dec_writedata(path, 246, 2317, FLAG_KANJI));
	do_writedata(FILENAME_MESSAGES,			dec_writedata(path, 0,   -1,   FLAG_KANA));
#if 0
	do_writedata(FILENAME_MAP_DATA,			map_writedata(path, buf_mapdata, buf_mapevent, buf_mapobject, buf_mapmove));
#endif

	/* ե򥯥 */
	file_close();

	/* Хåե */
	mon_freeact(buf_monact);
	free(buf_mondata);
	free(buf_monact);
	free(buf_monparty);
	free(buf_monenc);
	free(buf_monmap);
	free(buf_mapdata);
	free(buf_mapevent);
	free(buf_mapobject);
	free(buf_mapmove);
	free(buf_chrdata);
	free(buf_chrlvup);
	free(buf_itemdata);
	free(buf_magdata);
	free(buf_spdata);
	free(buf_landdata);
	free(buf_mixdata);
	free(buf_trdata);
	free(buf_storedata);
	buf_mondata   = NULL;
	buf_monact    = NULL;
	buf_monparty  = NULL;
	buf_monenc    = NULL;
	buf_monmap    = NULL;
	buf_mapdata   = NULL;
	buf_mapevent  = NULL;
	buf_mapobject = NULL;
	buf_mapmove   = NULL;
	buf_chrdata   = NULL;
	buf_chrlvup   = NULL;
	buf_chrjob    = NULL;
	buf_chrcmd    = NULL;
	buf_chrabt    = NULL;
	buf_itemdata  = NULL;
	buf_magdata   = NULL;
	buf_spdata    = NULL;
	buf_landdata  = NULL;
	buf_mixdata   = NULL;
	buf_trdata    = NULL;
	buf_storedata = NULL;

	return EXIT_SUCCESS;
}

void usage(void)
{
	fprintf(stderr, "%s\n\n", str_copyright);
	fprintf(stderr,	"usage: ff5analyzer: [-hn] src-pathname [dst-pathname]\n"
					"    -h                - display this help and exit\n"
					"    -n                - dry run\n"
					"    src-pathname      - ROM image file\n"
					"    dst-pathname      - write files into directory\n");

	exit(EXIT_SUCCESS);
}
