/*
 *	ff5character.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2007-2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff5decode.h"
#include "ff5character.h"
#include "ff5item.h"

#define FILEPOS_CHRDATA			0x141e70	/* SNES: 0x117000 */
#define FILEPOS_CHRPARAM		0x1564e0	/* SNES: 0x11551e */
#define FILEPOS_CHRLVUP			0x145f70	/* SNES: 0x115000, 0x115129, 0x1151ef */
#define FILEPOS_CHRJOBDATA		0x15734c	/* SNES: 0x1156b0 */
#define FILEPOS_CHRJOBLVUP_ADDR	0x1559cc	/* SNES: 0x1152c0 */
#define FILEPOS_CHRJOBLVUP		0x155a30	/* SNES: 0x115300 */
#define FILEPOS_CHRJOBLVMAX		0x14b744	/* SNES: 0x1152ea */
#define FILEPOS_CHRJOBEQUIP		0x1573b4	/* SNES: 0x115708 */
#define FILEPOS_CHRJOBCMD		0x1566b4	/* SNES: 0x115760 */
#define FILEPOS_CHRJOBABT		0x156680	/* SNES: 0x1157b8 */
#define FILEPOS_CHRCMDDATA		0x147e18	/* SNES: 0x1159e0 */
#define FILEPOS_CHRCMDTYPE		0x145e66	/* SNES: 0x10ef00 */
#define FILEPOS_CHRCMDPARAM		0x1564f4	/* SNES: 0x115e00 */
#define FILEPOS_CHRCMDTIME		0x1481f8	/* SNES: 0x115da0 */
#define FILEPOS_CHRABTDATA		0x156490	/* SNES: 0x11638c */
#define FILEPOS_CHRABTEQUIP		0x1563F4	/* SNES: 0x1163ce */
#define FILEPOS_CHRABTPARAM		0x156358	/* SNES: 0x116308 */
#define NUM_CHR					5
#define NUM_CHRMEMBER			4
#define NUM_CHRJOB				26			/* SNES: 22 */
#define NUM_CHRCMD				99			/* SNES: 96 */
#define NUM_CHRABT				39			/* SNES: 33 */
#define NUM_CHREQUIP			16
#define NUM_CHRSPABT			16
#define SIZEOF_CHRDATA			100			/* SNES: 80 */
#define SIZEOF_CHRPARAM			4
#define SIZEOF_CHRLVUP			8
#define SIZEOF_CHRJOBDATA		4
#define SIZEOF_CHRJOBLVUP		4			/* SNES: 3 */
#define SIZEOF_CHRJOBLVMAX		1
#define SIZEOF_CHRJOBEQUIP		4
#define SIZEOF_CHRJOBCMD		4
#define SIZEOF_CHRJOBABT		2
#define SIZEOF_CHRCMDDATA		8
#define SIZEOF_CHRCMDTYPE		1
#define SIZEOF_CHRCMDTIME		1
#define SIZEOF_CHRCMDPARAM		4
#define SIZEOF_OFFSETADDR		4
#define MSGID_CHRNAME			3088
#define MSGID_CHRJOBNAME		3756
#define MSGID_CHRCMDNAME		3095
#define MSGID_CHRABTNAME		2602
#define MSGID_CHRJOGMES			3730
#define MSGID_CHRCMDMES			3192
#define MSGID_CHRABTMES			2563

static void chr_setname(void);
static void chr_setjobname(void);
static void chr_setcmdname(void);
static void chr_setabtname(void);
static char tbl_chrname[NUM_CHR][256];		/* 饯̾Ǽơ֥ */
static char tbl_jobname[NUM_CHRJOB][256];	/* ֤̾Ǽơ֥ */
static char tbl_jobname[NUM_CHRJOB][256];	/* ֤̾Ǽơ֥ */
static char tbl_cmdname[NUM_CHRCMD][256];	/* ޥɥӥƥ̾Ǽơ֥ */
static char tbl_abtname[NUM_CHRABT][256];	/* ӥƥ̾Ǽơ֥ */
static int flag_setchrname = 0;				/* tbl_chrname˥ǡǼƤ뤫ɤ */
static int flag_setjobname = 0;				/* tbl_jobname˥ǡǼƤ뤫ɤ */
static int flag_setcmdname = 0;				/* tbl_cmdname˥ǡǼƤ뤫ɤ */
static int flag_setabtname = 0;				/* tbl_abtname˥ǡǼƤ뤫ɤ */

static const char *str_abt_effect[NUM_CHRSPABT] =
{
	"Ĥ",
	"Ȥʤ",
	"᡼椫",
	"å",
	"顼˥",
	"ޤۤХꥢ",
	"Ϥɤ",
	"󥿡",

	"ˤȤ夦",
	"󤻤",
	"",
	"С",
	"Τ",
	"礦Ƥ",
	"Ȥ",
	"Ф",
};

const char *chr_getname(int index)
{
	/* å */
	if (index < 0 || NUM_CHR <= index)
	{
		assert(0);
		return "";
	}

	if (!flag_setchrname)
		chr_setname();

	return tbl_chrname[index];
}

const char *chr_getjobname(int index)
{
	/* å */
	if (index < 0 || NUM_CHRJOB <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setjobname)
		chr_setjobname();

	return tbl_jobname[index];
}

const char *chr_getcmdname(int index)
{
	/* å */
	if (index < 0 || NUM_CHRCMD <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setcmdname)
		chr_setcmdname();

	return tbl_cmdname[index];
}

const char *chr_getabtname(int index)
{
	/* å */
	if (index < 0 || NUM_CHRABT <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setabtname)
		chr_setabtname();

	return tbl_abtname[index];
}

const char *chr_getcmdtypename(int index)
{
	static const char *str_cmdtype[] =
	{
		"00h",			"ƥ",		"",		"ܤ",
		"",		"ޤ",		"",			"",
		"㥯",		"Ȥ󤺤",		"̤",		"֤ɤ",
		"",		"夦",	"",	"֤󤷤",
		"ʤ",		"ߤͤ",		"ˤʤ",		"̤",
		"ɤ֤",		"ͤ餦",		"ߤ줦",	"Ӥ",
		"٤",		"ߤ֤",		"ʤ",		"Ĥ",
		"Ȥ館",		"Ϥʤ",		"礦",	"Τ",
		"",			"",		"",		"ߤ01",
		"",		"",	"ߤ02",		"",
		"",		"ɤ",		"Τޤ",		"ޤۤ",
		"",		"",		"ѥȥꥪ",		"ɥɥ",
		"Ҥ礦",		"ޤ",	"й",		"ĥ󥵡",
		"",		"ᤤ䤯",		"󤳤",		"褲",
		"ۤ",		"",		"󤮤",		"Ҥä",
		"3Ch",			"3Dh",			"3Eh",			"3Fh",
	};

	/* å */
	if (index < 0 || (int)(sizeof(str_cmdtype) / sizeof(str_cmdtype[0])) <= index)
	{
		assert(0);
		return "";
	}

	return str_cmdtype[index];
}

static void chr_setname(void)
{
	int i;

	/* 饯̾ǡ */
	for (i = 0; i < NUM_CHR; i++)
		*tbl_chrname[i] = '\0';

	/* 饯̾ǡɤ߹ */
	for (i = 0; i < NUM_CHR; i++)
		dec_msg(tbl_chrname[i], sizeof(tbl_chrname[i]), MSGID_CHRNAME + i, FLAG_KANA);

	flag_setchrname = 1;
}

static void chr_setjobname(void)
{
	int i;

	/* ֤̾ǡ */
	for (i = 0; i < NUM_CHRJOB; i++)
		*tbl_jobname[i] = '\0';

	/* ֤̾ǡɤ߹ */
	for (i = 0; i < NUM_CHRJOB; i++)
		dec_msg(tbl_jobname[i], sizeof(tbl_jobname[i]), MSGID_CHRJOBNAME + i, FLAG_KANA);

	flag_setjobname = 1;
}

static void chr_setcmdname(void)
{
	int i;

	/* ޥɥӥƥ̾ǡ */
	for (i = 0; i < NUM_CHRCMD; i++)
		*tbl_cmdname[i] = '\0';

	/* ޥɥӥƥ̾ǡɤ߹ */
	for (i = 0; i < 83; i++)
		dec_msg(tbl_cmdname[i], sizeof(tbl_cmdname[i]), MSGID_CHRCMDNAME + i, FLAG_KANA);

	for (; i < 87; i++)
		snprintf(tbl_cmdname[i], sizeof(tbl_cmdname[i]), "%s%02d", "ޥ", i);

	for (; i < NUM_CHRCMD; i++)
		dec_msg(tbl_cmdname[i], sizeof(tbl_cmdname[i]), MSGID_CHRCMDNAME + i - 4, FLAG_KANA);

	flag_setcmdname = 1;
}

static void chr_setabtname(void)
{
	int i;

	/* ӥƥ̾ǡ */
	for (i = 0; i < NUM_CHRABT; i++)
		*tbl_abtname[i] = '\0';

	/* ӥƥ̾ǡɤ߹ */
	for (i = 0; i < NUM_CHRABT; i++)
		dec_msg(tbl_abtname[i], sizeof(tbl_abtname[i]), MSGID_CHRABTNAME + i, FLAG_KANA);

	flag_setabtname = 1;
}

chrdata_t *chr_readdata(void)
{
	int i, j;
	chrdata_t *buf_chrdata = NULL;
	chrparam_t tbl_chrparam[NUM_CHR];

	/* 饯ΥǡǼХåեݤ */
	if ((buf_chrdata = malloc(sizeof(chrdata_t) * NUM_CHR)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 饯Υѥ᡼徺ͤɤ߹ */
	for (i = 0; i < NUM_CHR; i++)
	{
		unsigned char buf[SIZEOF_CHRPARAM];

		file_read(buf, FILEPOS_CHRPARAM + SIZEOF_CHRPARAM * i, SIZEOF_CHRPARAM);
		tbl_chrparam[i].strength    = buf[0];
		tbl_chrparam[i].speed       = buf[1];
		tbl_chrparam[i].stamina     = buf[2];
		tbl_chrparam[i].magic_power = buf[3];
	}

	/* 饯νǡɤ߹ */
	for (i = 0; i < NUM_CHRMEMBER; i++)
	{
		unsigned char buf[SIZEOF_CHRDATA];

		file_read(buf, FILEPOS_CHRDATA + SIZEOF_CHRDATA * i, SIZEOF_CHRDATA);

		/* ͷϥǡ */
		buf_chrdata[i].id                   = buf[ 0] & 0x07;
		buf_chrdata[i].job                  = buf[ 1];
		buf_chrdata[i].level                = buf[ 2];
		buf_chrdata[i].experience_point     = bits_make32(buf[4], buf[5], buf[6], buf[7]);
		buf_chrdata[i].hit_point            = (buf[11] << 8) | buf[10];
		buf_chrdata[i].magic_point          = (buf[15] << 8) | buf[14];
		buf_chrdata[i].caught_monster       = buf[31];
		buf_chrdata[i].equip_weight         = buf[45];
		buf_chrdata[i].strength             = buf[46];
		buf_chrdata[i].speed                = buf[47];
		buf_chrdata[i].stamina              = buf[48];
		buf_chrdata[i].magic_power          = buf[49];
		buf_chrdata[i].evade_rate           = buf[54];
		buf_chrdata[i].defense_power        = buf[55];
		buf_chrdata[i].magic_evade_rate     = buf[56];
		buf_chrdata[i].magic_defense_power  = buf[57];

		/* ե饰ϥǡ */
		buf_chrdata[i].flag_backrow = ((buf[0] & 0x80) != 0x00);
		buf_chrdata[i].flag_exist   = ((buf[0] & 0x40) == 0x00);
		buf_chrdata[i].flag_female  = ((buf[0] & 0x08) != 0x00);

		/* ޥ */
		for (j = 0; j < 4; j++)
			buf_chrdata[i].cmd_no[j] = buf[j + 32];

		/* ƥǡ */
		for (j = 0; j < 7; j++)
		{
			const int item_no = (buf[16 + j * 2 + 1] << 8) | buf[16 + j * 2];

			if (j < 5)
				/* Ƭ, , , ɶ, ɶ */
				buf_chrdata[i].equip_no[j] = item_no;
			else if (buf_chrdata[i].equip_no[j - 2] == 0x00)
				/* ,  */
				buf_chrdata[i].equip_no[j - 2] = item_no;
		}
	}
	for (; i < NUM_CHR; i++)
	{
		/* ͷϥǡ */
		buf_chrdata[i].id                   = i;

		/* ΰ򥯥ꥢ */
		buf_chrdata[i].job                  = 0;
		buf_chrdata[i].level                = 0;
		buf_chrdata[i].experience_point     = 0;
		buf_chrdata[i].hit_point            = 0;
		buf_chrdata[i].magic_point          = 0;
		buf_chrdata[i].caught_monster       = 0;
		buf_chrdata[i].equip_weight         = 0;
		buf_chrdata[i].strength             = 0;
		buf_chrdata[i].speed                = 0;
		buf_chrdata[i].stamina              = 0;
		buf_chrdata[i].magic_power          = 0;
		buf_chrdata[i].evade_rate           = 0;
		buf_chrdata[i].defense_power        = 0;
		buf_chrdata[i].magic_evade_rate     = 0;
		buf_chrdata[i].magic_defense_power  = 0;

		/* ե饰ϥǡ */
		buf_chrdata[i].flag_backrow = 0;
		buf_chrdata[i].flag_exist   = 0;
		buf_chrdata[i].flag_female  = 0;

		/* ޥ */
		for (j = 0; j < 4; j++)
			buf_chrdata[i].cmd_no[j] = 0;

		/* ƥǡ */
		for (j = 0; j < 5; j++)
			buf_chrdata[i].equip_no[j] = 0;
	}

	/* 饯Υѥ᡼徺ͤɤ߹ */
	for (i = 0; i < NUM_CHR; i++)
	{
		const int id = buf_chrdata[i].id;

		buf_chrdata[i].increase = tbl_chrparam[id];
	}

	return buf_chrdata;
}

chrlvup_t *chr_readlvup(void)
{
	int i;
	chrlvup_t *buf_chrlvup = NULL;

	/* 饯ĹǡǼХåեݤ */
	if ((buf_chrlvup = malloc(sizeof(chrlvup_t) * NUM_CHRLV)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ٥뤴Ȥ߷׷и͡ܥҥåȥݥȡܥޥåݥȤɤ߹ */
	for (i = 0; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRLVUP];

		file_read(buf, FILEPOS_CHRLVUP + SIZEOF_CHRLVUP * i, SIZEOF_CHRLVUP);
		buf_chrlvup[i].experience_point = bits_make32(buf[0], buf[1], buf[2], buf[3]);
		buf_chrlvup[i].hit_point        = (buf[5] << 8) | buf[4];
		buf_chrlvup[i].magic_point      = (buf[7] << 8) | buf[6];
	}

	return buf_chrlvup;
}

chrjob_t *chr_readjob(void)
{
	int i, j;
	long tbl_offset[NUM_CHRJOB];
	chrjob_t *buf_chrjob = NULL;
	/* const size_t len_joblvup_offset = 0x5a30; */	/* SNES: 0x5300 */

	/* ֤γƼǡǼХåեݤ */
	if ((buf_chrjob = malloc(sizeof(chrjob_t) * NUM_CHRJOB)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ֤δܥѥ᡼ɤ߹ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBDATA];

		file_read(buf, FILEPOS_CHRJOBDATA + SIZEOF_CHRJOBDATA * i, SIZEOF_CHRJOBDATA);
		buf_chrjob[i].chrparam.strength    = buf[0];
		buf_chrjob[i].chrparam.speed       = buf[1];
		buf_chrjob[i].chrparam.stamina     = buf[2];
		buf_chrjob[i].chrparam.magic_power = buf[3];
	}

	/* ֤Ǥ륢ƥΥǡɤ߹ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBEQUIP];

		file_read(buf, FILEPOS_CHRJOBEQUIP + SIZEOF_CHRJOBEQUIP * i, SIZEOF_CHRJOBEQUIP);
		buf_chrjob[i].bits_equip_weapon    = (buf[1] << 8) | buf[0];
		buf_chrjob[i].bits_equip_protector = (buf[3] << 8) | buf[2];
	}

	/* ֤δܥޥɤɤ߹ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBCMD];

		file_read(buf, FILEPOS_CHRJOBCMD + SIZEOF_CHRJOBCMD * i, SIZEOF_CHRJOBCMD);
		for (j = 0; j < 4; j++)
			buf_chrjob[i].cmd_no[j] = buf[j];
	}

	/* ֤Ȥӥƥθ̤ɤ߹ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBABT];

		file_read(buf, FILEPOS_CHRJOBABT + SIZEOF_CHRJOBABT * i, SIZEOF_CHRJOBABT);
		buf_chrjob[i].bits_abt_effect = (buf[1] << 8) | buf[0];
	}

	/* ֤Ȥκǹ⥸֥٥ɤ߹ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBLVMAX];

		file_read(buf, FILEPOS_CHRJOBLVMAX + SIZEOF_CHRJOBLVMAX * i, SIZEOF_CHRJOBLVMAX);
		buf_chrjob[i].level_max = buf[0];
	}

	/* ֥٥ΥǡǼƤ륢ɥ쥹ɤ߹ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_CHRJOBLVUP_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = bits_make32(buf[0], buf[1], buf[2], 0x00);
	}

	/* ֥٥Υǡɤ߹ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		for (j = 0; j < buf_chrjob[i].level_max; j++)
		{
			unsigned char buf[SIZEOF_CHRJOBLVUP];

			file_read(buf, tbl_offset[i] + SIZEOF_CHRJOBLVUP * j, SIZEOF_CHRJOBLVUP);
			buf_chrjob[i].level[j].ability_point = (buf[1] << 8) | buf[0];
			buf_chrjob[i].level[j].ability_no    = buf[2];
		}
		for (; j < NUM_CHRJOBLVUPMAX; j++)
		{
			buf_chrjob[i].level[j].ability_point = 0;
			buf_chrjob[i].level[j].ability_no    = 0;
		}
	}

	return buf_chrjob;
}

chrcmd_t *chr_readcmd(void)
{
	int i;
	chrcmd_t *buf_chrcmd = NULL;

	/* ޥɥӥƥγƼǡǼХåեݤ */
	if ((buf_chrcmd = malloc(sizeof(chrcmd_t) * NUM_CHRCMD)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ޥɥӥƥγƼǡɤ߹ */
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		unsigned char buf[SIZEOF_CHRCMDDATA];

		file_read(buf, FILEPOS_CHRCMDDATA + SIZEOF_CHRCMDDATA * i, SIZEOF_CHRCMDDATA);
		mag_setparam(&buf_chrcmd[i].param, buf, sizeof(buf));

		buf_chrcmd[i].flag_human_attack      = ((buf[7] & 0x01) != 0x00);
		buf_chrcmd[i].flag_swdslap_attack    = ((buf[7] & 0x02) != 0x00);
		buf_chrcmd[i].flag_ignore_defense    = ((buf[7] & 0x04) != 0x00);
		buf_chrcmd[i].flag_multiplier_half   = ((buf[7] & 0x08) != 0x00);
		buf_chrcmd[i].flag_multiplier_double = ((buf[7] & 0x10) != 0x00);
		buf_chrcmd[i].flag_attack_half       = ((buf[7] & 0x20) != 0x00);
		buf_chrcmd[i].flag_attack_double     = ((buf[7] & 0x40) != 0x00);
		buf_chrcmd[i].flag_always_hit        = ((buf[7] & 0x80) != 0x00);
	}

	/* ޥɥӥƥθ̥פɤ߹ */
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		unsigned char buf[SIZEOF_CHRCMDTYPE];

		file_read(buf, FILEPOS_CHRCMDTYPE + SIZEOF_CHRCMDTYPE * i, SIZEOF_CHRCMDTYPE);
		buf_chrcmd[i].effect_type = buf[0];
	}

	/* ޥɥӥƥθȯưޤǤλ֤ɤ߹ */
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		unsigned char buf[SIZEOF_CHRCMDTIME];

		file_read(buf, FILEPOS_CHRCMDTIME + SIZEOF_CHRCMDTIME * i, SIZEOF_CHRCMDTIME);
		buf_chrcmd[i].time_wait = buf[0] & 0x7f;
	}

	/* ޥɥӥƥΥѥ᡼徺ͤɤ߹ */
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		unsigned char buf[SIZEOF_CHRCMDPARAM];

		file_read(buf, FILEPOS_CHRCMDPARAM + SIZEOF_CHRCMDPARAM * i, SIZEOF_CHRCMDPARAM);
		buf_chrcmd[i].chrparam.strength    = buf[0];
		buf_chrcmd[i].chrparam.speed       = buf[1];
		buf_chrcmd[i].chrparam.stamina     = buf[2];
		buf_chrcmd[i].chrparam.magic_power = buf[3];
	}

	return buf_chrcmd;
}

chrabt_t *chr_readabt(void)
{
	int i;
	chrabt_t *buf_chrabt = NULL;

	/* ӥƥγƼǡǼХåեݤ */
	if ((buf_chrabt = malloc(sizeof(chrabt_t) * NUM_CHRABT)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ӥƥΥѥ᡼徺ͤɤ߹ */
	for (i = 0; i < NUM_CHRABT; i++)
	{
		unsigned char buf[SIZEOF_CHRCMDPARAM];

		file_read(buf, FILEPOS_CHRABTPARAM + SIZEOF_CHRCMDPARAM * i, SIZEOF_CHRCMDPARAM);
		buf_chrabt[i].chrparam.strength    = buf[0];
		buf_chrabt[i].chrparam.speed       = buf[1];
		buf_chrabt[i].chrparam.stamina     = buf[2];
		buf_chrabt[i].chrparam.magic_power = buf[3];
	}

	/* ӥƥθ̤ɤ߹ */
	for (i = 0; i < NUM_CHRABT; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBABT];

		file_read(buf, FILEPOS_CHRABTDATA + SIZEOF_CHRJOBABT * i, SIZEOF_CHRJOBABT);
		buf_chrabt[i].bits_abt_effect = (buf[1] << 8) | buf[0];
	}

	/* ӥƥǤ륢ƥΥǡɤ߹ */
	for (i = 0; i < NUM_CHRABT; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBEQUIP];

		file_read(buf, FILEPOS_CHRABTEQUIP + SIZEOF_CHRJOBEQUIP * i, SIZEOF_CHRJOBEQUIP);
		buf_chrabt[i].bits_equip_weapon    = (buf[1] << 8) | buf[0];
		buf_chrabt[i].bits_equip_protector = (buf[3] << 8) | buf[2];
	}

	return buf_chrabt;
}

int chr_writeparam(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i;
	FILE *fp = NULL;
	/* const char *str_sex[2] = {"", ""}; */
	const char *str_nothing = "---------------------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯ΥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯ΥǡϤ */
	fprintf(fp, "No.      ̾           "
				" LV HP MP   EXP   "
				"   ®  \n");
	for (i = 0; i < NUM_CHR; i++)
	{
		fprintf(fp, "%03d(%02X): %-12s : ", i, i, chr_getname(ptr_chrdata[i].id));

		if (i < NUM_CHRMEMBER)
			fprintf(fp, "%3d, %5d, %5d, %8ld",
					ptr_chrdata[i].level,
					ptr_chrdata[i].hit_point,
					ptr_chrdata[i].magic_point,
					ptr_chrdata[i].experience_point);
		else
			fprintf(fp, "%27s", str_nothing);

		fprintf(fp, ", %+3d, %+3d, %+3d, %+3d\n",
				ptr_chrdata[i].increase.strength,
				ptr_chrdata[i].increase.speed,
				ptr_chrdata[i].increase.stamina,
				ptr_chrdata[i].increase.magic_power);
	}

	return 1;
}

int chr_writeequip(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯ƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯ƥǡϤ */
	fprintf(fp, "No.      ̾           "
				"                "
				"                "
				"Ƭ                  "
				"                  "
				"\n");
	for (i = 0; i < NUM_CHRMEMBER; i++)
	{
		const char *ptr_name[5];

		for (j = 0; j < 5; j++)
		{
			ptr_name[j] = (ptr_chrdata[i].equip_no[j] != 0x00 && ptr_chrdata[i].equip_no[j] != 0x80)
							? item_getname(ptr_chrdata[i].equip_no[j])
							: STR_NOTHING_18;
		}

		fprintf(fp, "%03d(%02X): %-12s : %-18s, %-18s, %-18s, %-18s, %-18s\n", i, i,
					chr_getname(i),
					ptr_name[3],
					ptr_name[4],
					ptr_name[0],
					ptr_name[1],
					ptr_name[2]);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writelvup(const char *ptr_filename, const chrlvup_t *ptr_chrlvup)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrlvup == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯ĹǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯ĹǡϤ */
	fprintf(fp, "LV       EXP    NEXT HP MP\n");
	for (i = 0; i < NUM_CHRLV; i++)
	{
		fprintf(fp, "%2d: %8ld, %6ld, %5d, %5d\n", i + 1,
				ptr_chrlvup[i].experience_point,
				(i + 1 < NUM_CHRLV) ? ptr_chrlvup[i + 1].experience_point - ptr_chrlvup[i].experience_point : 0,
				ptr_chrlvup[i].hit_point,
				ptr_chrlvup[i].magic_point);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writejob(const char *ptr_filename, const chrjob_t *ptr_chrjob)
{
	int i, j, k;
	FILE *fp = NULL;
	const int tbl_equip_weapon[NUM_CHREQUIP] =
	{
		2,	/* = "ʥ" */
		3,	/* = "Ǧ" */
		4,	/* = "" */
		5,	/* = "η" */
		6,	/* = "" */
		7,	/* = "" */
		8,	/* = "" */
		9,	/* = "" */

		10,	/* = "å" */
		11,	/* = "" */
		12,	/* = "̤1" */
		13,	/* = "" */
		14,	/* = "è" */
		15,	/* = "" */
		16,	/* = "٥" */
		0,
	};
	const int tbl_equip_protector[NUM_CHREQUIP] =
	{
		17,	/* = "" */
		18,	/* = "" */
		19,	/* = "Ʈ˹" */
		21,	/* = "ɤꤳ" */
		22,	/* = "" */
		23,	/* = "Ʈ" */
		24,	/* = "" */
		1,	/* = "ɶ/" */

		25, /* = "" */
		26, /* = "" */
		28,	/* = "" */
		29,	/* = "¾" */
		20,	/* = "ˡ˹" */
		31,	/* = "̤2" */
		0,
		0,
	};
	const char *str_equip_all  = "(٤)";
	const char *str_equip_none = "(ʤ)";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrjob == NULL)
	{
		assert(0);
		return 0;
	}

	/* ֤γƼǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ֤γƼǡϤ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X): %s]\n", i, i, chr_getjobname(i));

		/* ơ */
		fprintf(fp, "     ®  \n"
					" %3d, %3d, %3d, %3d\n",
				ptr_chrjob[i].chrparam.strength,
				ptr_chrjob[i].chrparam.speed,
				ptr_chrjob[i].chrparam.stamina,
				ptr_chrjob[i].chrparam.magic_power);

		/* ǽʥƥ */
		fprintf(fp, "Ǥ륢ƥ\n");

		/*  */
		fprintf(fp, "  : ");
		if (ptr_chrjob[i].bits_equip_weapon == 0xffff)
			fprintf(fp, "%s", str_equip_all);
		else if (ptr_chrjob[i].bits_equip_weapon == 0x00)
			fprintf(fp, "%s", str_equip_none);
		else
		{
			k = 0;
			for (j = 0; j < NUM_CHREQUIP; j++)
			{
				if ((ptr_chrjob[i].bits_equip_weapon & (1U << j)) != 0x00)
				{
					const int equip_type = tbl_equip_weapon[j];

					if (k++ != 0)
						fprintf(fp, ", ");
					fprintf(fp, "%s", item_getequiptype(equip_type));
				}
			}

			/*  (ɶΥե饰) */
			if ((ptr_chrjob[i].bits_equip_protector & (1U << 10)) != 0x00)
			{
				if (k++ != 0)
					fprintf(fp, ", ");
				fprintf(fp, "%s", item_getequiptype(30));
			}
		}
		fputc('\n', fp);

		/* ɶ */
		fprintf(fp, "  : ");
		if (ptr_chrjob[i].bits_equip_protector == 0xffff)
			fprintf(fp, "%s", str_equip_all);
		else if (ptr_chrjob[i].bits_equip_protector == 0x00)
			fprintf(fp, "%s", str_equip_none);
		else
		{
			k = 0;
			for (j = 0; j < NUM_CHREQUIP; j++)
			{
				if ((ptr_chrjob[i].bits_equip_protector & (1U << j)) != 0x00)
				{
					const int equip_type = tbl_equip_protector[j];

					if (k++ != 0)
						fprintf(fp, ", ");

					/* 1: / */
					if (equip_type == 1)
						fprintf(fp, "%s/%s", item_getequiptype(1), item_getequiptype(27));
					else
						fprintf(fp, "%s", item_getequiptype(equip_type));
				}
			}
		}
		fputc('\n', fp);

		/* ӥƥ */
		if (ptr_chrjob[i].bits_abt_effect != 0x00)
		{
			fprintf(fp, "No. \n");

			k = 0;
			for (j = 0; j < NUM_CHRSPABT; j++)
			{
				if ((ptr_chrjob[i].bits_abt_effect & (1U << j)) != 0x00)
				{
					fprintf(fp, "%2d: %s\n", k, str_abt_effect[j]);
					k++;
				}
			}
		}

		/* ޥ */
		fprintf(fp, "No. ޥ\n");
		for (j = 0; j < 4; j++)
		{
			fprintf(fp, "%2d: %s\n", j,
					(ptr_chrjob[i].cmd_no[j] != 0x00) ? chr_getcmdname(ptr_chrjob[i].cmd_no[j]) : STR_NOTHING_10);
		}

		/* ֥٥ */
		if (ptr_chrjob[i].level_max > 0)
		{
			fprintf(fp, "LV   ABP  ӥƥ\n");
			for (j = 0; j < ptr_chrjob[i].level_max; j++)
			{
				int no = ptr_chrjob[i].level[j].ability_no & 0x7f;

				fprintf(fp, "%2d:  %3d, %s\n", j + 1,
						ptr_chrjob[i].level[j].ability_point,
						((ptr_chrjob[i].level[j].ability_no & 0x80) == 0x00) ? chr_getcmdname(no) : chr_getabtname(no));
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmdparam(const char *ptr_filename, const chrcmd_t *ptr_chrcmd)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrcmd == NULL)
	{
		assert(0);
		return 0;
	}

	/* ޥɥӥƥοͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ޥɥӥƥοͷϥǡϤ */
	fprintf(fp, "No.      ̾         "
				"   Ψ ̿Ψ   ³  "
				"Ի "
				"     ®        \n");
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		fprintf(fp, (ptr_chrcmd[i].param.effect_rate != 0) ?
					"%03d(%02X): %-10s :  %5d,%3d/16, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d, %4Xh, %4Xh\n" :
					"%03d(%02X): %-10s :  %5d, %5d, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d, %4Xh, %4Xh\n",
				i, i, chr_getcmdname(i),
				ptr_chrcmd[i].param.effect_power,
				ptr_chrcmd[i].param.effect_rate,
				ptr_chrcmd[i].param.hit_rate,
				ptr_chrcmd[i].param.effect_time,
				ptr_chrcmd[i].param.fixed_value,
				ptr_chrcmd[i].time_wait,
				ptr_chrcmd[i].chrparam.strength,
				ptr_chrcmd[i].chrparam.speed,
				ptr_chrcmd[i].chrparam.stamina,
				ptr_chrcmd[i].chrparam.magic_power,
				ptr_chrcmd[i].effect_type,
				ptr_chrcmd[i].param.effect_type);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmdelem(const char *ptr_filename, const chrcmd_t *ptr_chrcmd)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrcmd == NULL)
	{
		assert(0);
		return 0;
	}

	/* ޥɥӥƥ°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ޥɥӥƥ°ϥǡϤ */
	fprintf(fp, "No.      ̾         "
				"°\n");
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		fprintf(fp, "%03d(%02X): %-10s : ", i, i, chr_getcmdname(i));

		/* ° */
		str_printelem(fp, ptr_chrcmd[i].param.elem_effect, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmdcategory(const char *ptr_filename, const chrcmd_t *ptr_chrcmd)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrcmd == NULL)
	{
		assert(0);
		return 0;
	}

	/* ޥɥӥƥΥƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ޥɥӥƥΥƥǡϤ */
	fprintf(fp, "No.      ̾         "
				"ƥ\n");
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		fprintf(fp, "%03d(%02X): %-10s : ", i, i, chr_getcmdname(i));

		/* ƥ */
		str_printcategory(fp, ptr_chrcmd[i].param.category, CATEGORY_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmdstate(const char *ptr_filename, const chrcmd_t *ptr_chrcmd)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrcmd == NULL)
	{
		assert(0);
		return 0;
	}

	/* ޥɥӥƥξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ޥɥӥƥξѲϥǡϤ */
	fprintf(fp, "No.      ̾         "
				"Ѳ1        2                3                4\n");
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		fprintf(fp, "%03d(%02X): %-10s : ", i, i, chr_getcmdname(i));

		/* Ѳ */
		str_printstate(fp, ptr_chrcmd[i].param.state_effect, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmdflag(const char *ptr_filename, const chrcmd_t *ptr_chrcmd)
{
	int i;
	FILE *fp = NULL;
	const char *str_double[3] =
	{
		"Ⱦ",
		" -- ",
		" 2",
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrcmd == NULL)
	{
		assert(0);
		return 0;
	}

	/* ޥɥӥƥΥե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ޥɥӥƥΥե饰ϥǡϤ */
	fprintf(fp, "No.      ̾         "
				"̵ ɬ "
				"Ψ  "
				"ϰ           "
				"ͭ²\n");
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		const int modifier_multiplier = 1
			- (ptr_chrcmd[i].flag_multiplier_half   ? 1 : 0)
			+ (ptr_chrcmd[i].flag_multiplier_double ? 1 : 0);
		const int modifier_attack_power = 1
			- (ptr_chrcmd[i].flag_attack_half   ? 1 : 0)
			+ (ptr_chrcmd[i].flag_attack_double ? 1 : 0);

		fprintf(fp, "%03d(%02X): %-10s : ", i, i, chr_getcmdname(i));

		/* ե饰 */
		fprintf(fp, " %2s   %2s  ",
			str_getflag(ptr_chrcmd[i].flag_ignore_defense),
			str_getflag(ptr_chrcmd[i].flag_always_hit));

		/* ΨϤν */
		fprintf(fp, "%-4s %-4s   ",
			str_double[modifier_multiplier],
			str_double[modifier_attack_power]);

		/* ϰ */
		if (ptr_chrcmd[i].param.auto_direction && ptr_chrcmd[i].param.direction == 1)
			/* ȯưưŪ˷ꤹ */
			fprintf(fp, "-----------------");
		else
			mag_printtarget(fp, &ptr_chrcmd[i].param);

		/* ͭ󥹥 */
		if (ptr_chrcmd[i].param.montype_effect != 0x00)
		{
			fprintf(fp, "  ");
			str_printmontype(fp, ptr_chrcmd[i].param.montype_effect, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writeabtparam(const char *ptr_filename, const chrabt_t *ptr_chrabt)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrabt == NULL)
	{
		assert(0);
		return 0;
	}

	/* ӥƥοͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ӥƥοͷϥǡϤ */
	fprintf(fp, "No.      ̾               "
				"     ®  \n");
	for (i = 0; i < NUM_CHRABT; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %3d, %3d, %3d, %3d\n", i, i,
				chr_getabtname(i),
				ptr_chrabt[i].chrparam.strength,
				ptr_chrabt[i].chrparam.speed,
				ptr_chrabt[i].chrparam.stamina,
				ptr_chrabt[i].chrparam.magic_power);
#if 0
		fprintf(fp, "%03d(%02X): %-16s : %4X,%4X,%4X\n", i, i,
				chr_getabtname(i),
				ptr_chrabt[i].bits_equip_weapon,
				ptr_chrabt[i].bits_equip_protector,
				ptr_chrabt[i].bits_abt_effect);
#endif
	}

	return 1;
}
