/*
 *	ff5define.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2007-2009 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "ff5define.h"

#define STR_FLAG_TRUE	""
#define STR_FLAG_FALSE	""

static const char *str_element[NUM_ELEMENT] =
{
	"",
	"ɹ",
	"",
	"",
	"",
	"",
	"",
	"",
};

static const char *str_category[NUM_CATEGORY] =
{
	"Ũ",	/* ƤΤ虜 */
	"",	/* ˡ */
	"",	/* ˡ */
	"",	/* ˡ */
	"",	/* ˡ */
	"",	/* ȹ */
	"",	/* ֹ */
	"",	/* ܹ */
};

static const char *str_state[NUM_STATE] =
{
	"",	/* Ű */
	"",	/* ӡ */
	"",	/*  */
	"",	/* ӥƥ */
	"",	/* ߥ˥ޥ */
	"",	/*  */
	"",	/* в */
	"",	/* Ʈǽ */

	"ʬ",	/* ʬ1 */
	"ʬ",	/* ʬ2 */
	"",	/*  */
	"",	/* С */
	"",	/*  */
	"",	/*  */
	"̲",	/* ̲ */
	"Ϸ",	/* Ϸ */

	"",	/* ꥸ */
	"̵",	/* ̵Ũ */
	"",	/*  */
	"®",	/* إ */
	"",	/* ȥå */
	"",	/*  */
	"ʪ",	/* ץƥ */
	"ȿ",	/* ե쥯 */

	"",	/* Ƥ */
	"",	/* λ */
	"",	/* ΤäƤ */
	"",	/* å */
	"",	/*  */
	"",	/* Ƥ */
	"",	/* ʪ */
	"æ",	/* Υæ */
};

static const char *str_state_flag[NUM_STATE_FLAG] =
{
	STR_NOTHING,	/* ̤1 */
	STR_NOTHING,	/* ̤2 */
	STR_NOTHING,	/* ̤3 */
	"",			/*  */
	"",			/*  */
	STR_NOTHING,	/* ̤4 */
	"",			/* ޤ */
	"",			/* ܤ */
};

static const char *str_montype[NUM_MONTYPE] =
{
	"ǥå",
	"",			/* ̤ */
	"",
	"ӥ",
	"ɥ饴",
	"ܥ",
	"",
	"ʹ",
};

const char *str_getflag(int flag)
{
	return (flag != 0x00) ? STR_FLAG_TRUE : STR_FLAG_FALSE;
}

const char *str_getelem(elem_t elem, int index)
{
	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* å */
	if (index < 0 || NUM_ELEMENT <= index)
	{
		assert(0);
		return "";
	}

	return ((elem & (1U << index)) != 0x00) ? str_element[index] : "";
}

char *str_setelem(char *str, elem_t elem, size_t bufsize)
{
	int i;

	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	for (i = 0; i < NUM_ELEMENT; i++)
	{
		if ((elem & (1U << i)) != 0x00)
			strncat(str, str_element[i], bufsize - strlen(str) - 1);
	}

	return str;
}

void str_printelem(FILE *stream, elem_t elem, elem_t elem_mask)
{
	int i;

	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* å */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_ELEMENT; i++)
	{
		if ((elem_mask & (1U << i)) != 0x00)
			fprintf(stream, "%-2s", ((elem & (1U << i)) != 0x00) ? str_element[i] : STR_NOTHING);
	}
}

const char *str_getcategory(category_t category, int index)
{
	assert(sizeof(str_category) / sizeof(str_category[0]) == NUM_CATEGORY);

	/* å */
	if (index < 0 || NUM_CATEGORY <= index)
	{
		assert(0);
		return "";
	}

	return ((category & (1U << index)) != 0x00) ? str_category[index] : "";
}

char *str_setcategory(char *str, category_t category, size_t bufsize)
{
	int i;

	assert(sizeof(str_category) / sizeof(str_category[0]) == NUM_CATEGORY);

	for (i = 0; i < NUM_CATEGORY; i++)
	{
		if ((category & (1U << i)) != 0x00)
			strncat(str, str_category[i], bufsize - strlen(str) - 1);
	}

	return str;
}


void str_printcategory(FILE *stream, category_t category, category_t category_mask)
{
	int i;

	assert(sizeof(str_category) / sizeof(str_category[0]) == NUM_CATEGORY);

	/* å */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_CATEGORY; i++)
	{
		if ((category_mask & (1U << i)) != 0x00)
			fprintf(stream, "%-2s", ((category & (1U << i)) != 0x00) ? str_category[i] : STR_NOTHING);
	}
}

const char *str_getstate(state_t state, int index)
{
	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* å */
	if (index < 0 || NUM_STATE <= index)
	{
		assert(0);
		return "";
	}

	return ((state & (1U << index)) != 0x00) ? str_state[index] : "";
}

char *str_setstate(char *str, state_t state, size_t bufsize)
{
	int i;

	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	for (i = 0; i < NUM_STATE; i++)
	{
		if ((state & (1U << i)) != 0x00)
			strncat(str, str_state[i], bufsize - strlen(str) - 1);
	}

	return str;
}

char *str_setstate_f(char *str, state_t state, size_t bufsize)
{
	int i;

	assert(sizeof(str_state_flag) / sizeof(str_state_flag[0]) == NUM_STATE_FLAG);

	for (i = 0; i < NUM_STATE_FLAG; i++)
	{
		if ((state & (1U << i)) != 0x00)
			strncat(str, str_state_flag[i], bufsize - strlen(str) - 1);
	}

	return str;
}

void str_printstate(FILE *stream, state_t state, state_t state_mask)
{
	int i, count = 0;

	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* å */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_STATE; i++)
	{
		if ((i % 8) == 0)
		{
			if (count > 0 && state_mask > (1U << i))
				fprintf(stream, " ");
			count = 0;
		}

		if ((state_mask & (1U << i)) != 0x00)
		{
			fprintf(stream, "%-2s", ((state & (1U << i)) != 0x00) ? str_state[i] : STR_NOTHING);
			count++;
		}
	}
}

const char *str_getmontype(montype_t montype, int index)
{
	assert(sizeof(str_montype) / sizeof(str_montype[0]) == NUM_MONTYPE);

	/* å */
	if (index < 0 || NUM_MONTYPE <= index)
	{
		assert(0);
		return "";
	}

	return ((montype & (1U << index)) != 0x00) ? str_montype[index] : "";
}

void str_printmontype(FILE *stream, montype_t montype, montype_t montype_mask)
{
	int i, count = 0;

	assert(sizeof(str_montype) / sizeof(str_montype[0]) == NUM_MONTYPE);

	/* å */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_MONTYPE; i++)
	{
		if ((montype_mask & (1U << i)) != 0x00)
		{
			if ((montype & (1U << i)) != 0x00)
			{
				if (count > 0)
					fprintf(stream, " ");
				count++;
			}

			fprintf(stream, "%s", str_getmontype(montype, i));
		}
	}
}
