/*
 *	ff5item.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2007-2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff5decode.h"
#include "ff5item.h"

#define FILEPOS_ITEMDATA_WEAPON		0x155d58	/* SNES: 0x110000 */
#define FILEPOS_ITEMDATA_PROTECTOR	0x146e18	/* SNES: 0x110600 */
#define FILEPOS_ITEMDATA_TOOL		0x149c04	/* SNES: 0x110a80 */
#define FILEPOS_ITEMELEM			0x148380	/* SNES: 0x112580 */
#define FILEPOS_ITEMSTATE			0x1483e0	/* SNES: 0x1126c0 */
#define FILEPOS_ITEMPRICE			0x15671c	/* SNES: 0x112a00 */
#define FILEPOS_ITEMDATA_MAGLAMP	0x145edd	/* SNES: 0x10eee7 */
#define NUM_ITEM_WEAPON				128
#define NUM_ITEM_PROTECTOR			96
#define NUM_ITEM_TOOL				36			/* SNES: 32 */
#define NUM_ITEM_EQUIP				(NUM_ITEM_WEAPON + NUM_ITEM_PROTECTOR)					/* = 224 */
#define NUM_ITEM					(NUM_ITEM_WEAPON + NUM_ITEM_PROTECTOR + NUM_ITEM_TOOL)	/* = 260 */
#define NUM_IMPITEM					25
#define NUM_ITEMELEM				12			/* SNES: 64 */
#define NUM_ITEMSTATE				30			/* SNES: 64 */
#define NUM_EQUIPTYPE				32
#define NUM_ITEM_MAGLAMP			15
#define SIZEOF_ITEMDATA_WEAPON		12
#define SIZEOF_ITEMDATA_PROTECTOR	12
#define SIZEOF_ITEMDATA_TOOL		8
#define SIZEOF_ITEMELEM				8			/* SNES: 5 */
#define SIZEOF_ITEMSTATE			8			/* SNES: 7 */
#define SIZEOF_ITEMPRICE			12			/* SNES: 2 */
#define SIZEOF_ITEMDATA_MAGLAMP		1
#define REGION_WEAPON_BEGIN			0
#define REGION_WEAPON_END			(REGION_WEAPON_BEGIN + NUM_ITEM_WEAPON - 1)			/* = 127 */
#define REGION_PROTECTOR_BEGIN		(REGION_WEAPON_END + 1)								/* = 128 */
#define REGION_PROTECTOR_END		(REGION_PROTECTOR_BEGIN + NUM_ITEM_PROTECTOR - 1)	/* = 223 */
#define REGION_TOOL_BEGIN			(REGION_PROTECTOR_END + 1)							/* = 224 */
#define REGION_TOOL_END				(REGION_TOOL_BEGIN + NUM_ITEM_TOOL - 1)				/* = 259 */
#define MSGID_ITEMNAME				3470
#define MSGID_IMPITEMNAME			5717
#define MSGID_ITEMMES				5423
#define MSGID_IMPITEMMES			(MSGID_ITEMMES + NUM_ITEM)	/* = 5683 */

typedef struct itemelem_tag
{
	elem_t elem_half;		/* Ⱦ° */
	elem_t elem_none;		/* ̵° */
	elem_t elem_evade;		/* ° */
	elem_t elem_absorb;		/* ۼ° */
	elem_t elem_double;		/* ° */
} itemelem_t;

typedef struct itemstate_tag
{
	state_t state_defense;	/* Ѳؤ */
	state_t state_apriori;	/* ǽ餫ŬѤѲ */
} itemstate_t;

static void item_setname(void);

static char tbl_itemname[NUM_ITEM][256];	/* ƥ̾Ǽơ֥ */
static int flag_setitemname = 0;			/* tbl_itemname˥ǡǼƤ뤫ɤ */

static const int tbl_increase[8][2] =
{
	{ 0,  1},
	{ 0,  2},
	{ 0,  3},
	{-1,  1},
	{-1,  0},
	{-5,  5},
	{-5,  0},
	{ 0,  5},
};

const char *item_getname(int index)
{
	/* å */
	if (index < 0 || NUM_ITEM <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setitemname)
		item_setname();

	return tbl_itemname[index];
}

const char *item_getequiptype(int index)
{
	static const char *str_equiptype[NUM_EQUIPTYPE] =
	{
		STR_NOTHING_16,
		"ɶ",
		"ʥ",
		"Ǧ",
		"",
		"η",
		"",
		"",

		"",
		"",
		"å",
		"",
		"(̤1)",
		"",
		"è",
		"",

		"٥",
		"",
		"",
		"Ʈ˹",
		"ˡ˹",
		"ɤꤳ",
		"",
		"Ʈ",

		"",
		"",
		"",
		"",
		"ɶ",
		"¾",
		"",
		"(̤2)",
	};

	/* å */
	if (index < 0 || NUM_EQUIPTYPE <= index)
	{
		assert(0);
		return "";
	}

	return str_equiptype[index];
}

static void item_setname(void)
{
	int i;

	/* ƥ̾ǡ */
	for (i = 0; i < NUM_ITEM; i++)
		*tbl_itemname[i] = '\0';

	/* ƥ̾ǡɤ߹ */
	for (i = 0; i < NUM_ITEM; i++)
		dec_msg(tbl_itemname[i], sizeof(tbl_itemname[i]), MSGID_ITEMNAME + i, FLAG_KANA);

	flag_setitemname = 1;
}

itemdata_t *item_readdata(void)
{
	int i;
	itemdata_t *buf_itemdata = NULL;
	itemelem_t tbl_element[NUM_ITEMELEM];
	itemstate_t tbl_state[NUM_ITEMSTATE];
	const int tbl_time_delay[4] = { 0, 10, 20, 30 };

	/* ƥγƼǡǼХåեݤ */
	if ((buf_itemdata = malloc(sizeof(itemdata_t) * NUM_ITEM)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ƥ(ɶ)°ǡɤ߹ */
	for (i = 0; i < NUM_ITEMELEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMELEM];

		file_read(buf, FILEPOS_ITEMELEM + SIZEOF_ITEMELEM * i, SIZEOF_ITEMELEM);

		tbl_element[i].elem_half   = buf[3];
		tbl_element[i].elem_none   = buf[2];
		tbl_element[i].elem_evade  = buf[1];
		tbl_element[i].elem_absorb = buf[0];
		tbl_element[i].elem_double = buf[4];
	}

	/* ƥ(ɶ)ξѲǡɤ߹ */
	for (i = 0; i < NUM_ITEMSTATE; i++)
	{
		unsigned char buf[SIZEOF_ITEMSTATE];

		file_read(buf, FILEPOS_ITEMSTATE + SIZEOF_ITEMSTATE * i, SIZEOF_ITEMSTATE);

		tbl_state[i].state_defense = bits_make32(buf[4], buf[5], buf[6], 0x00);
		tbl_state[i].state_apriori = bits_make32(buf[0], buf[1], buf[2], buf[3]);
	}

	/* ƥβʥǡɤ߹ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMPRICE];

		file_read(buf, FILEPOS_ITEMPRICE + SIZEOF_ITEMPRICE * i, SIZEOF_ITEMPRICE);

		buf_itemdata[i].price_buy  = bits_make32(buf[4], buf[5], buf[6], buf[7]);
		buf_itemdata[i].price_sell = ((buf[0] & 0x80) != 0x00) ? -1 :
		                             ((buf[0] & 0x40) != 0x00) ?  5 : buf_itemdata[i].price_buy / 2;
	}

	/* ƥγƼǡɤ߹ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		/* ƥμȽ̤ */
		if (REGION_WEAPON_BEGIN <= i && i <= REGION_WEAPON_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_WEAPON];

			file_read(buf, FILEPOS_ITEMDATA_WEAPON + SIZEOF_ITEMDATA_WEAPON * (i - REGION_WEAPON_BEGIN), SIZEOF_ITEMDATA_WEAPON);

			/* ͷϥǡ */
			buf_itemdata[i].message_no                    = buf[ 6];
			buf_itemdata[i].param.weapon.attack_power     = buf[11];
			buf_itemdata[i].param.weapon.equip_type       = buf[ 7] & 0x1f;
			buf_itemdata[i].param.weapon.using_effect_no  = buf[10] & 0x7f;
			buf_itemdata[i].param.weapon.data.effect_type = buf[ 0];
			buf_itemdata[i].param.weapon.data.time_wait   = tbl_time_delay[buf[4] & 0x03];

			/* ƥǡ */
			buf_itemdata[i].param.weapon.data.category    = buf[ 5];

			/* ե饰ϥǡ */
			buf_itemdata[i].param.weapon.flag_reflect             = ((buf_itemdata[i].param.weapon.data.category & 0x0f) != 0x00)
																	&& ((buf[8] & 0x80) == 0x00);
			buf_itemdata[i].param.weapon.flag_break               = ((buf[10] & 0x80) != 0x00);
			buf_itemdata[i].param.weapon.flag_throw               = ((buf[ 7] & 0x40) == 0x00);
			buf_itemdata[i].param.weapon.flag_enchant             = ((buf[ 9] & 0x04) != 0x00);
			buf_itemdata[i].param.weapon.flag_initiative          = ((buf[ 9] & 0x20) != 0x00);
			buf_itemdata[i].param.weapon.flag_evade               = ((buf[ 9] & 0x40) != 0x00);
			buf_itemdata[i].param.weapon.flag_evade_f             = ((buf[ 9] & 0x80) != 0x00);
			buf_itemdata[i].param.weapon.flag_additional_effect   = ((buf[ 9] & 0x08) != 0x00);
			buf_itemdata[i].param.weapon.flag_additional_effect_r = ((buf[ 9] & 0x02) != 0x00);
			buf_itemdata[i].param.weapon.flag_wonder_wand         = ((buf[ 9] & 0x01) != 0x00);
			buf_itemdata[i].param.weapon.flag_2handed             = ((buf[ 6] & 0x40) != 0x00);
			buf_itemdata[i].param.weapon.flag_2handed_f           = ((buf[ 6] & 0x80) != 0x00);

			/* ϰϥǡ */
			buf_itemdata[i].param.weapon.data.direction           = ((buf[ 4] & 0x08) != 0x00) ? 1 : 0;
			buf_itemdata[i].param.weapon.data.change_direction    = ((buf[ 4] & 0x10) != 0x00) ? 1 : 0;
			buf_itemdata[i].param.weapon.data.change_target       = ((buf[ 4] & 0x20) != 0x00) ? 1 : 0;
			buf_itemdata[i].param.weapon.data.range               = ((buf[ 4] & 0x40) != 0x00) ? 1 : 0;
			buf_itemdata[i].param.weapon.data.change_range        = ((buf[ 4] & 0x80) != 0x00) ? 1 : 0;
			buf_itemdata[i].param.weapon.data.auto_direction      = (buf[4] == 0x00 || buf[4] == 0x08);

			/* ɲø̥ǡ */
			if (buf_itemdata[i].param.weapon.flag_additional_effect ||
				buf_itemdata[i].param.weapon.flag_additional_effect_r)
			{
				buf_itemdata[i].param.weapon.additional_effect_rate = buf[2];
				buf_itemdata[i].param.weapon.additional_effect_no   = buf[3];
			}
			else
			{
				buf_itemdata[i].param.weapon.additional_effect_rate = 0;
				buf_itemdata[i].param.weapon.additional_effect_no   = 0;
			}

			if ((buf[8] & 0x80) != 0x00)
			{
				/* ѥ᡼Ѳǡ */
				buf_itemdata[i].param.weapon.increase.strength =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x40) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.weapon.increase.speed =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x20) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.weapon.increase.stamina =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x10) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.weapon.increase.magic_power =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x08) != 0x00) ? 1 : 0];

				buf_itemdata[i].param.weapon.elem_powerup = 0x00;
			}
			else
			{
				/* ѥå°ǡ */
				buf_itemdata[i].param.weapon.increase.strength    = 0;
				buf_itemdata[i].param.weapon.increase.speed       = 0;
				buf_itemdata[i].param.weapon.increase.stamina     = 0;
				buf_itemdata[i].param.weapon.increase.magic_power = 0;

				buf_itemdata[i].param.weapon.elem_powerup = buf[8] & 0x7f;
			}

			/* ѥǡ() */
			buf_itemdata[i].param.weapon.data.effect_power      =    0;
			buf_itemdata[i].param.weapon.data.hit_rate          =  100;
			buf_itemdata[i].param.weapon.data.critical_hit_rate =    0;
			buf_itemdata[i].param.weapon.data.magic_point       =    0;
			buf_itemdata[i].param.weapon.data.elem_effect       = 0x00;
			buf_itemdata[i].param.weapon.data.state_effect      = 0x00;
			buf_itemdata[i].param.weapon.data.montype_effect    = 0x00;
			buf_itemdata[i].param.weapon.flag_backrow           =    0;
			buf_itemdata[i].param.weapon.flag_magic_attack      =    0;

			/* ѥǡ() */
			switch (buf_itemdata[i].param.weapon.data.effect_type)
			{
			case 0x30:	/* Ǽꡦ(ƥҥåΨ̤ѡ̤) */
			case 0x37:	/* (ƥҥåΨɲøȯưΨɲøֹ̤) */
				buf_itemdata[i].param.weapon.data.critical_hit_rate = buf[1];
				break;

			case 0x31:	/* (°ɲøȯưΨɲøֹ̤) */
			case 0x32:	/* ʥաû(°ɲøȯưΨɲøֹ̤) */
			case 0x33:	/* (°ɲøȯưΨɲøֹ̤) */
				buf_itemdata[i].param.weapon.data.elem_effect = buf[1];
				break;

			case 0x00:	/* ǤϤʤ(̿ΨɲøȯưΨɲøֹ̤) */
			case 0x34:	/* ࡦϥޡ(̿ΨɲøȯưΨɲøֹ̤) */
				buf_itemdata[i].param.weapon.data.hit_rate = buf[1];
				break;

			case 0x38:	/* (̿ΨɲøȯưΨɲøֹ̤) */
			case 0x3a:	/* ࡦ٥(̿ΨɲøȯưΨɲøֹ̤) */
				buf_itemdata[i].param.weapon.data.hit_rate = buf[1];
				buf_itemdata[i].param.weapon.flag_backrow  = 1;
				break;

			case 0x35:	/* : ³Ū⤷ϰŪξѲȤʤ(̿ΨѲȯưΨѲ) */
				buf_itemdata[i].param.weapon.data.hit_rate          = buf[1];
				buf_itemdata[i].param.weapon.additional_effect_rate = buf[2] & 0x7f;
				buf_itemdata[i].param.weapon.data.state_effect      = ((buf[2] & 0x80) != 0x00) ? bits_make32(buf[3], 0x00, 0x00, 0x00) : bits_make32(0x00, buf[3], 0x00, 0x00);
				buf_itemdata[i].param.weapon.flag_backrow           = 1;
				break;

			case 0x36:	/* : °ƥҥåȤΤ빶(̿ΨƥҥåΨ°) */
				buf_itemdata[i].param.weapon.data.hit_rate          = buf[1];
				buf_itemdata[i].param.weapon.data.critical_hit_rate = buf[2];
				buf_itemdata[i].param.weapon.data.elem_effect       = buf[3];
				buf_itemdata[i].param.weapon.flag_backrow           = 1;
				break;

			case 0x39:	/* ٥: ޥΤ/ƥ󥫡٥(̤ѡ̤ѡ̤) */
				buf_itemdata[i].param.weapon.flag_backrow      = 1;
				buf_itemdata[i].param.weapon.flag_magic_attack = 1;
				break;

			case 0x3b:	/* å(̿Ψ̤ѡ°) */
				buf_itemdata[i].param.weapon.data.hit_rate     = buf[1];
				buf_itemdata[i].param.weapon.data.elem_effect  = buf[3];
				buf_itemdata[i].param.weapon.flag_backrow      = 1;
				buf_itemdata[i].param.weapon.flag_magic_attack = 1;
				break;

			case 0x3c:	/* ޥåݥȤ񤷤ƥƥ(̿ΨޥåݥȤ񤷤ɲä빶ϡޥåݥ) */
				buf_itemdata[i].param.weapon.data.hit_rate     = buf[1];
				buf_itemdata[i].param.weapon.data.effect_power = buf[2] + buf_itemdata[i].param.weapon.attack_power;
				buf_itemdata[i].param.weapon.data.magic_point  = buf[3];
				break;

			case 0x28:	/* ᡼(̿Ψ᡼(̥Х)᡼(̥Х)) */
				buf_itemdata[i].param.weapon.data.hit_rate     = buf[1];
				buf_itemdata[i].param.weapon.data.effect_power = (buf[3] << 8) | buf[2];
				buf_itemdata[i].param.weapon.flag_backrow      = 1;
				buf_itemdata[i].param.weapon.flag_magic_attack = 1;
				break;

			case 0x49:	/* /ꥹΥå(̿ΨɲøȯưΨɲøֹ̤) */
				buf_itemdata[i].param.weapon.data.hit_rate     = buf[1];
				buf_itemdata[i].param.weapon.flag_backrow      = 1;
				buf_itemdata[i].param.weapon.flag_magic_attack = 1;
				break;

			case 0x0d:	/* ҥåȥݥȤۼ빶(̿Ψ̤͡) */
				buf_itemdata[i].param.weapon.data.hit_rate     = buf[1];
				buf_itemdata[i].param.weapon.data.effect_power = buf[2];
				buf_itemdata[i].param.weapon.flag_backrow      = 1;
				buf_itemdata[i].param.weapon.flag_magic_attack = 1;
				break;

			case 0x06:	/* °(̤ѡ͡°) */
				buf_itemdata[i].param.weapon.data.effect_power = buf[2];
				buf_itemdata[i].param.weapon.data.elem_effect  = buf[3];
				buf_itemdata[i].param.weapon.flag_backrow      = 1;
				buf_itemdata[i].param.weapon.flag_magic_attack = 1;
				break;

			case 0x6c:	/* μŨ˥ƥҥåȤ빶(̤ѡͭ͡󥹥) */
				buf_itemdata[i].param.weapon.data.effect_power   = buf[2];
				buf_itemdata[i].param.weapon.data.montype_effect = buf[3];
				buf_itemdata[i].param.weapon.flag_backrow        = 1;
				buf_itemdata[i].param.weapon.flag_magic_attack   = 1;
				break;

			case 0x72:	/* μŨ˥ƥҥåȤ빶(ͭ󥹥סɲøȯưΨɲøֹ̤) */
				buf_itemdata[i].param.weapon.data.montype_effect = buf[1];
				buf_itemdata[i].param.weapon.flag_backrow  = 1;
				break;

			case 0x73:	/* : μŨ˥ƥҥåȤ빶(ͭ󥹥סɲøȯưΨɲøֹ̤) */
				buf_itemdata[i].param.weapon.data.montype_effect = buf[1];
				break;

			case 0x8e:	/* °(̿ΨɲøȯưΨɲøֹ̤) */
				buf_itemdata[i].param.weapon.data.hit_rate     = buf[1];
				buf_itemdata[i].param.weapon.data.elem_effect  = ELEM_POISON;
				buf_itemdata[i].param.weapon.flag_backrow      = 1;
				buf_itemdata[i].param.weapon.flag_magic_attack = 1;
				break;

			case 0x64:	/* ʥ(̤ѡɲøȯưΨɲøֹ̤) */
			case 0x6e:	/* ֥쥤֥֥쥤(̤ѡ̤ѡ̤) */
			case 0x7f:	/* ʤ(ɲøȯưΨɲøֹ̤) */
			default:
				break;
			}

			buf_itemdata[i].type = ITEM_TYPE_WEAPON;
		}
		else if (REGION_PROTECTOR_BEGIN <= i && i <= REGION_PROTECTOR_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_PROTECTOR];

			file_read(buf, FILEPOS_ITEMDATA_PROTECTOR + SIZEOF_ITEMDATA_PROTECTOR * (i - REGION_PROTECTOR_BEGIN), SIZEOF_ITEMDATA_PROTECTOR);

			/* ͷϥǡ */
			buf_itemdata[i].message_no                          = buf[6];
			buf_itemdata[i].param.protector.defense_power       = buf[1];
			buf_itemdata[i].param.protector.evade_rate          = buf[0];
			buf_itemdata[i].param.protector.magic_defense_power = buf[3];
			buf_itemdata[i].param.protector.magic_evade_rate    = buf[2];
			buf_itemdata[i].param.protector.weight              = buf[5];
			buf_itemdata[i].param.protector.equip_type          = buf[7] & 0x1f;
			buf_itemdata[i].param.protector.protector_type      = buf[4];

			/* °ϥǡ */
			buf_itemdata[i].param.protector.elem_half   = tbl_element[buf[10]].elem_half;
			buf_itemdata[i].param.protector.elem_none   = tbl_element[buf[10]].elem_none;
			buf_itemdata[i].param.protector.elem_evade  = tbl_element[buf[10]].elem_evade;
			buf_itemdata[i].param.protector.elem_absorb = tbl_element[buf[10]].elem_absorb;
			buf_itemdata[i].param.protector.elem_double = tbl_element[buf[10]].elem_double;

			/* Ѳϥǡ */
			buf_itemdata[i].param.protector.state_defense = tbl_state[buf[11]].state_defense;
			buf_itemdata[i].param.protector.state_apriori = tbl_state[buf[11]].state_apriori;

			/* ե饰ϥǡ */
			buf_itemdata[i].param.protector.flag_evade       = ((buf[9] & 0x40) != 0x00);
			buf_itemdata[i].param.protector.flag_magic_evade = ((buf[9] & 0x80) != 0x00);
			buf_itemdata[i].param.protector.flag_steal       = ((buf[9] & 0x10) != 0x00);
			buf_itemdata[i].param.protector.flag_capture     = ((buf[9] & 0x01) != 0x00);
			buf_itemdata[i].param.protector.flag_monk_fight  = ((buf[9] & 0x20) != 0x00);
			buf_itemdata[i].param.protector.flag_mp_half     = ((buf[9] & 0x08) != 0x00);
			buf_itemdata[i].param.protector.flag_sword_dance = ((buf[9] & 0x04) != 0x00);
			buf_itemdata[i].param.protector.flag_undead      = ((buf[9] & 0x02) != 0x00);

			if ((buf[8] & 0x80) != 0x00)
			{
				/* ѥ᡼Ѳǡ */
				buf_itemdata[i].param.protector.increase.strength =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x40) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.protector.increase.speed =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x20) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.protector.increase.stamina =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x10) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.protector.increase.magic_power =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x08) != 0x00) ? 1 : 0];

				buf_itemdata[i].param.protector.elem_powerup = 0x00;
			}
			else
			{
				/* ѥå°ǡ */
				buf_itemdata[i].param.protector.increase.strength    = 0;
				buf_itemdata[i].param.protector.increase.speed       = 0;
				buf_itemdata[i].param.protector.increase.stamina     = 0;
				buf_itemdata[i].param.protector.increase.magic_power = 0;

				buf_itemdata[i].param.protector.elem_powerup = buf[8] & 0x7f;
			}

			buf_itemdata[i].type = ITEM_TYPE_PROTECTOR;
		}
		else if (REGION_TOOL_BEGIN <= i && i <= REGION_TOOL_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_TOOL];

			file_read(buf, FILEPOS_ITEMDATA_TOOL + SIZEOF_ITEMDATA_TOOL * (i - REGION_TOOL_BEGIN), SIZEOF_ITEMDATA_TOOL);

			/* ͷϥǡ */
			buf_itemdata[i].message_no = buf[6];
			buf_itemdata[i].type       = ITEM_TYPE_TOOL;

			/* ե饰ϥǡ */
			buf_itemdata[i].param.tool.flag_throw      = ((buf[7] & 0x40) == 0x00);
			buf_itemdata[i].param.tool.flag_tool       = ((buf[7] & 0x20) == 0x00);
			buf_itemdata[i].param.tool.flag_drink      = ((buf[7] & 0x10) == 0x00);
			buf_itemdata[i].param.tool.flag_magic_lamp = ((buf[7] & 0x08) != 0x00);
			buf_itemdata[i].param.tool.flag_mix        = ((buf[7] & 0x02) == 0x00);

			/* ˡǡѤι¤Ĵ: buf[6] <=> buf[7] */
			buf[6] = buf[7];
			buf[7] = buf_itemdata[i].message_no; /* buf[6] */

			mag_setparam(&buf_itemdata[i].param.tool.data, buf, sizeof(buf));
		}
		else
			assert(0);
	}

	return buf_itemdata;
}

int item_writeparam(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥοͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥοͷϥǡϤ() */
	fprintf(fp, "No.       ̾                 "
				"      "
				"  ̿Ψ   CHΨ "
				"Ի "
				"     ®     \n");
	for (i = REGION_WEAPON_BEGIN; i <= REGION_WEAPON_END; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s : %6ld,%6ld, %5d, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d, %4Xh\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price_buy,
				ptr_itemdata[i].price_sell,
				ptr_itemdata[i].param.weapon.attack_power,
				ptr_itemdata[i].param.weapon.data.effect_power,
				ptr_itemdata[i].param.weapon.data.hit_rate,
				ptr_itemdata[i].param.weapon.data.critical_hit_rate,
				ptr_itemdata[i].param.weapon.data.time_wait,
				ptr_itemdata[i].param.weapon.increase.strength,
				ptr_itemdata[i].param.weapon.increase.speed,
				ptr_itemdata[i].param.weapon.increase.stamina,
				ptr_itemdata[i].param.weapon.increase.magic_power,
				ptr_itemdata[i].param.weapon.data.effect_type);
	}

	/* ƥοͷϥǡϤ(ɶ) */
	fprintf(fp, "No.       ̾                 "
				"      "
				"ɸ Ψ  Ψ    "
				"     ®  \n");
	for (i = REGION_PROTECTOR_BEGIN; i <= REGION_PROTECTOR_END; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s : %6ld,%6ld, %5d, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price_buy,
				ptr_itemdata[i].price_sell,
				ptr_itemdata[i].param.protector.defense_power,
				ptr_itemdata[i].param.protector.evade_rate,
				ptr_itemdata[i].param.protector.magic_defense_power,
				ptr_itemdata[i].param.protector.magic_evade_rate,
				ptr_itemdata[i].param.protector.weight,
				ptr_itemdata[i].param.protector.increase.strength,
				ptr_itemdata[i].param.protector.increase.speed,
				ptr_itemdata[i].param.protector.increase.stamina,
				ptr_itemdata[i].param.protector.increase.magic_power);
	}

	/* ƥοͷϥǡϤ(ƻ) */
	fprintf(fp, "No.       ̾                 "
				"      "
				"   Ψ ̿Ψ  "
				"Ի "
				"  \n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		fprintf(fp, (ptr_itemdata[i].param.tool.data.effect_rate != 0) ?
					"%03d(%03X): %-18s : %6ld,%6ld, %5d,%3d/16, %5d, %5d, %5d, %4Xh\n" :
					"%03d(%03X): %-18s : %6ld,%6ld, %5d, %5d, %5d, %5d, %5d, %4Xh\n",
				i, i, item_getname(i),
				ptr_itemdata[i].price_buy,
				ptr_itemdata[i].price_sell,
				ptr_itemdata[i].param.tool.data.effect_power,
				ptr_itemdata[i].param.tool.data.effect_rate,
				ptr_itemdata[i].param.tool.data.hit_rate,
				ptr_itemdata[i].param.tool.data.fixed_value,
				ptr_itemdata[i].param.tool.data.time_wait,
				ptr_itemdata[i].param.tool.data.effect_type);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeelem(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥ°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥ°ϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"Ⱦ/°     "
				"̵°          "
				"ۼ°          "
				"°          "
				"ѥå°\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			fprintf(fp, " ");

			/* ° */
			str_printelem(fp, ptr_itemdata[i].param.weapon.data.elem_effect, ELEM_ALL);
			fprintf(fp, "                                                        ");

			/* ѥå° */
			str_printelem(fp, ptr_itemdata[i].param.weapon.elem_powerup, ELEM_ALL);
			break;

		case ITEM_TYPE_PROTECTOR:
			fprintf(fp, " ");

			/* Ⱦ° */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_half, ELEM_ALL);
			fprintf(fp, "  ");

			/* ̵° */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_none | ptr_itemdata[i].param.protector.elem_evade, ELEM_ALL);
			fprintf(fp, "  ");

			/* ۼ° */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_absorb, ELEM_ALL);
			fprintf(fp, "  ");

			/* ° */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_double, ELEM_ALL);
			fprintf(fp, "  ");

			/* ѥå° */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_powerup, ELEM_ALL);
			break;

		case ITEM_TYPE_TOOL:
			fprintf(fp, " ");

			/* ° */
			str_printelem(fp, ptr_itemdata[i].param.tool.data.elem_effect, ELEM_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writecategory(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥΥƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥΥƥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"ƥ\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s :", i, i, item_getname(i));

		if (ptr_itemdata[i].type == ITEM_TYPE_WEAPON)
		{
			fprintf(fp, " ");
			str_printcategory(fp, ptr_itemdata[i].param.weapon.data.category, CATEGORY_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writestate(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥξѲϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"Ѳ1    2                3                 "
				"Ѳ1    2                3                4\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			break;

		case ITEM_TYPE_PROTECTOR:
			fprintf(fp, " ");

			/* Ѳؤ */
			str_printstate(fp, ptr_itemdata[i].param.protector.state_defense, STATE_ALL & ~STATE_OTHER);
			fprintf(fp, "  ");

			/* ǽ餫ŬѤѲ */
			str_printstate(fp, ptr_itemdata[i].param.protector.state_apriori, STATE_ALL);
			break;

		case ITEM_TYPE_TOOL:
			fprintf(fp, "                                                     ");

			/* Ѳ */
			str_printstate(fp, ptr_itemdata[i].param.tool.data.state_effect, STATE_ALL);
			break;
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeflag(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥΥե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥΥե饰ϥǡϤ() */
	fprintf(fp, "No.       ̾                 "
				"          "
				"ꤲ  ȿ  ξ ξ  "
				"ͭ²\n");
	for (i = REGION_WEAPON_BEGIN; i <= REGION_WEAPON_END; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s : %-16s,  %2s   %2s   %2s   %2s   %2s   %2s", i, i,
				item_getname(i),
				item_getequiptype(ptr_itemdata[i].param.weapon.equip_type),
				str_getflag(ptr_itemdata[i].param.weapon.flag_throw),
				str_getflag(ptr_itemdata[i].param.weapon.flag_backrow),
				str_getflag(ptr_itemdata[i].param.weapon.flag_reflect),
				str_getflag(ptr_itemdata[i].param.weapon.flag_enchant),
				str_getflag(ptr_itemdata[i].param.weapon.flag_2handed),
				str_getflag(ptr_itemdata[i].param.weapon.flag_2handed_f));

		/* ͭ󥹥 */
		if (ptr_itemdata[i].param.weapon.data.montype_effect != 0x00)
		{
			fprintf(fp, "   ");
			str_printmontype(fp, ptr_itemdata[i].param.weapon.data.montype_effect, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	/* ƥΥե饰ϥǡϤ(ɶ) */
	fprintf(fp, "No.       ̾                 "
				"\n");
	for (i = REGION_PROTECTOR_BEGIN; i <= REGION_PROTECTOR_END; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s : %-16s\n", i, i,
				item_getname(i),
				item_getequiptype(ptr_itemdata[i].param.protector.equip_type));
	}

	/* ƥΥե饰ϥǡϤ() */
	fprintf(fp, "No.       ̾                 "
				"ꤲ   Ĵ ȿ  "
				"ϰ\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s :  %2s   %2s   %2s   %2s   %2s   ", i, i,
				item_getname(i),
				str_getflag(ptr_itemdata[i].param.tool.flag_throw),
				str_getflag(ptr_itemdata[i].param.tool.flag_tool),
				str_getflag(ptr_itemdata[i].param.tool.flag_drink),
				str_getflag(ptr_itemdata[i].param.tool.flag_mix),
				str_getflag(ptr_itemdata[i].param.tool.data.flag_reflect));

		/* ϰ */
		mag_printtarget(fp, &ptr_itemdata[i].param.tool.data);

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeeffect(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥθ̥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥθ̥ǡϤ */
	fprintf(fp, "No.       ̾                 "
				"˲ ɲ Ѳ  "
				"Ѹ          "
				"ϰ           "
				"ɲø          "
				"ȯưΨ\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		char str[256] = "";
		const char *ptr_name_using      = STR_NOTHING_16;
		const char *ptr_name_additional = STR_NOTHING_16;

		fprintf(fp, "%03d(%03X): %-18s :", i, i, item_getname(i));

		if (ptr_itemdata[i].type == ITEM_TYPE_WEAPON)
		{
			/* Ѹ */
			ptr_name_using = mag_getname(ptr_itemdata[i].param.weapon.using_effect_no);

			/* ɲø */
			if (ptr_itemdata[i].param.weapon.data.state_effect != 0x00)
			{
				char str_state[64] = "";

				snprintf(str, sizeof(str), "%s: %s",
							"Ѳ",
							str_setstate(str_state, ptr_itemdata[i].param.weapon.data.state_effect, sizeof(str_state)));
				ptr_name_additional = str;
			}
			else if (ptr_itemdata[i].param.weapon.flag_additional_effect_r)
				ptr_name_additional = chr_getcmdname(ptr_itemdata[i].param.weapon.additional_effect_no);
			else if (ptr_itemdata[i].param.weapon.flag_additional_effect)
				ptr_name_additional = mag_getname(ptr_itemdata[i].param.weapon.additional_effect_no);

			/* Ƽե饰/Ѹ */
			fprintf(fp, "  %2s   %2s   %2s   %-16s, ",
					str_getflag(ptr_itemdata[i].param.weapon.flag_break),
					str_getflag(ptr_itemdata[i].param.weapon.flag_additional_effect),
					str_getflag(ptr_itemdata[i].param.weapon.flag_additional_effect_r),
					ptr_name_using);

			/* ϰ */
			mag_printtarget(fp, &ptr_itemdata[i].param.weapon.data);

			/* ɲø */
			fprintf(fp, ", %-16s,  %5d",
					ptr_name_additional,
					ptr_itemdata[i].param.weapon.additional_effect_rate);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeparambug(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥοͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	for (j = 0; j < 2; j++)
	{
		long filepos_itemdata = 0, sizeof_itemdata  = 0;

		switch (j)
		{
		case 0:
			/* Ʈ˻ѤɶǡΥɥ쥹 */
			filepos_itemdata = FILEPOS_ITEMDATA_PROTECTOR + SIZEOF_ITEMDATA_PROTECTOR * NUM_ITEM_PROTECTOR;
			sizeof_itemdata  = SIZEOF_ITEMDATA_PROTECTOR;
			break;

		case 1:
			/* ̤ɽɶǡΥɥ쥹 */
			filepos_itemdata = FILEPOS_ITEMDATA_TOOL;
			sizeof_itemdata  = SIZEOF_ITEMDATA_TOOL;
			break;

		default:
			assert(0);
		}

		/* ƥοͷϥǡϡϤ */
		fprintf(fp, "No.       ̾                 "
					"ɸ Ψ  Ψ    "
					"     ®  \n");
		for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
		{
			unsigned char buf[SIZEOF_ITEMDATA_PROTECTOR];
			protector_t tmp_protector;

			file_read(buf, filepos_itemdata + sizeof_itemdata * (i - REGION_TOOL_BEGIN), SIZEOF_ITEMDATA_PROTECTOR);

			/* ͷϥǡ */
			tmp_protector.defense_power       = buf[1];
			tmp_protector.evade_rate          = buf[0];
			tmp_protector.magic_defense_power = buf[3];
			tmp_protector.magic_evade_rate    = buf[2];
			tmp_protector.weight              = buf[5];

			if ((buf[8] & 0x80) != 0x00)
			{
				/* ѥ᡼Ѳǡ */
				tmp_protector.increase.strength =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x40) != 0x00) ? 1 : 0];
				tmp_protector.increase.speed =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x20) != 0x00) ? 1 : 0];
				tmp_protector.increase.stamina =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x10) != 0x00) ? 1 : 0];
				tmp_protector.increase.magic_power =
					tbl_increase[buf[8] & 0x07][((buf[8] & 0x08) != 0x00) ? 1 : 0];
			}
			else
			{
				tmp_protector.increase.strength    = 0;
				tmp_protector.increase.speed       = 0;
				tmp_protector.increase.stamina     = 0;
				tmp_protector.increase.magic_power = 0;
			}

			fprintf(fp, "%03d(%03X): %-18s :  %5d, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d\n", i, i,
					item_getname(i),
					tmp_protector.defense_power,
					tmp_protector.evade_rate,
					tmp_protector.magic_defense_power,
					tmp_protector.magic_evade_rate,
					tmp_protector.weight,
					tmp_protector.increase.strength,
					tmp_protector.increase.speed,
					tmp_protector.increase.stamina,
					tmp_protector.increase.magic_power);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeelembug(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	itemelem_t tbl_element[256];

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥ°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥ(ɶ)°ǡɤ߹ */
	for (i = 0; i < 256; i++)
	{
		unsigned char buf[SIZEOF_ITEMELEM];

		file_read(buf, FILEPOS_ITEMELEM + SIZEOF_ITEMELEM * i, SIZEOF_ITEMELEM);

		tbl_element[i].elem_half   = buf[3];
		tbl_element[i].elem_none   = buf[2];
		tbl_element[i].elem_absorb = buf[0];
		tbl_element[i].elem_double = buf[4];
	}

	/* ƥ°ϥǡϡϤ */
	fprintf(fp, "No.       ̾                 "
				"Ⱦ/°     "
				"̵°          "
				"ۼ°          "
				"°          "
				"ѥå°\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		unsigned char buf[SIZEOF_ITEMDATA_PROTECTOR];
		protector_t tmp_protector;

		file_read(buf, FILEPOS_ITEMDATA_PROTECTOR + SIZEOF_ITEMDATA_PROTECTOR * (i - REGION_PROTECTOR_BEGIN), SIZEOF_ITEMDATA_PROTECTOR);

		/* °ϥǡ */
		tmp_protector.elem_half   = tbl_element[buf[10]].elem_half;
		tmp_protector.elem_none   = tbl_element[buf[10]].elem_none;
		tmp_protector.elem_absorb = tbl_element[buf[10]].elem_absorb;
		tmp_protector.elem_double = tbl_element[buf[10]].elem_double;

		/* ѥå°ǡ */
		tmp_protector.elem_powerup = ((buf[8] & 0x80) == 0x00) ? buf[8] & 0x7f : 0x00;

		fprintf(fp, "%03d(%03X): %-18s : ", i, i, item_getname(i));

		/* Ⱦ° */
		str_printelem(fp, tmp_protector.elem_half, ELEM_ALL);
		fprintf(fp, "  ");

		/* ̵° */
		str_printelem(fp, tmp_protector.elem_none, ELEM_ALL);
		fprintf(fp, "  ");

		/* ۼ° */
		str_printelem(fp, tmp_protector.elem_absorb, ELEM_ALL);
		fprintf(fp, "  ");

		/* ° */
		str_printelem(fp, tmp_protector.elem_double, ELEM_ALL);
		fprintf(fp, "  ");

		/* ѥå° */
		str_printelem(fp, tmp_protector.elem_powerup, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writestatbug(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	itemstate_t tbl_state[256];

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥ(ɶ)ξѲǡɤ߹ */
	for (i = 0; i < 256; i++)
	{
		unsigned char buf[SIZEOF_ITEMSTATE];

		file_read(buf, FILEPOS_ITEMSTATE + SIZEOF_ITEMSTATE * i, SIZEOF_ITEMSTATE);

		tbl_state[i].state_defense = bits_make32(buf[4], buf[5], buf[6], 0x00);
		tbl_state[i].state_apriori = bits_make32(buf[0], buf[1], buf[2], buf[3]);
	}

	/* ƥξѲϥǡϡϤ */
	fprintf(fp, "No.       ̾                 "
				"Ѳ1    2                3                "
				"Ѳ1    2                3                4\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		unsigned char buf[SIZEOF_ITEMDATA_PROTECTOR];
		protector_t tmp_protector;

		file_read(buf, FILEPOS_ITEMDATA_PROTECTOR + SIZEOF_ITEMDATA_PROTECTOR * (i - REGION_PROTECTOR_BEGIN), SIZEOF_ITEMDATA_PROTECTOR);

		/* Ѳϥǡ */
		tmp_protector.state_defense = tbl_state[buf[11]].state_defense;
		tmp_protector.state_apriori = tbl_state[buf[11]].state_apriori;

		fprintf(fp, "%03d(%03X): %-18s : ", i, i, item_getname(i));

		/* Ѳؤ */
		str_printstate(fp, tmp_protector.state_defense, STATE_ALL & ~STATE_OTHER);
		fprintf(fp, "  ");

		/* ǽ餫ŬѤѲ */
		str_printstate(fp, tmp_protector.state_apriori, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeflagbug(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥΥե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥΥե饰ϥǡϡϤ */
	fprintf(fp, "No.       ̾                 "
				" Ի ٤ MPȾ  Ʈ  \n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		unsigned char buf[SIZEOF_ITEMDATA_PROTECTOR];
		protector_t tmp_protector;

		file_read(buf, FILEPOS_ITEMDATA_PROTECTOR + SIZEOF_ITEMDATA_PROTECTOR * (i - REGION_PROTECTOR_BEGIN), SIZEOF_ITEMDATA_PROTECTOR);

		/* ե饰ϥǡ */
		tmp_protector.flag_evade       = ((buf[9] & 0x40) != 0x00);
		tmp_protector.flag_magic_evade = ((buf[9] & 0x80) != 0x00);
		tmp_protector.flag_steal       = ((buf[9] & 0x10) != 0x00);
		tmp_protector.flag_capture     = ((buf[9] & 0x01) != 0x00);
		tmp_protector.flag_monk_fight  = ((buf[9] & 0x20) != 0x00);
		tmp_protector.flag_mp_half     = ((buf[9] & 0x08) != 0x00);
		tmp_protector.flag_sword_dance = ((buf[9] & 0x04) != 0x00);
		tmp_protector.flag_undead      = ((buf[9] & 0x02) != 0x00);

		fprintf(fp, "%03d(%03X): %-18s :  %2s   %2s   %2s   %2s   %2s   %2s   %2s   %2s\n", i, i,
				item_getname(i),
				str_getflag(tmp_protector.flag_capture),
				str_getflag(tmp_protector.flag_undead),
				str_getflag(tmp_protector.flag_sword_dance),
				str_getflag(tmp_protector.flag_mp_half),
				str_getflag(tmp_protector.flag_steal),
				str_getflag(tmp_protector.flag_monk_fight),
				str_getflag(tmp_protector.flag_evade),
				str_getflag(tmp_protector.flag_magic_evade));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
