/*
 *	ff5magic.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2007-2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff5decode.h"
#include "ff5magic.h"

#define FILEPOS_MAGDATA			0x14a3c4	/* SNES: 0x110b80 */
#define FILEPOS_MAGTIME			0x145e42	/* SNES: 0x10ecee */	/* unused */
#define NUM_MAG					256
#define NUM_MAGTIME				4
#define SIZEOF_MAGDATA			8
#define SIZEOF_MAGTIME			1
#define MSGID_MAGNAME			4738
#define MSGID_MAGMES			3782

static void mag_setname(void);

static char tbl_magname[NUM_MAG][256];	/* ˡ̾Ǽơ֥ */
static int flag_setmagname = 0;			/* tbl_magname˥ǡǼƤ뤫ɤ */

const char *mag_getname(int index)
{
	/* å */
	if (index < 0 || NUM_MAG <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmagname)
		mag_setname();

	return tbl_magname[index];
}

static void mag_setname(void)
{
	int i;

	/* ˡ̾ǡ */
	for (i = 0; i < NUM_MAG; i++)
		*tbl_magname[i] = '\0';

	/* ˡ̾ǡɤ߹ */
	for (i = 0; i < NUM_MAG; i++)
		dec_msg(tbl_magname[i], sizeof(tbl_magname[i]), MSGID_MAGNAME + i, FLAG_KANA);

	/* ص塢(No.215(0D7)) ̾Τѹ */
	strncat(tbl_magname[215], "!", sizeof(tbl_magname[215]) - strlen(tbl_magname[215]) - 1);

	flag_setmagname = 1;
}

magdata_t *mag_readdata(void)
{
	int i;
	magdata_t *buf_magdata = NULL;

	/* ˡγƼǡǼХåեݤ */
	if ((buf_magdata = malloc(sizeof(magdata_t) * NUM_MAG)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ˡγƼǡɤ߹ */
	for (i = 0; i < NUM_MAG; i++)
	{
		unsigned char buf[SIZEOF_MAGDATA];

		file_read(buf, FILEPOS_MAGDATA + SIZEOF_MAGDATA * i, SIZEOF_MAGDATA);
		mag_setparam(&buf_magdata[i], buf, sizeof(buf));
	}

	return buf_magdata;
}

int mag_setparam(magdata_t *ptr_magdata, unsigned char *ptr_buf, size_t bufsize)
{
	const int tbl_time_delay[4] = { 0, 10, 20, 30 };
	const int tbl_learning_rate[4] = { 0, 100, 50, 10 };

	/* å */
	if (ptr_magdata == NULL || ptr_buf == NULL || bufsize < SIZEOF_MAGDATA)
		return 0;

	/* ͷϥǡ */
	ptr_magdata->magic_point = ptr_buf[7] & 0x7f;
	ptr_magdata->effect_type = ptr_buf[0];
	ptr_magdata->attacks_num = ptr_buf[6] & 0x07;
	ptr_magdata->time_wait   = tbl_time_delay[ptr_buf[4] & 0x03];

	if ((ptr_buf[6] & 0xf0) == 0x00)
		ptr_magdata->learning_rate = 0;
	else
	{
		int index;
		for (index = 0; index < 4; index++)
		{
			if ((ptr_buf[6] & (0x80 >> index)) != 0x00)
			{
				ptr_magdata->learning_rate = tbl_learning_rate[index];
				break;
			}
		}
	}

	/* ƥǡ */
	ptr_magdata->category    = ptr_buf[5];

	/* ե饰ϥǡ */
	ptr_magdata->flag_reflect     = ((ptr_buf[7] & 0x80) == 0x00);
	ptr_magdata->flag_roulette    = 0; /*  */
	ptr_magdata->flag_cure        = ((ptr_buf[4] & 0x04) != 0x00);
	ptr_magdata->direction        = ((ptr_buf[4] & 0x08) != 0x00) ? 1 : 0;
	ptr_magdata->change_direction = ((ptr_buf[4] & 0x10) != 0x00) ? 1 : 0;
	ptr_magdata->change_target    = ((ptr_buf[4] & 0x20) != 0x00) ? 1 : 0;
	ptr_magdata->range            = ((ptr_buf[4] & 0x40) != 0x00) ? 1 : 0;
	ptr_magdata->change_range     = ((ptr_buf[4] & 0x80) != 0x00) ? 1 : 0;
	ptr_magdata->auto_direction   = (ptr_buf[4] == 0x00 || ptr_buf[4] == 0x08);

	/* ѥǡ() */
	ptr_magdata->effect_rate    = 0;
	ptr_magdata->hit_rate       = 0;
	ptr_magdata->effect_time    = 0;
	ptr_magdata->effect_level   = 0;
	ptr_magdata->fixed_value    = 0;
	ptr_magdata->elem_effect    = 0x00;
	ptr_magdata->state_effect   = 0x00;
	ptr_magdata->montype_effect = 0x00;

	switch (ptr_magdata->effect_type)
	{
	case 0x03:	/* ˡ(եȡղ°ղþѲ1(³Ū)) */
		ptr_magdata->elem_effect  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);
		break;

	case 0x04:	/* ˡ(եȡղ°ղþѲ2(Ū)) */
		ptr_magdata->elem_effect  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(0x00, ptr_buf[3], 0x00, 0x00);
		break;

	case 0x05:	/* ˡ(եȡղ°) */
		/* (0x20: ޥåݥȤۼ, 0x40: ҥåȥݥȤۼ, 0x80: Ϥ 100 ɲäоݤɸ(ۤ)̵빶) */
		ptr_magdata->elem_effect = ptr_buf[2];
		ptr_magdata->fixed_value = ptr_buf[3];
		break;

	case 0x06:	/* °(̤ѡ͡°) */
	case 0x2d:	/* °(ӥƥȤ̵Ǥ)(̤ѡ͡°) */
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->elem_effect  = ptr_buf[3];
		break;

	case 0x07:	/* ѲȤʤΨ᡼(̿ΨΨѲ2(Ū)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_rate  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(0x00, ptr_buf[3], 0x00, 0x00);
		break;

	case 0x08:	/* ɸ(ۤ)̵ι(̤ѡ͡°) */
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->elem_effect  = ptr_buf[3];
		break;

	case 0x09:	/* ᡼Ψι(̤ѡΨ°) */
		ptr_magdata->fixed_value = ptr_buf[2];
		ptr_magdata->elem_effect = ptr_buf[3];
		break;

	case 0x0a:	/* ʪⰷι(̿Ψ͡°) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->elem_effect  = ptr_buf[3];
		break;

	case 0x0b:	/*  10100 °(°̿ΨѲ1(³Ū) or Ѳ2(Ū)) */
		ptr_magdata->elem_effect  = ptr_buf[1];
		ptr_magdata->hit_rate     = ptr_buf[2] & 0x7f;
		ptr_magdata->state_effect = ((ptr_buf[2] & 0x80) != 0x00) ? bits_make32(ptr_buf[3], 0x00, 0x00, 0x00) : bits_make32(0x00, ptr_buf[3], 0x00, 0x00);
		break;

	case 0x0c:	/* Ѳ(å)Ȥʤɸ(ۤ)̵°(°͡³) */
		ptr_magdata->elem_effect  = ptr_buf[1];
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->effect_time  = ptr_buf[3];
		break;

	case 0x0d:	/* ҥåȥݥȤۼ빶(̿Ψ̤͡) */
	case 0x0e:	/* ޥåݥȤۼ빶(̿Ψ̤͡) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_power = ptr_buf[2];
		break;

	case 0x0f:	/* λ๶(̿Ψ̤ѡ̤) */
		ptr_magdata->hit_rate = ptr_buf[1];
		break;

	case 0x13:	/* Ѳ(̿Ψ³֡Ѳ2(Ū)) */
	case 0x3d:	/* Ѳȼλ๶(̿Ψ³֡Ѳ2(Ū)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_time  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(0x00, ptr_buf[3], 0x00, 0x00);
		break;

	case 0x10:	/* (̤ѡ̤͡) */
	case 0x11:	/* (ɸñΤλ̸)(̤ѡ̤͡) */
	case 0x1b:	/* ҥåȥݥȤۼ빶(ۼʬʬ)(̤ѡ̤͡) */
	case 0x87:	/* ɥ쥤󥿥å(̤ѡ̤͡) */
		ptr_magdata->effect_power = ptr_buf[2];
		break;

	case 0x12:	/* Ѳ(̿Ψ̤ѡѲ1(³Ū)) */
	case 0x4f:	/* 줤(̿Ψ̤ѡѲ1(³Ū)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->state_effect = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);
		break;

	case 0x14:	/* Ѳ(̿Ψ³֡Ѳ3(ˡ)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_time  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[3], 0x00);
		break;

	case 0x15:	/* Ѳ(ȿž)(̿Ψ̤ѡѲ1(³Ū)) */
	case 0x17:	/* Ѳ(Իϥ󥹥ؤü)(̿Ψ̤ѡѲ1(³Ū)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_time  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);
		break;

	case 0x16:	/* Ѳ(ֲȤʤ)(̿ΨѲ3(ˡ)Ѳ3(ˡ)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
/*		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[2], 0x00); */
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[3], 0x00);
		break;

	case 0x18:	/* ¨ˡ(Իϥ󥹥ؤṳ̈ʤ)(̿Ψ̤ѡ̤) */
		ptr_magdata->hit_rate = ptr_buf[1];
		break;

	case 0x19:	/* Ѳ(Ѳ1(³Ū)Ѳ2(Ū)Ѳ3(ˡ)) */
		ptr_magdata->state_effect = bits_make32(~ptr_buf[1], ~ptr_buf[2], ~ptr_buf[3], 0x00);
		break;

	case 0x27:	/* Ѳ(Ѳ1(³Ū)Ѳ2(Ū)Ѳ3(ˡ)) */
		ptr_magdata->state_effect = bits_make32(ptr_buf[1], ptr_buf[2], ptr_buf[3], 0x00);
		break;

	case 0x1a:	/* Ʈǽ(̿Ψ͡Ψ) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->effect_rate  = ptr_buf[3];
		break;

	case 0x1c:	/* °ۼ̵(ۼ°̵°°) */
		ptr_magdata->elem_effect = ptr_buf[1] | ptr_buf[2] | ptr_buf[3];
		break;

	case 0x1f:	/* ߥ塼(Ψ̤ѡ̤) */
	case 0x20:	/* ƥݡȤ󤺤顦(Ψ̤ѡ̤) */
		ptr_magdata->hit_rate = ptr_buf[1];
		break;

	case 0x23:	/* (̤ѡ̤) */
	case 0x46:	/* ˤʤ(񥮥η᡼η̤) */
		ptr_magdata->fixed_value = ptr_buf[2];
		break;

	case 0x24:	/* ̸ΥҥåȥݥȲ(͡Ψ̤) */
	case 0x25:	/* ̸ΥޥåݥȲ(͡Ψ̤) */
		ptr_magdata->effect_power = ptr_buf[1];
		ptr_magdata->fixed_value  = ptr_buf[2];
		break;

	case 0x28:	/* ᡼(̿Ψ᡼(̥Х)᡼(̥Х)) */
		ptr_magdata->hit_rate    = ptr_buf[1];
		ptr_magdata->fixed_value = MIN(((ptr_buf[3] << 8) | ptr_buf[2]), 9999);
		break;

	case 0x29:	/* Ѳ(̿ΨΥȡѲ4(¾)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->fixed_value  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, 0x00, ptr_buf[3]);
		break;

	case 0x2a:	/* Ѳ(å)ȤʤΨ᡼°(°Ψ³) */
		ptr_magdata->elem_effect = ptr_buf[1];
		ptr_magdata->effect_rate = ptr_buf[2];
		ptr_magdata->effect_time = ptr_buf[3];
		break;

	case 0x2b:	/* ʬФѲȤʤΨ᡼°(°ΨѲ1(³Ū)) */
		ptr_magdata->elem_effect  = ptr_buf[1];
		ptr_magdata->effect_rate  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);
		break;

	case 0x2c:	/* Ѳ(2Τɤ餫θ)(Ѳ1(³Ū)³֡Ѳ2(Ū)) */
		ptr_magdata->state_effect = bits_make32(ptr_buf[1], ptr_buf[3], 0x00, 0x00);
		ptr_magdata->effect_time  = ptr_buf[2];
		break;

	case 0x2e:	/* ˤߤΤĤ뤮(̿Ψ͡Ѳ1(³Ū)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);
		break;

	case 0x2f:	/* [̤] 󥹥פ˾Ѳ(󥹥ס̤ѡѲ1(³Ū)) */
		ptr_magdata->montype_effect = ptr_buf[1];
		ptr_magdata->state_effect   = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);

	case 0x3e:	/* äѤ(λ̤ˤʤΨ³֡Ѳ4) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_time  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, 0x00, ptr_buf[3]);
		break;

	case 0x3f:	/* ᡼Ψ꤫Ʈǽ˾Ѳȼ(̤ѡΨ̤) */
		ptr_magdata->fixed_value = ptr_buf[2];
		break;

	case 0x40:	/* ѹ(̿Ψѹμ̡̤) */
		ptr_magdata->hit_rate = ptr_buf[1];
		break;

	case 0x41:	/* [̤]  12 ܤĹ(ϡ̤ѡ̤) */
		ptr_magdata->effect_power = ptr_buf[1];
		break;

	case 0x42:	/*  + Ѳ(ѥ᡼μ̡͡Ѳ1(³Ū) or Ѳ2(Ū)) */
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->state_effect = ((ptr_buf[1] & 0x80) != 0x00) ? bits_make32(~ptr_buf[3], 0x00, 0x00, 0x00) : bits_make32(0x00, ~ptr_buf[3], 0x00, 0x00);
		break;

	case 0x43:	/* ̤(Ψ̤ѡ̤) */
	case 0x45:	/* ʤ(̿Ψ̤ѡ̤) */
	case 0x49:	/* /ꥹΥå(̿ΨɲøȯưΨɲøֹ̤) */
		ptr_magdata->hit_rate = ptr_buf[1];
		break;

	case 0x47:	/* ʤ(̤ѡ³֡Ѳ3(ˡ)) */
		ptr_magdata->effect_time  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[3], 0x00);
		break;

	case 0x4b:	/* ٥ϾѲ(٥롦̤ѡѲ1(³Ū)) */
		ptr_magdata->effect_level = ptr_buf[1];
		ptr_magdata->state_effect = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);
		break;

	case 0x4c:	/* ٥ϾѲ(٥롦³֡Ѳ2(Ū)) */
		ptr_magdata->effect_level = ptr_buf[1];
		ptr_magdata->effect_time  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(0x00, ptr_buf[3], 0x00, 0x00);
		break;

	case 0x4d:	/* ٥Ψ᡼(٥롦ΨѲ2(Ū)) */
		ptr_magdata->effect_level = ptr_buf[1];
		ptr_magdata->effect_time  = 60;
		ptr_magdata->effect_rate  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(0x00, ptr_buf[3], 0x00, 0x00);
		break;

	case 0x4e:	/* ٥Ϲ(٥롦̤͡) */
		ptr_magdata->effect_level = ptr_buf[1];
		ptr_magdata->effect_power = ptr_buf[2];
		break;

	case 0x50:	/* ֥ѥ(̿Ψ̤ѡѲ1(³Ū)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->state_effect = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);
		break;

	case 0x51:	/* ơ徺̤륢ƥ(̿ΨоݤȤʤ륹ơ徺) */
		ptr_magdata->hit_rate    = ptr_buf[1];
		ptr_magdata->fixed_value = ptr_buf[3];
		break;

	case 0x52:	/* Ѳ(åפȤʤ)(̿ΨѲ3(ˡ)Ѳ3(ˡ)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[3], 0x00);
		break;

	case 0x53:	/* ޥåݥȤФΨ᡼(̿ΨΨ̤) */
		ptr_magdata->hit_rate    = ptr_buf[1];
		ptr_magdata->effect_rate = ptr_buf[2];
		break;

	case 0x54:	/* ҥåȥݥȤθʬ᡼Ȥ빶(̿Ψ̤ѡ̤) */
	case 0x6d:	/* ҥåȥݥȤθʬ᡼Ȥ빶(ҥåȥݥȵۼ)(̿Ψ̤ѡ̤) */
		ptr_magdata->hit_rate = ptr_buf[1];
		break;

	case 0x55:	/* 椦(̤ѡ̤ѡѲ1(³Ū)) */
		ptr_magdata->state_effect = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);
		break;

	case 0x56:	/* [̤] Ѳ(ȿž)(̿Ψ̤ѡѲ4) */
		ptr_magdata->hit_rate = ptr_buf[1];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, 0x00, ptr_buf[3]);
		break;

	case 0x57:	/* åפȼѲ(2ξθ)(Ѳ1(³Ū)³֡Ѳ2(Ū)) */
		ptr_magdata->effect_time  = ptr_buf[2];
		ptr_magdata->state_effect = bits_make32(ptr_buf[1], ptr_buf[3], 0x00, 0x00);
		break;

	case 0x58:	/* ޥɥ֥饹(̿Ψ͡³) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->effect_time  = ptr_buf[3];
		break;

	case 0x5b:	/* [̤] $2079 += arg[2] */
		ptr_magdata->fixed_value  = ptr_buf[2];
		break;

	case 0x5c:	/* ۤ褦(̿Ψ̤ѡѲ1(³Ū)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->state_effect = bits_make32(ptr_buf[3], 0x00, 0x00, 0x00);
		break;

	case 0x5d:	/* Ѳ()(̿Ψ̤ѡ̤) */
		ptr_magdata->hit_rate = ptr_buf[1];
		break;

	case 0x5f:	/* μŨˤΤͭʾѲ(å)Ȥʤɸ(ۤ)̵ι(ͭ󥹥ס͡³) */
		ptr_magdata->montype_effect = ptr_buf[1];
		ptr_magdata->effect_power   = ptr_buf[2];
		ptr_magdata->effect_time    = ptr_buf[3];
		break;

	case 0x6b:	/* 󥹥ղáѲղá°ѥå(󥹥סѲ1(³Ū)°) */
		ptr_magdata->montype_effect = ptr_buf[1];
		ptr_magdata->state_effect   = bits_make32(ptr_buf[2], 0x00, 0x00, 0x00);
		ptr_magdata->elem_effect    = ptr_buf[3];
		break;

	case 0x6c:	/* μŨ˥ƥҥåȤ빶(̤ѡͭ͡󥹥) */
		ptr_magdata->effect_power   = ptr_buf[2];
		ptr_magdata->montype_effect = ptr_buf[3];
		break;

	case 0x70:	/* ۡ(̿Ψ̤ѡѲ4(¾)) */
		ptr_magdata->hit_rate     = ptr_buf[1];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, 0x00, ptr_buf[3]);
		break;

	case 0x82:	/* 褲(̤ѡ͡°) */
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->elem_effect  = ptr_buf[3];
		break;

	case 0x83:	/* 󤳤(̤μࡦΥȡ?) */
		ptr_magdata->fixed_value  = ptr_buf[2];
		ptr_magdata->effect_power = ptr_buf[3];
		break;

	case 0x85:	/* (°/ɲø̤ȯưΨ̤͡μ) */
		ptr_magdata->effect_power = ptr_buf[2];
		/*
			- 01: Υ
			- 02: 顼Υ
			- 03: Υ
			- 04: ɥ饴󥭥Υ
			- 06: ѥΥ
			- 08: ɥΥ
			- 09: ⡼륭Υ
			- 05: Υ
			- 07: ɥΥ
			- 0a: °
		*/
		if (ptr_buf[3] == 0x0a)
			ptr_magdata->elem_effect = ptr_buf[1];
		else
			ptr_magdata->hit_rate    = ptr_buf[1];
		break;

	case 0x88:	/* ѲȤʤ°(Ѳ1(³Ū)͡°) */
		ptr_magdata->state_effect  = bits_make32(ptr_buf[1], 0x00, 0x00, 0x00);
		ptr_magdata->effect_power  = ptr_buf[2];
		ptr_magdata->elem_effect   = ptr_buf[3];
		break;

	case 0x89:	/* ѲȤʤ°(Ѳ2(Ū)͡°) */
		ptr_magdata->state_effect  = bits_make32(0x00, ptr_buf[1], 0x00, 0x00);
		ptr_magdata->effect_power  = ptr_buf[2];
		ptr_magdata->elem_effect   = ptr_buf[3];
		break;

	case 0x8a:	/* ѲȤʤ°(Ѳ3(ˡ)͡°) */
		ptr_magdata->state_effect  = bits_make32(0x00, 0x00, ptr_buf[1], 0x00);
		ptr_magdata->effect_power  = ptr_buf[2];
		ptr_magdata->elem_effect   = ptr_buf[3];
		break;

	case 0x8b:	/* ѲȤʤ°(Ѳ4(¾)͡°) */
		ptr_magdata->state_effect  = bits_make32(0x00, 0x00, 0x00, ptr_buf[1]);
		ptr_magdata->effect_power  = ptr_buf[2];
		ptr_magdata->elem_effect   = ptr_buf[3];
		break;

	case 0x8d:	/* ѥ˥å쥤(̤ѡ͡°)*/
		ptr_magdata->effect_power = ptr_buf[2];
		ptr_magdata->elem_effect  = ptr_buf[3];
		break;

	case 0x00:	/* ʤ(̿ΨɲøȯưΨɲøֹ̤) */
	case 0x01:	/* (̤ѡ̤ѡ̤) */
	case 0x02:	/* Ȥ虜(̤ѡ̤ѡ̤) */
	case 0x1d:	/* 饤֥顦٤롦ߤ֤(μ̡̤ѡ̤) */
	case 0x1e:	/* ԡ(̤ѡ̤ѡ̤) */
	case 0x21:	/* ꥿(̤ѡ̤ѡ̤) */
	case 0x22:	/* å(̤ѡ̤ѡ̤) */
	case 0x26:	/* ҥåȥݥȡޥåݥȲ(ѥ᡼μ̡̤ѡ̤) */
	case 0x30:	/* Ǽꡦ(ƥҥåΨ̤ѡ̤) */
	case 0x44:	/* Ȥ(̤ѡ̤ѡ̤) */
	case 0x48:	/* Ȥ館(̤ѡ̤ѡ̤) */
	case 0x4a:	/* ɤ(̤) */
	case 0x59:	/* 礸Τ(̤ѡ̤ѡ̤) */
	case 0x5a:	/* ۥ磻ȥ(̤ѡ̤ѡ̤) */
	case 0x5e:	/* ѥ᡼Ѳ(̤ѡ̤ѡѲоݤȤʤǽ) */
	case 0x60:	/* 󥹥иѹ(̤ѡ̤ѡ̤) */
	case 0x61:	/* ȡ(̤ѡ̤ѡ̤) */
	case 0x62:	/* ӥ֥(̤ѡ̤ѡ̤) */
	case 0x63:	/* ɥ(̤ѡ̤ѡ̤) */
	case 0x65:	/* å(0x800x800x80) */
	case 0x66:	/* åƥ(̤ѡ̤ѡ̤) */
	case 0x67:	/* 󤻤(̤ѡ̤ѡ̤) */
	case 0x68:	/* 夦礦(̤ѡ̤ѡ̤) */
	case 0x69:	/* Ĥ(̤ѡ̤ѡ̤) */
	case 0x6a:	/* 󥹥(̤ѡ̤ѡ̤) */
	case 0x6f:	/* !(̤ѡ̤ѡ̤) */
	case 0x71:	/* ٥(̤ѡ̤ѡ̤) */
	case 0x74:	/* 0x7E */
	case 0x75:	/* 0x7F */
	case 0x7e:	/* ̿椷ʤ(̤ѡ̤ѡ̤) */
	case 0x7f:	/* ʤ(̤ѡ̤ѡ̤) */
	case 0x80:	/* (̤ѡ̤ѡ̤) */
	case 0x81:	/* Ҥä(̤ѡ̤ѡ̤) */
	case 0x84:	/* ۤ(̤ѡ̤ѡ̤)	*/
	case 0x8c:	/* : ޤޤʾѲΤΤ줫Ѳ(̤ѡ̤ѡ̤) */
	default:
		break;
	}

	return 1;
}

int mag_writeparam(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡοͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡοͷϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"MP    Ψ ̿Ψ   ³ ٥  "
				"Ի "
				"  \n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, (ptr_magdata[i].effect_rate != 0) ?
					"%03d(%03X): %-18s :  %5d, %5d,%3d/16, %5d, %5d, %5d, %5d, %5d, %4Xh\n" :
					"%03d(%03X): %-18s :  %5d, %5d, %5d, %5d, %5d, %5d, %5d, %5d, %4Xh\n",
				i, i, mag_getname(i),
				ptr_magdata[i].magic_point,
				ptr_magdata[i].effect_power,
				ptr_magdata[i].effect_rate,
				ptr_magdata[i].hit_rate,
				ptr_magdata[i].effect_time,
				ptr_magdata[i].effect_level,
				ptr_magdata[i].fixed_value,
				ptr_magdata[i].time_wait,
				ptr_magdata[i].effect_type);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeelem(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡ°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡ°ϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"°\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s : ", i, i, mag_getname(i));

		/* ° */
		str_printelem(fp, ptr_magdata[i].elem_effect, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writecategory(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡΥƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡΥƥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"ƥ\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s : ", i, i, mag_getname(i));

		/* ƥ */
		str_printcategory(fp, ptr_magdata[i].category, CATEGORY_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writestate(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡξѲϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"Ѳ1        2                3                4\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s : ", i, i, mag_getname(i));

		/* Ѳ */
		str_printstate(fp, ptr_magdata[i].state_effect, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeflag(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡΥե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡΥե饰ϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"ȿ  "
				"ϰ           "
				"ͭ²\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		/* Ƽե饰 */
		fprintf(fp, "%03d(%03X): %-18s :  %2s   ", i, i, mag_getname(i),
				str_getflag(ptr_magdata[i].flag_reflect));

		/* ϰ */
		mag_printtarget(fp, &ptr_magdata[i]);

		/* ͭ󥹥 */
		if (ptr_magdata[i].montype_effect != 0x00)
		{
			fprintf(fp, "  ");
			str_printmontype(fp, ptr_magdata[i].montype_effect, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

void mag_printtarget(FILE *stream, const magdata_t *ptr_magdata)
{
	const char *str_direction[2][2] =
	{
		{"̣",	"̣/Ũ"},
		{"Ũ",		"Ũ/̣"},
	};
	const char *str_range[2][2] =
	{
		{"ñ",	"ñ/"},
		{"",	"/ñ"},
	};
	const char *str_self[2] =
	{
		"ʬ",
		"Ũ ñ/ư",
	};
	const char *str_roulette = "롼å";

	/* å */
	if (stream == NULL || ptr_magdata == NULL)
	{
		assert(0);
		return;
	}

	/* ˡθϰϥǡϤ */
	if (ptr_magdata->flag_roulette)
		fprintf(stream, "%-17s", str_roulette);
	else if (ptr_magdata->auto_direction)
		fprintf(stream, "%-17s", str_self[ptr_magdata->direction]);
	else
		fprintf(stream, "%-7s %-9s",
				str_direction[ptr_magdata->direction][ptr_magdata->change_direction],
				str_range[ptr_magdata->range][ptr_magdata->change_range]);
}
