/*
 *	ff5misc.c
 *		Copyright (C)2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2007-2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff5decode.h"
#include "ff5misc.h"
#include "ff5magic.h"

#define FILEPOS_TRDATA		0x142548	/* SNES: 0x113210 */
#define FILEPOS_STOREDATA	0x157844	/* SNES: 0x112d40 */
#define FILEPOS_MAGPRICE	0x15741c	/* SNES: 0x112c00 */
#define NUM_TREASURE		337			/* SNES: 252 */
#define NUM_STORE			64
#define NUM_STORENAME		8
#define NUM_MAGPRICE		128
#define OFFSET_MONENC		348			/* SNES: 320 */
#define SIZEOF_TRDATA		8			/* SNES: 4 */
#define SIZEOF_STOREDATA	20			/* SNES: 9 */
#define SIZEOF_MAGPRICE		8			/* SNES: 2 */
#define MSGID_STORENAME		5760

treasure_t *tr_readdata(const mapdata_t *ptr_mapdata)
{
	int i, j;
	treasure_t *buf_trdata = NULL;

	/* ȢγƼǡǼХåեݤ */
	if ((buf_trdata = malloc(sizeof(treasure_t) * NUM_TREASURE)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ȢγƼǡɤ߹ */
	for (i = 0; i < NUM_TREASURE; i++)
	{
		unsigned char buf[SIZEOF_TRDATA];

		file_read(buf, FILEPOS_TRDATA + SIZEOF_TRDATA * i, SIZEOF_TRDATA);

		buf_trdata[i].map.id = 0;
		buf_trdata[i].map.x  = buf[0];
		buf_trdata[i].map.y  = buf[1];

		/* ȢȤμȽ̤ */
		if ((buf[2] & 0x80) != 0)
		{
			buf_trdata[i].monster_no = (buf[2] & 0x3f) + OFFSET_MONENC;
			buf_trdata[i].type       = ((buf[2] & 0x40) != 0) ? TREASURE_TYPE_MAGIC :
			                           TREASURE_TYPE_ITEM;
		}
		else
		{
			buf_trdata[i].monster_no = 0;
			buf_trdata[i].type       = ((buf[2] & 0x40) != 0) ? TREASURE_TYPE_ITEM :
			                           ((buf[2] & 0x20) != 0) ? TREASURE_TYPE_MAGIC :
			                           TREASURE_TYPE_GIL;
		}

		/* ȢȤΥǡɤ߹ */
		switch (buf_trdata[i].type)
		{
		case TREASURE_TYPE_ITEM:
			buf_trdata[i].param.item_no = (buf[5] << 8) | buf[4];
			break;

		case TREASURE_TYPE_MAGIC:
			buf_trdata[i].param.magic_no = (buf[5] << 8) | buf[4];
			break;

		case TREASURE_TYPE_GIL:
			{
				long gil = buf[4];

				for (j = 0; j < (buf[2] & 0x1f); j++)
					gil *= 10;

				buf_trdata[i].param.gil = gil;
			}
			break;

		default:
			assert(0);
		}
	}

	/* Ȣ֤Ƥޥåפɤ߹ */
	for (i = 0; i < NUM_MAP; i++)
	{
		const int index = ptr_mapdata[i].treasure_id;
		int num = ptr_mapdata[i].treasure_num;

		assert(index + num <= NUM_TREASURE);

		if (num > 0 && index + num <= NUM_TREASURE)
		{
			for (; num > 0; num--)
				buf_trdata[index + num - 1].map.id = ptr_mapdata[i].id;
		}
	}

	return buf_trdata;
}

store_t *store_readdata(void)
{
	int i, j;
	store_t *buf_storedata = NULL;

	/* ŹγƼǡǼХåեݤ */
	if ((buf_storedata = malloc(sizeof(store_t) * NUM_STORE)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ŹγƼǡɤ߹ */
	for (i = 0; i < NUM_STORE; i++)
	{
		unsigned char buf[SIZEOF_STOREDATA];

		file_read(buf, FILEPOS_STOREDATA + SIZEOF_STOREDATA * i, SIZEOF_STOREDATA);

		buf_storedata[i].type_no   = buf[0] & 0x7f;
		buf_storedata[i].flag_half = ((buf[0] & 0x80) != 0x00);

		switch (buf_storedata[i].type_no)
		{
			case 0x00:
				buf_storedata[i].type = STORE_TYPE_MAGIC;
				break;

			case 0x01:
			case 0x02:
			case 0x03:
			case 0x04:
			case 0x05:
			case 0x06:
			case 0x07:
				buf_storedata[i].type = STORE_TYPE_ITEM;
				break;

			default:
				assert(0);
		}

		for (j = 0; j < NUM_STORE_ITEM; j++)
		{
			switch (buf_storedata[i].type)
			{
			case STORE_TYPE_ITEM:
				buf_storedata[i].param.item_no[j] = (buf[2 + j * 2 + 1] << 8) | buf[2 + j * 2];
				break;

			case STORE_TYPE_MAGIC:
				buf_storedata[i].param.magic_no[j] = buf[2 + j * 2];
				break;

			default:
				assert(0);
			}
		}
	}

	return buf_storedata;
}

int tr_writedata(const char *ptr_filename, const treasure_t *ptr_trdata,
	const monenc_t *ptr_monenc, const monparty_t *ptr_monparty, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_trdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ȢγƼǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ȢγƼǡϤ */
	fprintf(fp, "No.       ʪ               "
				"  x   y                                      "
				"  Ũ   иŨѡƥ\n");
	for (i = 0; i < NUM_TREASURE; i++)
	{
		fprintf(fp, "%03d(%03X): ", i, i);

		switch (ptr_trdata[i].type)
		{
		case TREASURE_TYPE_ITEM:
			fprintf(fp, "%-16s", item_getname(ptr_trdata[i].param.item_no));
			break;

		case TREASURE_TYPE_MAGIC:
			fprintf(fp, "%-16s", mag_getname(ptr_trdata[i].param.magic_no));
			break;

		case TREASURE_TYPE_GIL:
			{
				char str_gil[256] = "";

				snprintf(str_gil, sizeof(str_gil), "%ld", ptr_trdata[i].param.gil);
				fprintf(fp, "%-16s", str_gil);
			}
			break;

		default:
			assert(0);
		}

		/*  */
		fprintf(fp,	" : %3u,%3u, %-38s",
				ptr_trdata[i].map.x,
				ptr_trdata[i].map.y,
				map_getname(ptr_trdata[i].map.id));

		/* 󥹥 */
		fprintf(fp,	" :  %2s", str_getflag(ptr_trdata[i].monster_no));
		if (ptr_trdata[i].monster_no != 0x00)
		{
			const int monenc_no = ptr_trdata[i].monster_no;
			int monparty_old = 0;

			fprintf(fp, "   No.%03d(%02X):  ", monenc_no, monenc_no);
			for (j = 0; j < 2; j++)
			{
				const int monparty_no = ptr_monenc[monenc_no].party_no[j];
				int index;
				int found = 0;

				if (j != 0 && monparty_no == monparty_old)
					continue;
				else if (j != 0)
					fprintf(fp, " / ");

				for (index = 0; index < 8; index++)
				{
					const int monster_no = ptr_monparty[monparty_no].monster_no[index];
					const int flag_absent = ((ptr_monparty[monparty_no].bits_exist & (1U << (7 - index))) == 0x00);
					int monster_num = 1;

					if (!flag_absent)
					{
						if (found++)
							fprintf(fp, ",  ");

						fprintf(fp, "%-16s", mon_getname(ptr_mondata[monster_no].id));

						for (; index < 7; index++)
						{
							if (monster_no == ptr_monparty[monparty_no].monster_no[index + 1])
								monster_num++;
							else
								break;
						}

						fprintf(fp, " %d", monster_num);
					}
				}

				monparty_old = monparty_no;
			}
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int store_writedata(const char *ptr_filename, const store_t *ptr_storedata, const itemdata_t *ptr_itemdata)
{
	int i, j;
	char tbl_storename[NUM_STORENAME][256];	/* Źμ̾Ǽơ֥ */
	long tbl_magprice[NUM_MAGPRICE];
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_storedata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ŹγƼǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* Źμ̾ǡɤ߹ */
	for (i = 0; i < NUM_STORENAME; i++)
		dec_msg(tbl_storename[i], sizeof(tbl_storename[i]), MSGID_STORENAME + i, FLAG_KANA);

	/* ˡʥǡɤ߹ */
	for (i = 0; i < NUM_MAGPRICE; i++)
	{
		unsigned char buf[SIZEOF_MAGPRICE];

		file_read(buf, FILEPOS_MAGPRICE + SIZEOF_MAGPRICE * i, SIZEOF_MAGPRICE);
		tbl_magprice[i] = bits_make32(buf[0], buf[1], buf[2], buf[3]);
	}

	for (i = 0; i < NUM_STORE; i++)
	{
		if (i != 0)
			fputc('\n', fp);

#		define IS_TOWN_OF_KARNAK_EVENT(v)	((v) == 0x3e || (v) == 0x3f)

		fprintf(fp, "[No.%03d(%03X): %s%s]\n"
					"    ̾\n", i, i,
				(ptr_storedata[i].type_no < NUM_STORENAME) ? tbl_storename[ptr_storedata[i].type_no] : "????",
				IS_TOWN_OF_KARNAK_EVENT(i) ? " (٥)" : (ptr_storedata[i].flag_half) ? " (Ⱦ)" : "");

		for (j = 0; j < NUM_STORE_ITEM; j++)
		{
			long price = 0;
			const char *ptr_name = NULL;

			switch (ptr_storedata[i].type)
			{
			case STORE_TYPE_ITEM:
				if (ptr_storedata[i].param.item_no[j] == 0x00)
					continue;
				price    = ptr_itemdata[ptr_storedata[i].param.item_no[j]].price_buy;
				ptr_name = item_getname(ptr_storedata[i].param.item_no[j]);
				break;

			case STORE_TYPE_MAGIC:
				if (ptr_storedata[i].param.magic_no[j] == 0x00)
					continue;
				price    = (i >= NUM_MAGPRICE) ? 0 : tbl_magprice[ptr_storedata[i].param.magic_no[j]];
				ptr_name = mag_getname(ptr_storedata[i].param.magic_no[j]);
				break;

			default:
				assert(0);
			}

			/* Ⱦ */
			if (ptr_storedata[i].flag_half)
				price /= IS_TOWN_OF_KARNAK_EVENT(i) ? 4 : 2;

			fprintf(fp, "%6ld, %s\n", price, ptr_name);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
