/*
 *	ff5monster.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2007-2009 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff5decode.h"
#include "ff5monster.h"
#include "ff5item.h"
#include "ff5character.h"
#include "ff5magic.h"

#define FILEPOS_MONDATA1		0x14d770	/* SNES: 0x100000 */
#define FILEPOS_MONDATA2		0x14b7f0
#define FILEPOS_MONITEM1		0x148504	/* SNES: 0x105000 */
#define FILEPOS_MONITEM2		0x149504
#define FILEPOS_MONCTRL1		0x14abc4	/* SNES: 0x105600 */
#define FILEPOS_MONCTRL2		0x14b3c4
#define FILEPOS_MONFREE			0x14825c	/* SNES: 0x108600 */
#define FILEPOS_MONTECH1		0x14728c	/* SNES: 0x109900 */
#define FILEPOS_MONTECH2		0x147a8c
#define FILEPOS_MONACT_ADDR1	0x2418b4	/* SNES: 0x109c00 */
#define FILEPOS_MONACT_ADDR2	0x2420b4
#define FILEPOS_MONACT			0x24240c	/* SNES: 0x109f00 */
#define FILEPOS_MONACT_EOF		0x246972	/* SNES: 0x10c7ff */
#define FILEPOS_MONPARTY		0x14fff0	/* SNES: 0x103000 */
#define FILEPOS_MONPARTYPOS		0x1545d4	/* SNES: 0x108900 */
#define FILEPOS_MONENCRND		0x141500	/* SNES: 0x106800 */
#define FILEPOS_MONENCFIXED		0x13a8bc	/* SNES: 0x107800 */
#define FILEPOS_MONENCALT		0x132570	/* SNES: 0x10ffe0 */
#define FILEPOS_MONMAP			0x13aafc	/* SNES: 0x107a00 */
#define NUM_MON1				288			/* ̾󥹥 */	/* SNES: 256 */
#define NUM_MON2				224			/* ܥ󥹥 */	/* SNES: 128 */
#define NUM_MON					(NUM_MON1 + NUM_MON2)				/* = 512 */
#define NUM_MONNAME				407
#define NUM_MONTECH				64
#define NUM_MONMES				238			/* SNES: 234 */
#define NUM_MONPARTY			639			/* SNES: 512 */
#define NUM_MONENCRND			284			/* SNES: 256 */
#define NUM_MONENCFIXED			144
#define NUM_MONENC				(NUM_MONENCRND + NUM_MONENCFIXED)	/* = 428 */
#define NUM_MONMAP				3
#define NUM_MONMAPAREA			8
#define SIZEOF_MONDATA			36			/* SNES: 32 */
#define SIZEOF_MONITEM			8			/* SNES: 4 */
#define SIZEOF_MONCTRL			4
#define SIZEOF_MONFREE			1
#define SIZEOF_MONTECH			4			/* SNES: 2 */
#define SIZEOF_MONPARTY			28			/* SNES: 16 */
#define SIZEOF_MONPARTYPOS		8
#define SIZEOF_MONENCRND		8
#define SIZEOF_MONENCFIXED		4
#define SIZEOF_MONMAP			8
#define SIZEOF_MONACT_ADDR		4			/* SNES: 2 */
#define SIZEOF_MONACT_COND		4
#define SIZEOF_MONACT_TBL		(FILEPOS_MONACT_EOF - FILEPOS_MONACT + 1)
#define MSGID_MONNAME			5018
#define MSGID_MONMES			2850
#define MSGID_MONTECHNAME		2641

static void mon_setname(void);
static void mon_settechname(void);
static void mon_setmes(void);
static const char *mon_gettarget(int index);
static const char *mon_gettechname(int index);
static const char *mon_getmes(int index);
static void mon_printact(FILE *stream, int index, const mondata_t *mondata);

static char tbl_monname[NUM_MONNAME][256];		/* 󥹥̾Ǽơ֥ */
static char tbl_montechname[NUM_MONTECH][256];	/* ֤Ȥ虜פ̾Ǽơ֥ */
static char tbl_monmes[NUM_MONMES][256];		/* 󥹥ΥåǼơ֥ */
static int flag_setmonname     = 0;				/* tbl_monname˥ǡǼƤ뤫ɤ */
static int flag_setmontechname = 0;				/* tbl_montechname˥ǡǼƤ뤫ɤ */
static int flag_setmonmes      = 0;				/* tbl_monmes˥ǡǼƤ뤫ɤ */

const char *mon_getname(int index)
{
	/* å */
	if (index < 0 || NUM_MONNAME <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmonname)
		mon_setname();

	return tbl_monname[index];
}

static const char *mon_gettechname(int index)
{
	/* å */
	if (index < 0 || NUM_MONTECH <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmontechname)
		mon_settechname();

	return tbl_montechname[index];
}

static const char *mon_getmes(int index)
{
	/* å */
	if (index < 0 || NUM_MONMES <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmonmes)
		mon_setmes();

	return tbl_monmes[index];
}

static const char *mon_gettarget(int index)
{
	static const char *str_target[] =
	{
		"Хå", "", "", "եꥹ",
		"", "󥹥1", "󥹥2", "󥹥3",
		"󥹥4", "󥹥5", "󥹥6", "󥹥7",
		"󥹥8", "󥹥ʬ", "ʬʳΥ󥹥", "󥹥",
		"ʬʳΥ󥹥ï", "󥹥ï", "Υ饯", "Υ饯",
		"Υ饯ï", "Υ饯ï", "饯", "饯",
		"饯ï", "饯ï", "ƮǽΥ饯", "ƮǽΥ饯ï",
		"ե쥯֤Υ󥹥", "ե쥯֤Υ󥹥ï", "λ֤Υ󥹥", "λ֤Υ󥹥ï",
		"HP50%̤Υ󥹥", "HP50%̤Υ󥹥ï", "ե쥯֤Υ饯ï", "饯",
		"饯ï", "ƮǽΥ󥹥", "ƮǽΥ󥹥ï", "27h",
		"ƾ֤ΥХå", "ƾ֤Υ", "ƾ֤Υ", "ƾ֤Υեꥹ",
		"ƾ֤Υ", "Ǹ˹ư饯/Ũï", "󤻤饯", "ƮǽΥХå",
		"ƮǽΥ", "ƮǽΥ", "ƮǽΥեꥹ", "ƮǽΥ",
	};

	/* å */
	if (index < 0 || (int)(sizeof(str_target) / sizeof(str_target[0])) <= index)
	{
		assert(0);
		return "";
	}

	return str_target[index];
}

static void mon_setname(void)
{
	int i;

	/* 󥹥̾ǡ */
	for (i = 0; i < NUM_MONNAME; i++)
		*tbl_monname[i] = '\0';

	/* 󥹥̾ǡɤ߹ */
	for (i = 0; i < NUM_MONNAME; i++)
		dec_msg(tbl_monname[i], sizeof(tbl_monname[i]), MSGID_MONNAME + i, FLAG_KANA);

	flag_setmonname = 1;
}

static void mon_settechname(void)
{
	int i;

	/* 󥹥Ρ֤Ȥ虜פ̾ǡ */
	for (i = 0; i < NUM_MONTECH; i++)
		*tbl_montechname[i] = '\0';

	/* 󥹥Ρ֤Ȥ虜פ̾ǡɤ߹ */
	for (i = 0; i < NUM_MONTECH; i++)
		dec_msg(tbl_montechname[i], sizeof(tbl_montechname[i]), MSGID_MONTECHNAME + i, FLAG_KANA);

	flag_setmontechname = 1;
}

static void mon_setmes(void)
{
	int i;
	char *ptr;

	/* 󥹥Υåǡ */
	for (i = 0; i < NUM_MONMES; i++)
		*tbl_monmes[i] = '\0';

	/* 󥹥Υåǡɤ߹ */
	for (i = 0; i < NUM_MONMES; i++)
	{
		dec_msg(tbl_monmes[i], sizeof(tbl_monmes[i]), MSGID_MONMES + i, FLAG_KANJI);

		/* Ԥ򵭹֤ */
		while ((ptr = strchr(tbl_monmes[i], '\n')) != NULL)
			*ptr = '/';
	}

	flag_setmonmes = 1;
}

mondata_t *mon_readdata(void)
{
	int i, j;
	mondata_t *buf_mondata = NULL;

	/* 󥹥γƼǡǼХåեݤ */
	if ((buf_mondata = malloc(sizeof(mondata_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥νƥǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONITEM];

		if (i >= NUM_MON1)
			file_read(buf, FILEPOS_MONITEM2 + SIZEOF_MONITEM * (i - NUM_MON1), SIZEOF_MONITEM);
		else
			file_read(buf, FILEPOS_MONITEM1 + SIZEOF_MONITEM * i, SIZEOF_MONITEM);

		for (j = 0; j < 4; j++)
			buf_mondata[i].item_no[j] = (buf[j * 2 + 1] << 8) | buf[j * 2];
	}

	/* 󥹥Ρ֤Ĥ׹ưǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONCTRL];

		if (i >= NUM_MON1)
			file_read(buf, FILEPOS_MONCTRL2 + SIZEOF_MONCTRL * (i - NUM_MON1), SIZEOF_MONCTRL);
		else
			file_read(buf, FILEPOS_MONCTRL1 + SIZEOF_MONCTRL * i, SIZEOF_MONCTRL);

		for (j = 0; j < SIZEOF_MONCTRL; j++)
			buf_mondata[i].control_no[j] = buf[j];
	}

	/* 󥹥Ρ֤Ϥʤġ׹ưǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i >= NUM_MON1)
			buf_mondata[i].free_no = 0xff;
		else
		{
			unsigned char buf[SIZEOF_MONFREE];

			file_read(buf, FILEPOS_MONFREE + SIZEOF_MONFREE * i, SIZEOF_MONFREE);
			buf_mondata[i].free_no = buf[0];
		}
	}

	/* 󥹥Ρ֤Ȥ虜׹ưǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONTECH];

		if (i >= NUM_MON1)
			file_read(buf, FILEPOS_MONTECH2 + SIZEOF_MONTECH * (i - NUM_MON1), SIZEOF_MONTECH);
		else
			file_read(buf, FILEPOS_MONTECH1 + SIZEOF_MONTECH * i, SIZEOF_MONTECH);

		buf_mondata[i].flag_tech_attack_up    = ((buf[0] & 0x01) != 0x00);
		buf_mondata[i].flag_tech_defence_down = ((buf[0] & 0x80) != 0x00);
		buf_mondata[i].flag_tech_become_old   = ((buf[0] & 0x40) != 0x00);
		buf_mondata[i].flag_tech_poison       = ((buf[0] & 0x20) != 0x00);
		buf_mondata[i].flag_tech_darkness     = ((buf[0] & 0x10) != 0x00);
		buf_mondata[i].flag_tech_paralyze     = ((buf[0] & 0x08) != 0x00);
		buf_mondata[i].flag_tech_confuse      = ((buf[0] & 0x04) != 0x00);
		buf_mondata[i].flag_tech_slip         = ((buf[0] & 0x02) != 0x00);

		buf_mondata[i].tech_no = buf[1];
	}

	/* 󥹥γƼǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONDATA];

		if (i >= NUM_MON1)
			file_read(buf, FILEPOS_MONDATA2 + SIZEOF_MONDATA * (i - NUM_MON1), SIZEOF_MONDATA);
		else
			file_read(buf, FILEPOS_MONDATA1 + SIZEOF_MONDATA * i, SIZEOF_MONDATA);

		/* ͷϥǡ */
		buf_mondata[i].level               = buf[32];
		buf_mondata[i].hit_point           = (buf[ 9] << 8) | buf[ 8];
		buf_mondata[i].magic_point         = (buf[11] << 8) | buf[10];
		buf_mondata[i].experience_point    = (buf[13] << 8) | buf[12];
		buf_mondata[i].gil                 = (buf[15] << 8) | buf[14];
		buf_mondata[i].speed               = buf[ 0];
		buf_mondata[i].attack_power        = buf[ 1];
		buf_mondata[i].defense_power       = buf[ 4];
		buf_mondata[i].evade_rate          = buf[ 3];
		buf_mondata[i].magic_attack_power  = buf[ 5];
		buf_mondata[i].magic_defense_power = buf[ 6];
		buf_mondata[i].magic_evade_rate    = buf[ 7];
		buf_mondata[i].weapon_no           = buf[16];
		buf_mondata[i].multiplier          = buf[ 2];
		buf_mondata[i].id                  = (buf[31] << 8) | buf[30];

		/* ѥ᡼黻ФǤǡ */
		buf_mondata[i].stamina             = buf_mondata[i].level;
		buf_mondata[i].magic_multiplier    = MIN(buf_mondata[i].level, 99) * MIN(buf_mondata[i].magic_attack_power, 99) / 256 + 4;

		/* °ϥǡ */
		buf_mondata[i].elem_none   = buf[17];
		buf_mondata[i].elem_absorb = buf[21];
		buf_mondata[i].elem_double = buf[23];

		/* ƥǡ */
		buf_mondata[i].category_weak = buf[22];

		/* Ѳϥǡ */
		buf_mondata[i].state_defense = bits_make32(buf[18], buf[19], buf[20], 0x00);
		buf_mondata[i].state_apriori = bits_make32(buf[26], buf[27], buf[28], buf[29]);
		buf_mondata[i].montype       = buf[24];

		/* ե饰ϥǡ */
		buf_mondata[i].flag_capture = ((buf[25] & 0x08) == 0x00);
		buf_mondata[i].flag_control = ((buf[25] & 0x10) == 0x00);
		buf_mondata[i].flag_hide_hp = ((buf[25] & 0x80) != 0x00);
	}

	return buf_mondata;
}

monact_t *mon_readact(void)
{
	int i, j;
	monact_t *buf_monact = NULL;
	int tbl_offset[NUM_MON];
	unsigned char tbl_monact[SIZEOF_MONACT_TBL];
	const int num_monact_separator = 2;

	/* 󥹥ιưѥǡǼХåեݤ */
	if ((buf_monact = malloc(sizeof(monact_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥ιưѥǡ */
	for (i = 0; i < NUM_MON; i++)
	{
		buf_monact[i].ptn_num = 0;
		for (j = 0; j < (int)(sizeof(buf_monact[i].ptn) / sizeof(buf_monact[i].ptn[0])); j++)
		{
			buf_monact[i].ptn[j].condition_size = 0;
			buf_monact[i].ptn[j].action_size = 0;
			buf_monact[i].ptn[j].condition = NULL;
			buf_monact[i].ptn[j].action = NULL;
		}
	}

	/* 󥹥ιưѥǡǼƤ륢ɥ쥹(ؤΥեå)ɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONACT_ADDR];

		if (i >= NUM_MON1)
			file_read(buf, FILEPOS_MONACT_ADDR2 + SIZEOF_MONACT_ADDR * (i - NUM_MON1), SIZEOF_MONACT_ADDR);
		else
			file_read(buf, FILEPOS_MONACT_ADDR1 + SIZEOF_MONACT_ADDR * i, SIZEOF_MONACT_ADDR);

		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}

	/* 󥹥ιưѥǡɤ߹ */
	file_read(tbl_monact, FILEPOS_MONACT, SIZEOF_MONACT_TBL);

	for (i = 0; i < NUM_MON; i++)
	{
		int separator_num = 0;
		int ptn_num = 0;
		int flag_condition = 1;
		int index = 0, index_mark = 0;

		for (index = index_mark = tbl_offset[i]; index < SIZEOF_MONACT_TBL; index++)
		{
			if (tbl_monact[index] == 0xfe || tbl_monact[index] == 0xff)
			{
				const size_t len_read = index - index_mark;

				/* ŬѾξά */
				if (len_read == 0)
				{
					assert(tbl_monact[index] == 0xff);
					if (++separator_num >= num_monact_separator)
						break;
				}
				else
				{
					if (flag_condition)
					{
						/* ʹưѥǡγǧ */
						if (tbl_monact[index] != 0xfe)
							break;

						assert(tbl_monact[index] == 0xfe);
						assert(len_read % SIZEOF_MONACT_COND == 0);

						if ((buf_monact[i].ptn[ptn_num].condition = malloc(len_read)) == NULL)
							err_printf(STR_FATAL_ERROR_IN_MALLOC);
						memcpy(buf_monact[i].ptn[ptn_num].condition, &tbl_monact[index_mark], len_read);
						buf_monact[i].ptn[ptn_num].condition_size = len_read;
					}
					else
					{
						assert(ptn_num < (int)(sizeof(buf_monact[i].ptn) / sizeof(buf_monact[i].ptn[0])));

						if ((buf_monact[i].ptn[ptn_num].action = malloc(len_read)) == NULL)
							err_printf(STR_FATAL_ERROR_IN_MALLOC);
						memcpy(buf_monact[i].ptn[ptn_num].action, &tbl_monact[index_mark], len_read);
						buf_monact[i].ptn[ptn_num].action_size = len_read;
						ptn_num++;

						/* ŬѾνλ */
						if (tbl_monact[index] == 0xff)
							if (++separator_num >= num_monact_separator)
								break;
					}

					flag_condition = !flag_condition;
				}

				index_mark = index + 1;
			}
		}

		/* ʹưѥǡγǧ */
		/* assert(index < SIZEOF_MONACT_TBL); */

		buf_monact[i].ptn_num = ptn_num;
	}

	return buf_monact;
}

monparty_t *mon_readparty(void)
{
	int i, j;
	monparty_t *buf_monparty = NULL;

	/* 󥹥ΥѡƥǡǼХåեݤ */
	if ((buf_monparty = malloc(sizeof(monparty_t) * NUM_MONPARTY)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥ΥѡƥγƼǡɤ߹ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY];

		file_read(buf, FILEPOS_MONPARTY + SIZEOF_MONPARTY * i, SIZEOF_MONPARTY);

		/* ͷϥǡ */
		buf_monparty[i].bits_vfx      = buf[ 0];
		buf_monparty[i].ability_point = buf[ 2];
		buf_monparty[i].escape_rate   = buf[ 1] & 0x7f;
		buf_monparty[i].bits_exist    = buf[ 3];
		buf_monparty[i].music_no      = (buf[22] & 0x7f) >> 3;

		/* ե饰ϥǡ */
		buf_monparty[i].flag_escape      = ((buf[ 1] & 0x80) == 0x00);
		buf_monparty[i].flag_boss_effect = ((buf[23] & 0x01) != 0x00);
		buf_monparty[i].flag_last_effect = ((buf[23] & 0x02) != 0x00);
		buf_monparty[i].flag_magic       = ((buf[23] & 0x04) == 0x00);
		buf_monparty[i].flag_xleave      = ((buf[23] & 0x10) != 0x00);
		buf_monparty[i].flag_boss        = ((buf[23] & 0x20) != 0x00);
		buf_monparty[i].flag_mute        = ((buf[23] & 0x40) == 0x00);
		buf_monparty[i].flag_raid        = ((buf[23] & 0x80) != 0x00);

		for (j = 0; j < 8; j++)
		{
			buf_monparty[i].monster_no[j] = (buf[j * 2 + 5] << 8) | buf[j * 2 + 4];

			if (buf_monparty[i].flag_boss && buf_monparty[i].monster_no[j] != 0xff)
			{
				assert(buf_monparty[i].monster_no[j] + NUM_MON1 < NUM_MON);
				buf_monparty[i].monster_no[j] += NUM_MON1;
			}
		}
	}

	/* 󥹥Υѡƥ֥ǡɤ߹ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTYPOS];
		int position_front = 1;

		file_read(buf, FILEPOS_MONPARTYPOS + SIZEOF_MONPARTYPOS * i, SIZEOF_MONPARTYPOS);

		buf_monparty[i].bits_backrow = 0x00;
		for (j = 0; j < 8; j++)
		{
			buf_monparty[i].position_x[j] = buf[j] >> 4;
			buf_monparty[i].position_y[j] = buf[j] & 0x0f;

			if (position_front < buf_monparty[i].position_x[j])
				position_front = buf_monparty[i].position_x[j];
		}

		/* XXX: 󥹥ˤ뤫Ƚꤹ() */
		for (j = 0; j < 8; j++)
		{
			if (position_front - buf_monparty[i].position_x[j] >= 5)
				buf_monparty[i].bits_backrow |= 1U << (7 - j);
		}
	}

	return buf_monparty;
}

monenc_t *mon_readenc(void)
{
	int i, j;
	monenc_t *buf_monenc = NULL;

	/* 󥹥ơ֥ǡǼХåեݤ */
	if ((buf_monenc = malloc(sizeof(monenc_t) * NUM_MONENC)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ̾󥹥Υ󥹥ơ֥ǡɤ߹ */
	for (i = 0; i < NUM_MONENCRND; i++)
	{
		unsigned char buf[SIZEOF_MONENCRND];

		file_read(buf, FILEPOS_MONENCRND + SIZEOF_MONENCRND * i, SIZEOF_MONENCRND);
		for (j = 0; j < 4; j++)
			buf_monenc[i].party_no[j] = (buf[j * 2 + 1] << 8) | buf[j * 2];
	}

	/* 󥹥Υ󥹥ơ֥ǡɤ߹ */
	for (i = NUM_MONENCRND; i < NUM_MONENC; i++)
	{
		unsigned char buf[SIZEOF_MONENCFIXED];

		file_read(buf, FILEPOS_MONENCFIXED + SIZEOF_MONENCFIXED * (i - NUM_MONENCRND), SIZEOF_MONENCFIXED);
		for (j = 0; j < 2; j++)
			buf_monenc[i].party_no[j] = (buf[j * 2 + 1] << 8) | buf[j * 2];
	}

	return buf_monenc;
}

monmap_t *mon_readmap(void)
{
	int i, j, x, y;
	long filepos;
	monmap_t *buf_monmap = NULL;

	/* 󥹥ޥåץǡǼХåեݤ */
	if ((buf_monmap = malloc(sizeof(monmap_t) * NUM_MONMAP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥ޥåץǡɤ߹ */
	filepos = FILEPOS_MONMAP;
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < NUM_MONMAPAREA; y++)
		{
			for (x = 0; x < NUM_MONMAPAREA; x++)
			{
				unsigned char buf[SIZEOF_MONMAP];

				filepos += file_read(buf, filepos, SIZEOF_MONMAP);
				for (j = 0; j < 4; j++)
					buf_monmap[i].area[x][y].enc_no[j] = (buf[j * 2 + 1] << 8) | buf[j * 2];
			}
		}
	}

	return buf_monmap;
}

int mon_writeparam(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥οͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥οͷϥǡϤ */
	fprintf(fp, "No.       ̾                "
				" LV     HP     MP    EXP    "
				"  ®          ⹶   Ψ    ID\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :  %3d, %5ld, %5ld, %5ld, %5ld, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d\n", i, i,
				mon_getname(ptr_mondata[i].id),
				ptr_mondata[i].level,
				ptr_mondata[i].hit_point,
				ptr_mondata[i].magic_point,
				ptr_mondata[i].experience_point,
				ptr_mondata[i].gil,
				ptr_mondata[i].speed,
				ptr_mondata[i].attack_power,
				ptr_mondata[i].defense_power,
				ptr_mondata[i].evade_rate,
				ptr_mondata[i].magic_attack_power,
				ptr_mondata[i].magic_defense_power,
				ptr_mondata[i].magic_evade_rate,
				ptr_mondata[i].multiplier,
				ptr_mondata[i].magic_multiplier,
				ptr_mondata[i].id);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeelem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥°ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"̵°          "
				"ۼ°          "
				"°\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(ptr_mondata[i].id));

		/* ̵° */
		str_printelem(fp, ptr_mondata[i].elem_none, ELEM_ALL);
		fprintf(fp, "  ");

		/* ۼ° */
		str_printelem(fp, ptr_mondata[i].elem_absorb, ELEM_ALL);
		fprintf(fp, "  ");

		/* ° */
		str_printelem(fp, ptr_mondata[i].elem_double, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writestate(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ξѲϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"Ѳ1    2                3                 "
				"Ѳ1    2                3                4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(ptr_mondata[i].id));

		/* Ѳؤ */
		str_printstate(fp, ptr_mondata[i].state_defense, STATE_ALL & ~STATE_OTHER);
		fprintf(fp, "  ");

		/* ǽ餫ŬѤѲ */
		str_printstate(fp, ptr_mondata[i].state_apriori, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeflag(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Υե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Υե饰ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"  HP?? ²\n");
	for (i = 0; i < NUM_MON; i++)
	{
		/* Ƽե饰 */
		fprintf(fp, "%03d(%03X): %-16s :  %2s   %2s   %2s", i, i, mon_getname(ptr_mondata[i].id),
				str_getflag(ptr_mondata[i].flag_capture),
				str_getflag(ptr_mondata[i].flag_control),
				str_getflag(ptr_mondata[i].flag_hide_hp));

		/* 󥹥 */
		if ((ptr_mondata[i].montype & MONTYPE_ALL) != 0x00)
		{
			fprintf(fp, "  ");
			str_printmontype(fp, ptr_mondata[i].montype, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeitem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥νƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥νƥǡϤ */
	fprintf(fp, "No.       ̾               "
				"(Ψ)        "
				"(Ψ)        "
				"Ȥ              "
				"Ȥ(100%%)\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
			ptr_name[j] = (ptr_mondata[i].item_no[j] != 0x00) ? item_getname(ptr_mondata[i].item_no[j]) : STR_NOTHING_18;

		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(ptr_mondata[i].id));
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-18s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writefree(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Ρ֤Ϥʤġ׹ưǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Ρ֤Ϥʤġ׹ưǡϤ */
	fprintf(fp, "No.       ̾                 Ϥʤ\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name;

		if (ptr_mondata[i].free_no == 0xff)	/* 0xff = ʤ */
			ptr_name = STR_NOTHING_16;
		else
			ptr_name = mag_getname(ptr_mondata[i].free_no);

		fprintf(fp, "%03d(%03X): %-16s :  %2s   %-16s\n", i, i,
				mon_getname(ptr_mondata[i].id),
				str_getflag(ptr_mondata[i].flag_capture),
				ptr_name);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writectrl(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Ρ֤Ĥ׹ưǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Ρ֤Ĥ׹ưǡϤ */
	fprintf(fp, "No.       ̾               "
				"  "
				"Ĥ1         "
				"Ĥ2         "
				"Ĥ3         "
				"Ĥ4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
		{
			if (ptr_mondata[i].control_no[j] == 0x81)		/* 0x81 = "Ȥ虜" */
				ptr_name[j] = mon_gettechname(ptr_mondata[i].tech_no);
			else if (ptr_mondata[i].control_no[j] == 0xff)	/* 0xff = ʤ */
				ptr_name[j] = STR_NOTHING_16;
			else
				ptr_name[j] = mag_getname(ptr_mondata[i].control_no[j]);
		}

		fprintf(fp, "%03d(%03X): %-16s :  %2s   ", i, i,
				mon_getname(ptr_mondata[i].id),
				str_getflag(ptr_mondata[i].flag_control));
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writetech(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Ρ֤Ȥ虜ץǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Ρ֤Ȥ虜ץǡϤ */
	fprintf(fp, "No.       "
				"̾               "
				"Ȥ虜          "
				"徺 ɸ̵ Ϸ  Ű   å\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : %-18s,    %2s       %2s     %2s   %2s   %2s   %2s   %2s   %2s\n", i, i,
				mon_getname(ptr_mondata[i].id),
				mon_gettechname(ptr_mondata[i].tech_no),
				str_getflag(ptr_mondata[i].flag_tech_attack_up),
				str_getflag(ptr_mondata[i].flag_tech_defence_down),
				str_getflag(ptr_mondata[i].flag_tech_become_old),
				str_getflag(ptr_mondata[i].flag_tech_poison),
				str_getflag(ptr_mondata[i].flag_tech_darkness),
				str_getflag(ptr_mondata[i].flag_tech_paralyze),
				str_getflag(ptr_mondata[i].flag_tech_confuse),
				str_getflag(ptr_mondata[i].flag_tech_slip));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeweak(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥μƥΥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥μƥΥǡϤ */
	fprintf(fp, "No.       ̾               "
				"ƥ\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(ptr_mondata[i].id));

		/* ƥ */
		str_printcategory(fp, ptr_mondata[i].category_weak, CATEGORY_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemisc(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Τ¾ΥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Τ¾ΥǡϤ */
	fprintf(fp, "No.       "
				"̾               "
				"̾ﹶ\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : %-18s\n", i, i,
				mon_getname(ptr_mondata[i].id),
				item_getname((ptr_mondata[i].weapon_no != 0x00) ? ptr_mondata[i].weapon_no : 1));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeact(const char *ptr_filename, const monact_t *ptr_monact, const mondata_t *ptr_mondata)
{
	int i, j, k;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monact == NULL || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ιưѥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ιưѥǡϤ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X): %s]\n", i, i, mon_getname(ptr_mondata[i].id));

		for (j = 0; j < ptr_monact[i].ptn_num; j++)
		{
			int turn = 0;
			int turn_old = 0;
			int index = 0;
			int letter_num = 0;
			const size_t bufsize = ptr_monact[i].ptn[j].action_size;
			const unsigned char *ptr_condition = ptr_monact[i].ptn[j].condition;
			const unsigned char *ptr_action = ptr_monact[i].ptn[j].action;
			int *buf_turn = NULL;

			/* ѤΰХåեݤ */
			if ((buf_turn = malloc(sizeof(int) * bufsize)) == NULL)
				err_printf(STR_FATAL_ERROR_IN_MALLOC);

			for (k = 0; k < (int)bufsize; k++)
				buf_turn[k] = 0;

			/* ŬѾ */
			fprintf(fp, ": ");
			for (k = 0; k < (int)ptr_monact[i].ptn[j].condition_size / SIZEOF_MONACT_COND; k++,
				ptr_condition += SIZEOF_MONACT_COND)
			{
				if (k != 0)
					fprintf(fp, " & ");

				if (ptr_condition[0] == 0x00)
				{
					/* ̾(00 00 00 00) */
					fprintf(fp, "[̾]");
				}
				else if (ptr_condition[0] == 0x01)
				{
					/* ֤ξ֤Ȱפ(01 xx yy zz) */
					/*   xx: о */
					/*   yy: ѲӥåβХܤѤ뤫(04) */
					/*   zz: Ѳӥå */
					char str[256] = "";

					if (ptr_condition[2] == 4)
						str_setstate_f(str, (state_t)ptr_condition[3], sizeof(str));
					else
						str_setstate(str, (state_t)bits_make32(
							(ptr_condition[2] == 0) ? ptr_condition[3] : 0x00,
							(ptr_condition[2] == 1) ? ptr_condition[3] : 0x00,
							(ptr_condition[2] == 2) ? ptr_condition[3] : 0x00,
							(ptr_condition[2] == 3) ? ptr_condition[3] : 0x00), sizeof(str));

					fprintf(fp, "[%sξ֤ \"%s\" Ǥ]", mon_gettarget(ptr_condition[1]), str);
				}
				else if (ptr_condition[0] == 0x02)
				{
					/* ҥåȥݥȤ̤(02 xx yy zz) */
					/*   xx: о */
					/*   yy: ҥåȥݥ(̥Х) */
					/*   zz: ҥåȥݥ(̥Х) */
					fprintf(fp, "[%sHP%d̤Ǥ]", mon_gettarget(ptr_condition[1]), (ptr_condition[3] << 8) | ptr_condition[2]);
				}
				else if (ptr_condition[0] == 0x03)
				{
					/* ѿͤȰפ(03 00 yy zz) */
					/*   yy: ѿֹ */
					/*   zz: ѿ */
					fprintf(fp, "[ѿ%d \"%d\" Ǥ]", ptr_condition[2], ptr_condition[3]);
				}
				else if (ptr_condition[0] == 0x04)
				{
					/* Υ󥹥¸ߤʤ(04 00 yy 00) */
					/*   yy: 󥹥 */
					if (ptr_condition[2] == 0x00)
						fprintf(fp, "[ʬʳΥ󥹥ʤ]");
					else
						fprintf(fp, "[ʬʳΥ롼פʤ]");
				}
				else if (ptr_condition[0] == 0x05)
				{
					/* ֤˥󥹥¸ߤ(05 xx yy zz) */
					/*   xx: ե饰(0: ƤϤޤ, 1: ƤϤޤʤ) */
					/*   yy: ̤ (00 or F0) */
					/*   zz: (ӥå) */
					int index;
					char str_no[64] = "";

					for (index = 0; index < 8; index++)
					{
						if ((ptr_condition[3] & (1U << (7 - index))) != 0x00)
							snprintf(str_no, sizeof(str_no), "%s%s%d", str_no, (strlen(str_no) > 0) ? ", " : "", index + 1);
					}
					fprintf(fp, "[%sܤΥ󥹥и%s]", str_no, (ptr_condition[1] == 0x00) ? "" : "ʤ");
				}
				else if (ptr_condition[0] == 0x06 || ptr_condition[0] == 0x07)
				{
					/* ưФ륫󥿡(06 xx yy zz) */
					/* ưФ륫󥿡(07 xx yy zz) */
					/*   xx: ե饰(0: ƤϤޤ, 1: ƤϤޤʤ) */
					/*   yy: ư */
					/*   zz: °/ƥӥå */
					fprintf(fp, "[");

					switch (ptr_condition[2])
					{
					case 0x07:
						fprintf(fp, "ʤ餫ιư");
						break;
					case 0x2b:
						fprintf(fp, "ˡ");
						break;
					default:
						fprintf(fp, "ޥ \"%s\" ", chr_getcmdtypename(ptr_condition[2]));
					}

					if (ptr_condition[3] != 0x00)
					{
						char str[64] = "";

						if (ptr_condition[0] == 0x06)
							fprintf(fp, "(%s°)", str_setelem(str, (elem_t)ptr_condition[3], sizeof(str)));
						else
							fprintf(fp, "(ƥ: %s)", str_setcategory(str, (category_t)ptr_condition[3], sizeof(str)));
					}

					fprintf(fp, "%s]", (ptr_condition[1] != 0x00) ? "ʳ" : "");
				}
				else if (ptr_condition[0] == 0x08)
				{
					/* ˡФ륫󥿡(08 xx yy 00) */
					/*   xx: ե饰(0: ƤϤޤ, 1: ƤϤޤʤ) */
					/*   yy: ˡֹ */
					fprintf(fp, "[ˡ \"%s\" %s]", mag_getname(ptr_condition[2]), (ptr_condition[1] != 0x00) ? "ʳ" : "");
				}
				else if (ptr_condition[0] == 0x09)
				{
					/* ƥ५󥿡(09 xx yy 00) */
					/*   xx: ե饰(0: ƤϤޤ, 1: ƤϤޤʤ) */
					/*   yy: ƥֹ */
					fprintf(fp, "[ƥ \"%s\" %s]", item_getname(ptr_condition[2]), (ptr_condition[1] != 0x00) ? "ʳ" : "");
				}
				else if (ptr_condition[0] == 0x0a)
				{
					/* оФ륫󥿡(0A 00 00 zz) */
					/*   zz: ե饰(0: ñ, 1: ) */
					fprintf(fp, "[%s]", (ptr_condition[3] != 0x00) ? "" : "ñ");
				}
				else if (ptr_condition[0] == 0x0b)
				{
					/* ѿͤͤȰפ(0B xx yy zz) */
					/*   xx: о */
					/*   yy: ɥ쥹 (SNES: $7E2000, GBA: $0201EDFC) Υեå */
					/*   zz:  */
					if ((ptr_condition[2] >= 0x14 && ptr_condition[2] <= 0x18) ||
						(ptr_condition[2] >= 0x6c && ptr_condition[2] <= 0x6f))
					{
						const int index = (ptr_condition[2] >= 0x6c) ? 0x6c : 0x14;
						char str[256] = "";

						if (ptr_condition[2] - index == 4)
							str_setstate_f(str, (state_t)ptr_condition[3], sizeof(str));
						else
							str_setstate(str, (state_t)bits_make32(
								(ptr_condition[2] - index == 0) ? ptr_condition[3] : 0x00,
								(ptr_condition[2] - index == 1) ? ptr_condition[3] : 0x00,
								(ptr_condition[2] - index == 2) ? ptr_condition[3] : 0x00,
								(ptr_condition[2] - index == 3) ? ptr_condition[3] : 0x00), sizeof(str));

						fprintf(fp, "[%sѿ \"%s%d, %s\" Ǥ]", mon_gettarget(ptr_condition[1]),
								(index == 0x6c) ? "³" : "", ptr_condition[2] - index + 1, str);
					}
					else
					{
						fprintf(fp, "[%sѿ \"%02Xh, %02Xh\" Ǥ]", mon_gettarget(ptr_condition[1]), ptr_condition[2], ptr_condition[3]);
					}
				}
				else if (ptr_condition[0] == 0x0c)
				{
					/* XXX: Ʈϸ˷вᤷե졼ͰʾǤ(0C 00 yy zz) */
					/*   yy: ե졼 (̥Х) */
					/*   zz: ե졼 (̥Х) */
					fprintf(fp, "[Ʈϸηв֤%dե졼ʾǤ]",  (ptr_condition[3] << 8) | ptr_condition[2]);
				}
				else if (ptr_condition[0] == 0x0d)
				{
					/* ٥ȥե饰ꤵƤ(0D 00 yy zz) */
					/*   yy: ٥ȥե饰ֹ */
					/*   zz: ӥå */
					if (ptr_condition[2] == 0x03)
					{
						int index;
						char str[256] = "";

						for (index = 0; index < 8; index++)
						{
							if ((ptr_condition[3] & (1U << (7 - index))) != 0x00)
								snprintf(str, sizeof(str), "%s%s%s", str, (strlen(str) > 0) ? ", " : "", chr_getname(index));
						}
						fprintf(fp, "[%s󤻤]", str);
					}
					else
					{
						fprintf(fp, "[٥ȥե饰%d \"%02Xh\" Ǥ]", ptr_condition[2], ptr_condition[3]);
					}
				}
				else if (ptr_condition[0] == 0x0e)
				{
					/* ᡼Ф륫󥿡(0E 00 00 00) */
					fprintf(fp, "[᡼]");
				}
				else if (ptr_condition[0] == 0x0f)
				{
					/* եʥ륢å(0F 00 00 00) */
					fprintf(fp, "[եʥ륢å]");
				}
				else if (ptr_condition[0] == 0x10)
				{
					/* ̣饯 1 ͤǤ(10 00 00 00) */
					fprintf(fp, "[̣饯1ͤǤ]");
				}
				else if (ptr_condition[0] == 0x11)
				{
					/* ˡФ륫󥿡(11 00 00 00) */
					fprintf(fp, "[ˡ]");
				}
				else if (ptr_condition[0] == 0x12)
				{
					/* 饯ʤ(12 00 00 00) */
					fprintf(fp, "[饯ʤ]");
				}
				else
				{
					fprintf(fp, "[%02X %02X %02X %02X]", ptr_condition[0], ptr_condition[1], ptr_condition[2], ptr_condition[3]);
				}
			}

			fputc('\n', fp);

			/*  */
			turn = 1;
			for (index = 0; index < (int)bufsize; index++)
			{
				if (ptr_action[index] != 0xfc && ptr_action[index] != 0xfd)
				{
					buf_turn[index] = turn;
					turn++;
				}
				else
				{
					buf_turn[index] = turn;

					/* fc xx yy zz */
					if (ptr_action[index] == 0xfc || ptr_action[index + 1] < 0xf0)
					{
						for (k = 0; k < 3; k++)
						{
							index++;
							assert(index < (int)bufsize);
							buf_turn[index] = turn;
						}
						turn++;
					}
					/* fd f7 xx yy */
					else if (ptr_action[index + 1] == 0xf7)
					{
						const int skip_byte = ptr_action[index + 2];
						int skip_index = skip_byte + 3;

						/* ʣιư˹ưѥ󤬽λϡskip_byte  1  */
						if (index + skip_index == (int)bufsize)
							skip_index -= 1;

						for (k = 0; k < skip_index; k++)
						{
							index++;

							/* ʹưѥǡγǧ */
							/* Ƥưѥ󤬼ºݤ˻ȤƤ뤿ᡢassert Ϥʤ */
							if (index >= (int)bufsize) {
								fprintf(stderr, "ff5monster.c: No.%03d(%03X) AI script is corrupted.\n", i, i);
								break;
							}

							buf_turn[index] = turn;
						}
						turn++;
					}
					/* fd xx yy zz */
					else
					{
						for (k = 0; k < 3; k++)
						{
							index++;
							assert(index < (int)bufsize);
							buf_turn[index] = turn;
						}
					}
				}
			}

			/* ưѥ */
			for (index = 0; index < (int)bufsize; index++)
			{
				/* ɽ */
				turn = buf_turn[index];
				if (turn != turn_old)
					letter_num = fprintf(fp, "%d: ", turn);
				else
					fprintf(fp, "%*s", letter_num, "");
				turn_old = turn;

				/* ưѥǡβ */
				if (ptr_action[index] != 0xfc && ptr_action[index] != 0xfd)
				{
					mon_printact(fp, ptr_action[index], &ptr_mondata[i]);
				}
				else
				{
					/* ư(fc xx yy zz) */
					if (ptr_action[index] == 0xfc || ptr_action[index + 1] < 0xf0)
					{
						for (k = 0; k < 3; k++)
						{
							if (k != 0)
								fprintf(fp, ", ");
							index++;
							mon_printact(fp, ptr_action[index], &ptr_mondata[i]);
						}
					}
					/* ؿ(fd f* yy zz) */
					else
					{
						const unsigned char *ptr_sub = &ptr_action[index + 1];
						index += 3;

						if (ptr_sub[0] == 0xf2)
						{
							/* и(f2 xx yy) */
							/*   xx: Ū */
							/*       bit0-5: Ū */
							/*       bit6: ꤵ줿󥹥Τ줫 */
							/*       bit7: 0: HPѾ, 1: HP */
							/*   yy: (ӥå) */
							fprintf(fp, "[и(HP%s): ", ((ptr_sub[1] & 0x80) != 0x00) ? "" : "Ѿ");

							if (ptr_sub[2] == 0x00)
								fprintf(fp, "ʬȤȼʬʳΥ󥹥ï");
							else
							{
								int index;
								char str_no[64] = "";

								for (index = 0; index < 8; index++)
								{
									if ((ptr_sub[2] & (1U << (7 - index))) != 0x00)
										snprintf(str_no, sizeof(str_no), "%s%s%d", str_no, (strlen(str_no) > 0) ? ", " : "", index + 1);
								}

								fprintf(fp, "%sܤΥ󥹥%s", str_no, ((ptr_sub[1] & 0x40) != 0x00) ? "Τ줫" : "");
							}

							fprintf(fp, "]");
							fprintf(fp, " [и: %02Xh]", ptr_sub[1] & 0x3f);
						}
						else if (ptr_sub[0] == 0xf3)
						{
							/* ɸ(f3 xx yy) */
							/*   xx: о */
							/*   yy: ̤ (F0 or 00) */
							fprintf(fp, "[ɸ: %s]", mon_gettarget(ptr_sub[1]));
						}
						else if (ptr_sub[0] == 0xf4)
						{
							/* ѿ(f4 xx yy) */
							/*   yy: ѿֹ */
							/*   zz: ѿ */
							fprintf(fp, "[ѿ%d: %d]", ptr_sub[1], ptr_sub[2]);
						}
						else if (ptr_sub[0] == 0xf6)
						{
							/* åɽ(f6 xx yy) */
							/*   xx: åɽÿ */
							/*   yy: åֹ */
							fprintf(fp, "[åɽ: %s]", mon_getmes(ptr_sub[2]));
						}
						else if (ptr_sub[0] == 0xf7)
						{
							/* 1ʣιư(f7 xx yy) */
							/*   xx: ư (Хȿ) */
							/*   yy: F0  */
							fprintf(fp, "[ʣư]");
						}
						else if (ptr_sub[0] == 0xf8)
						{
							/* и(f8 xx yy) */
							/*   xx:  */
							/*   yy: ֹ */
							fprintf(fp, "[и: ");

							switch (ptr_sub[1])
							{
							case 0x00:
								fprintf(fp, "BGMѹ: %02Xh", ptr_sub[2]);
								break;
							case 0x01:
								fprintf(fp, "̲: %02Xh", ptr_sub[2]);
								break;
							case 0x03:
								fprintf(fp, "եå");
								break;
							case 0x07:
								fprintf(fp, ": %02Xh", ptr_sub[2]);
								break;
							case 0x09:
							case 0x0a:
								fprintf(fp, "%s%4s: %02Xh",
										(ptr_sub[1] == 0x09) ? "Ũư" : "Ũѿ",
										((ptr_sub[2] & 0x80) == 0x00) ? "ON" : "OFF", ptr_sub[2] & 0x7f);
								break;
							case 0x04:
							case 0x06:	/* αи */
							case 0x05:	/* طʤαи(1) */
							case 0x08:	/* طʤαи(2) */
							default:
								fprintf(fp, "%02Xh, %02Xh", ptr_sub[1], ptr_sub[2]);
							}

							fprintf(fp, "]");
						}
						else if (ptr_sub[0] == 0xf9)
						{
							/* ٥ȥե饰(f9 xx yy) */
							/*   xx: ٥ȥե饰ֹ */
							/*   yy: ӥå */
							fprintf(fp, "[٥ȥե饰%d: %02Xh]", ptr_sub[1], ptr_sub[2]);
						}
						else if (ptr_sub[0] == 0xfa)
						{
							/* ʬȤѿͤ/ȿž(XOR) (fa xx yy) */
							/*   xx: ɥ쥹 (SNES: $7E2000, GBA: $0201EDFC) Υեå */
							/*   yy: /ӥå */
							if ((ptr_sub[1] >= 0x14 && ptr_sub[1] <= 0x18) ||
								(ptr_sub[1] >= 0x6c && ptr_sub[1] <= 0x6f))
							{
								const int index = (ptr_sub[1] >= 0x6c) ? 0x6c : 0x14;
								char str[256] = "";

								if (ptr_sub[1] - index == 4)
									str_setstate_f(str, (state_t)ptr_sub[2], sizeof(str));
								else
									str_setstate(str, (state_t)bits_make32(
										(ptr_sub[1] - index == 0) ? ptr_sub[2] : 0x00,
										(ptr_sub[1] - index == 1) ? ptr_sub[2] : 0x00,
										(ptr_sub[1] - index == 2) ? ptr_sub[2] : 0x00,
										(ptr_sub[1] - index == 3) ? ptr_sub[2] : 0x00), sizeof(str));

								fprintf(fp, "[ӥåȿž: %s%d, %s]",
										(index == 0x6c) ? "³" : "", ptr_sub[1] - index + 1, str);
							}
							else
							{
								fprintf(fp, "[ӥå: %02Xh, %02Xh]", ptr_sub[1], ptr_sub[2]);
							}
						}
						else
						{
							fprintf(fp, "[%02X %02X %02X]", ptr_sub[0], ptr_sub[1], ptr_sub[2]);
						}
					}
				}

				fputc('\n', fp);
			}

			free(buf_turn);
			buf_turn = NULL;
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemes(const char *ptr_filename)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0')
	{
		assert(0);
		return 0;
	}

	/* 󥹥ΥåǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ΥåǡϤ */
	fprintf(fp, "No.      å\n");
	for (i = 0; i < NUM_MONMES; i++)
	{
		fprintf(fp, "%03d(%02X): %s\n", i, i, mon_getmes(i));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeparty(const char *ptr_filename, const monparty_t *ptr_monparty, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
#if 0
	const char *str_formation[2] = {"",""};
	const char *str_music[9] =
	{
		"Хȥ1",
		"Хȥ2",
		"Ǹ襤",
		"ӥå֥å¤λƮ",
		"Ʋǥ",
		"",
		"⡼Υơ",
		"(̵٤ǥ߹ݤθ̲)",
		"(ʤڤؤʤ)",
	};
#endif

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ΥѡƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ΥѡƥǡϤ */
	fprintf(fp, "No.       "
				" ABP ƨ  "
				" ƨ"
				"   "
				"и       "
				"󥹥1         "
				"󥹥2         "
				"󥹥3         "
				"󥹥4         "
				"󥹥5         "
				"󥹥6         "
				"󥹥7         "
				"󥹥8\n");
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		fprintf(fp, "%03d(%03X):  %3d, %3d,  %2s%2s%2s%2s%2s   ", i, i,
				ptr_monparty[i].ability_point,
				ptr_monparty[i].escape_rate,
				str_getflag(ptr_monparty[i].flag_escape),
				str_getflag(ptr_monparty[i].flag_boss),
				str_getflag(ptr_monparty[i].flag_raid),
				str_getflag(ptr_monparty[i].flag_magic),
				str_getflag(ptr_monparty[i].flag_mute));

		for (j = 0; j < 8; j++)
			fprintf(fp, "%2s", str_getflag((ptr_monparty[i].bits_exist & (1U << (7 - j))) != 0x00));

		fprintf(fp, "  ");

		mon_printparty(fp, &ptr_monparty[i], ptr_mondata);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeenc(const char *ptr_filename, const monenc_t *ptr_monenc, const monparty_t *ptr_monparty, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monenc == NULL || ptr_monparty == NULL || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ơ֥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ơ֥Ϥ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		const int len_monsters = (i < NUM_MONENCRND) ? 4 : 2;

		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X)]\n", i, i);
		for (j = 0; j < len_monsters; j++)
		{
			const int monparty_no = ptr_monenc[i].party_no[j];

			fprintf(fp, "%02d: %03d(%03X): ", j + 1, monparty_no, monparty_no);

			mon_printparty(fp, &ptr_monparty[monparty_no], ptr_mondata);
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemap(const char *ptr_filename, const monmap_t *ptr_monmap)
{
	int i, j, x, y;
	FILE *fp = NULL;
	const char *str_name[NUM_MONMAP][4] =
	{
		{"1(ʿ)",	"1()",		"1()",	"1()"},
		{"2(ʿ)",	"2()",		"2()",		"2()"},
		{"3(ʿ)",	"3()",		"3()",	"3()"},
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monmap == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ޥåץǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ޥåפϤ */
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (j = 0; j < 4; j++)
		{
			if (i != 0 || j != 0)
				fputc('\n', fp);

			fprintf(fp, "[No.%1d-%1d: %s]\n", i, j, str_name[i][j]);
			for (y = 0; y < NUM_MONMAPAREA; y++)
			{
				for (x = 0; x < NUM_MONMAPAREA; x++)
				{
					if (x != 0)
						fprintf(fp, " ");

					fprintf(fp, "%03d", ptr_monmap[i].area[x][y].enc_no[j]);
				}
				fputc('\n', fp);
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writelocation(const char *ptr_filename, const mapdata_t *ptr_mapdata,
	const monenc_t *ptr_monenc, const monparty_t *ptr_monparty, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mapdata == NULL ||
		ptr_monenc == NULL || ptr_monparty == NULL || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ǡϤ */
	for (i = 0; i < NUM_MAP; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X): %s]\n", i, i, map_getname(ptr_mapdata[i].id));

		/* 󥹥ơ֥Ϥ */
		if (ptr_mapdata[i].flag_encounter)
		{
			for (j = 0; j < 4; j++)
			{
				const int monparty_no = ptr_monenc[ptr_mapdata[i].enc_no].party_no[j];

				fprintf(fp, "%02d: %03d(%03X): ", j + 1, monparty_no, monparty_no);

				mon_printparty(fp, &ptr_monparty[monparty_no], ptr_mondata);
				fputc('\n', fp);
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

static void mon_printact(FILE *stream, int index, const mondata_t *ptr_mondata)
{
	/* å */
	if (stream == NULL || index < 0 || 0xff <= index || ptr_mondata == NULL)
	{
		assert(0);
		return;
	}

	switch (index)
	{
	case 0x81:		/* Ȥ虜 */
		fprintf(stream, "%s", mon_gettechname(ptr_mondata->tech_no));
		break;
	case 0xaa:		/*  */
		fprintf(stream, "(⤷ʤ)");
		break;
	case 0xa5:		/* Фʤ */
	case 0xb1:		/* ޤ */
		fprintf(stream, "()");
		break;
	case 0xde:		/* 椦Ф */
		fprintf(stream, "ʪ");
		break;
	case 0xec:
		fprintf(stream, "(󤻤)");
		break;
	case 0xed:
		fprintf(stream, "(󥹥)");
		break;
	case 0xee:
		fprintf(stream, "(󥹥иѹ)");
		break;
	case 0xef:		/* 夦礦 */
		fprintf(stream, "(Ʈ)");
		break;
	default:
		fprintf(stream, "%s", mag_getname(index));
	}
}

void mon_freeact(monact_t *ptr_monact)
{
	int i, j;

	/* å */
	if (ptr_monact == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_MON; i++)
	{
		for (j = 0; j < ptr_monact[i].ptn_num; j++)
		{
			if (ptr_monact[i].ptn[j].condition != NULL)
			{
				free(ptr_monact[i].ptn[j].condition);
				ptr_monact[i].ptn[j].condition = NULL;
			}
			if (ptr_monact[i].ptn[j].action != NULL)
			{
				free(ptr_monact[i].ptn[j].action);
				ptr_monact[i].ptn[j].action = NULL;
			}
		}
	}
}

void mon_printparty(FILE *stream, const monparty_t *ptr_monparty, const mondata_t *ptr_mondata)
{
	int i;

	/* å */
	if (stream == NULL || ptr_monparty == NULL || ptr_mondata == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < 8; i++)
	{
		char str_name[64];
		const char *ptr_name = NULL;
		const int flag_absent = ((ptr_monparty->bits_exist & (1U << (7 - i))) == 0x00);

		if (ptr_monparty->monster_no[i] == 0xff)
			break;

		ptr_name = mon_getname(ptr_mondata[ptr_monparty->monster_no[i]].id);
		if (*ptr_name == '\0')
		{
			snprintf(str_name, sizeof(str_name), "[No.%03d(%03X)]",
					ptr_monparty->monster_no[i], ptr_monparty->monster_no[i]);
			ptr_name = str_name;
		}

		if (i != 0)
			fprintf(stream, ", ");
		fprintf(stream, "%s%-16s%s",
				(flag_absent) ? "(" : " ",
				ptr_name,
				(flag_absent) ? ")" : " ");
	}
}
