/*
 *	ff6analyzer.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *  ff6a_analyzer
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "utility.h"
#include "file.h"
#include "ff6decode.h"
#include "ff6monster.h"
#include "ff6map.h"
#include "ff6character.h"
#include "ff6item.h"
#include "ff6magic.h"
#include "ff6misc.h"

#define FILENAME_MON_PARAM		"ff6monparam.txt"
#define FILENAME_MON_ELEM		"ff6monelem.txt"
#define FILENAME_MON_STATE		"ff6monstate.txt"
#define FILENAME_MON_FLAG		"ff6monflag.txt"
#define FILENAME_MON_ITEM		"ff6monitem.txt"
#define FILENAME_MON_RAGNA		"ff6monragna.txt"
#define FILENAME_MON_CTRL		"ff6monctrl.txt"
#define FILENAME_MON_SKETCH		"ff6monsketch.txt"
#define FILENAME_MON_RAGE		"ff6monrage.txt"
#define FILENAME_MON_TECH		"ff6montech.txt"
#define FILENAME_MON_MISC		"ff6monmisc.txt"
#define FILENAME_MON_ACT		"ff6monact.txt"
#define FILENAME_MON_MES		"ff6monmes.txt"
#define FILENAME_MON_EVENT		"ff6monevent.txt"
#define FILENAME_MON_PARTY		"ff6monparty.txt"
#define FILENAME_MON_ENC		"ff6monenc.txt"
#define FILENAME_MON_MAP		"ff6monmap.txt"
#define FILENAME_MON_LOCATION	"ff6monlocation.txt"
#define FILENAME_MON_VELDT		"ff6monveldt.txt"
#define FILENAME_CHR_PARAM		"ff6chrparam.txt"
#define FILENAME_CHR_EQUIP		"ff6chrequip.txt"
#define FILENAME_CHR_CMD		"ff6chrcmd.txt"
#define FILENAME_CHR_LVUP		"ff6chrlvup.txt"
#define FILENAME_ITEM_PARAM		"ff6itemparam.txt"
#define FILENAME_ITEM_ELEM		"ff6itemelem.txt"
#define FILENAME_ITEM_STATE		"ff6itemstate.txt"
#define FILENAME_ITEM_FLAG		"ff6itemflag.txt"
#define FILENAME_ITEM_EFFECT	"ff6itemeffect.txt"
#define FILENAME_ITEM_COL		"ff6itemcol.txt"
#define FILENAME_MAG_PARAM		"ff6magparam.txt"
#define FILENAME_MAG_ELEM		"ff6magelem.txt"
#define FILENAME_MAG_STATE		"ff6magstate.txt"
#define FILENAME_MAG_FLAG		"ff6magflag.txt"
#define FILENAME_ESP_DATA		"ff6espparam.txt"
#define FILENAME_TREASURE		"ff6treasure.txt"
#define FILENAME_TEXT_MENU		"ff6menu.txt"
#define FILENAME_TEXT_KANA		"ff6kana.txt"
#define FILENAME_TEXT_KANJI		"ff6kanji.txt"

#define STR_INVALID_OPTION			"Invalid option"
#define STR_INVALID_ROMIMAGE_FILE	"Invalid ROM-Image file: %s"
#define STR_INVALID_CDROM_FILES		"Invalid CD-ROM files"
#define STR_CANNOT_CREATE_FILE		"Cannot create file: %s"
#define STR_FILENAME_TOO_LONG		"Filename too long"

#define STR_READ_FROM_FILE			"Reading from %s\n"
#define STR_WRITE_TO_FILE			"Writing to %s\n"

/* ɽ */
static const char *str_copyright =
	"ff6analyzer   - Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>\n"
	"ff6a_analyzer - Modified by Inu";

static const char *tbl_gametitle[] =
{
	"FINAL FANTASY 6      ",
			/* 0x46, 0x49, 0x4E, 0x41, 0x4C, 0x20, 0x46, 0x41, 0x4E, 0x54, 0x41, 0x53, 0x59, 0x20, 0x36, */
			/* 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 */
	NULL,
			/* unused */
	"FF6ADVANCE\0\0",
			/* 0x46, 0x46, 0x36, 0x41, 0x44, 0x56, 0x41, 0x4E, 0x43, 0x45, 0x00, 0x00 */
};

void usage(void);

#define do_writedata(filename, function) do {        \
	char path[sizeof(str_dstpath) + 128] = "";       \
	snprintf(path, sizeof(path), "%s%s", str_dstpath, filename); \
	if (!flag_dry)                                    \
	{                                                 \
		fprintf(stderr, STR_WRITE_TO_FILE, path);     \
		if (!function)                                \
			err_printf(STR_CANNOT_CREATE_FILE, path); \
	}                                                 \
} while (0)

int main(int argc, char *argv[])
{
	int optind = 0;
	int flag_dry = 0;
	char str_srcpath[1024] = "";
	char str_dstpath[1024] = "";
	mondata_t  *buf_mondata  = NULL;
	monact_t   *buf_monact   = NULL;
	monparty_t *buf_monparty = NULL;
	monenc_t   *buf_monenc   = NULL;
	monmap_t   *buf_monmap   = NULL;
	mapdata_t  *buf_mapdata  = NULL;
	chrdata_t  *buf_chrdata  = NULL;
	chrlvup_t  *buf_chrlvup  = NULL;
	itemdata_t *buf_itemdata = NULL;
	magdata_t  *buf_magdata  = NULL;
	espdata_t  *buf_espdata  = NULL;
	treasure_t *buf_trdata   = NULL;

	errno = 0;
	err_setprogname(argv[0]);

	/* ޥɥ饤Ϥ */
	for (optind = 1; optind < argc; optind++)
	{
		const char *p = argv[optind];

		if (p == NULL)
			continue;

		if (*p != '-')
			break;
		else if (*(p + 1) == '-')
		{
			optind++;
			break;
		}

		while (*++p != '\0')
		{
			switch (*p)
			{
			case 'n':
				flag_dry = 1;
				break;
			case 'h':	/* ˡɽ */
			default:
				usage();
				break;
			}
		}
	}
	argc -= optind;
	argv += optind;

	/* ϥѥꤹ */
	if (argc < 1)
		usage();

	if (strlen(argv[0]) > sizeof(str_srcpath))
		err_printf(STR_FILENAME_TOO_LONG);
	transfilename(str_srcpath, argv[0], sizeof(str_srcpath));

	/* ϥѥꤹ */
	if (argc >= 2)
	{
		if (strlen(argv[1]) > sizeof(str_dstpath))
			err_printf(STR_FILENAME_TOO_LONG);
		transpath(str_dstpath, argv[1], sizeof(str_dstpath));
	}

	/* ե򥪡ץ󤹤 */
	fprintf(stderr, STR_READ_FROM_FILE, str_srcpath);
	if (!file_open(str_srcpath, tbl_gametitle[FILEMODE_GBA], FILEMODE_GBA))
		err_printf(STR_INVALID_ROMIMAGE_FILE, str_srcpath);

	/* ƼǡϡϤ */
	buf_mondata  = mon_readdata();
	buf_monact   = mon_readact();
	buf_monparty = mon_readparty();
	buf_monenc   = mon_readenc();
	buf_monmap   = mon_readmap();
	buf_mapdata  = map_readdata();
	buf_chrdata  = chr_readdata();
	buf_chrlvup  = chr_readlvup();
	buf_itemdata = item_readdata();
	buf_magdata  = mag_readdata();
	buf_espdata  = mag_readesp();
	buf_trdata   = tr_readdata();

	if (buf_mondata  == NULL ||
		buf_monact   == NULL ||
		buf_monparty == NULL ||
		buf_monenc   == NULL ||
		buf_monmap   == NULL ||
		buf_monmap   == NULL ||
		buf_chrdata  == NULL ||
		buf_chrlvup  == NULL ||
		buf_itemdata == NULL ||
		buf_magdata  == NULL ||
		buf_espdata  == NULL ||
		buf_trdata  == NULL)
	{
		err_printf(STR_INVALID_ROMIMAGE_FILE, str_srcpath);
	}

	/* ƼǡϤ */
	do_writedata(FILENAME_MON_PARAM,	mon_writeparam(path, buf_mondata));
	do_writedata(FILENAME_MON_ELEM,		mon_writeelem(path, buf_mondata));
	do_writedata(FILENAME_MON_STATE,	mon_writestate(path, buf_mondata));
	do_writedata(FILENAME_MON_FLAG,		mon_writeflag(path, buf_mondata));
	do_writedata(FILENAME_MON_ITEM,		mon_writeitem(path, buf_mondata));
	do_writedata(FILENAME_MON_RAGNA,	mon_writeragna(path, buf_mondata));
	do_writedata(FILENAME_MON_CTRL,		mon_writectrl(path, buf_mondata));
	do_writedata(FILENAME_MON_SKETCH,	mon_writesketch(path, buf_mondata));
	do_writedata(FILENAME_MON_RAGE,		mon_writerage(path, buf_mondata));
	do_writedata(FILENAME_MON_TECH,		mon_writetech(path, buf_mondata));
	do_writedata(FILENAME_MON_MISC,		mon_writemisc(path, buf_mondata));
	do_writedata(FILENAME_MON_ACT,		mon_writeact(path, buf_monact, buf_mondata));
	do_writedata(FILENAME_MON_MES,		mon_writemes(path));
	do_writedata(FILENAME_MON_EVENT,	mon_writeevent(path));
	do_writedata(FILENAME_MON_PARTY,	mon_writeparty(path, buf_monparty));
	do_writedata(FILENAME_MON_VELDT,	mon_writeveldt(path, buf_monparty));
	do_writedata(FILENAME_MON_ENC,		mon_writeenc(path, buf_monenc, buf_monparty));
	do_writedata(FILENAME_MON_MAP,		mon_writemap(path, buf_monmap));
	do_writedata(FILENAME_MON_LOCATION,	mon_writelocation(path, buf_mapdata, buf_monenc, buf_monparty));
	do_writedata(FILENAME_CHR_PARAM,	chr_writeparam(path, buf_chrdata));
	do_writedata(FILENAME_CHR_EQUIP,	chr_writeequip(path, buf_chrdata));
	do_writedata(FILENAME_CHR_CMD,		chr_writecmd(path, buf_chrdata));
	do_writedata(FILENAME_CHR_LVUP,		chr_writelvup(path, buf_chrlvup));
	do_writedata(FILENAME_ITEM_PARAM,	item_writeparam(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_ELEM,	item_writeelem(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_STATE,	item_writestate(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_FLAG,	item_writeflag(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_EFFECT,	item_writeeffect(path, buf_itemdata));
	do_writedata(FILENAME_ITEM_COL,		item_writecol(path, buf_itemdata));
	do_writedata(FILENAME_MAG_PARAM,	mag_writeparam(path, buf_magdata));
	do_writedata(FILENAME_MAG_ELEM,		mag_writeelem(path, buf_magdata));
	do_writedata(FILENAME_MAG_STATE,	mag_writestate(path, buf_magdata));
	do_writedata(FILENAME_MAG_FLAG,		mag_writeflag(path, buf_magdata));
	do_writedata(FILENAME_ESP_DATA,		mag_writeesp(path, buf_espdata));
	do_writedata(FILENAME_TREASURE,		tr_writedata(path, buf_trdata, buf_monenc, buf_monparty));
	do_writedata(FILENAME_TEXT_MENU,	dec_writedata(path, 0, -1, FLAG_KANA  | FLAG_SHORT));
	do_writedata(FILENAME_TEXT_KANA,	dec_writedata(path, 0, -1, FLAG_KANA  | FLAG_LONG));
	do_writedata(FILENAME_TEXT_KANJI,	dec_writedata(path, 0, -1, FLAG_KANJI | FLAG_LONG));

	/* ե򥯥 */
	file_close();

	/* Хåե */
	mon_freeact(buf_monact);
	free(buf_mondata);
	free(buf_monact);
	free(buf_monparty);
	free(buf_monenc);
	free(buf_monmap);
	free(buf_mapdata);
	free(buf_chrdata);
	free(buf_chrlvup);
	free(buf_itemdata);
	free(buf_magdata);
	free(buf_espdata);
	buf_mondata  = NULL;
	buf_monact   = NULL;
	buf_monparty = NULL;
	buf_monenc   = NULL;
	buf_monmap   = NULL;
	buf_mapdata  = NULL;
	buf_chrdata  = NULL;
	buf_chrlvup  = NULL;
	buf_itemdata = NULL;
	buf_magdata  = NULL;
	buf_espdata  = NULL;

	return EXIT_SUCCESS;
}

void usage(void)
{
	fprintf(stderr, "%s\n\n", str_copyright);
	fprintf(stderr,	"usage: ff6analyzer: [-hn] src-pathname [dst-pathname]\n"
					"    -h                - display this help and exit\n"
					"    -n                - dry run\n"
					"    src-pathname      - ROM image file\n"
					"    dst-pathname      - write files into directory\n");

	exit(EXIT_SUCCESS);
}
