/*
 *	ff6character.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff6decode.h"
#include "ff6item.h"
#include "ff6character.h"

#define FILEPOS_CHREXP			0x62abc8	/* SNES: 0x2d8220 */
#define FILEPOS_CHRHP			0x62ec28	/* SNES: 0x26f4a0 */
#define FILEPOS_CHRMP			0x62aa54	/* SNES: 0x26f502 */
#define FILEPOS_CHRNAME			0x799e8e	/* SNES: 0x0478c0 */
#define FILEPOS_CHRJOBNAME		0x79ab86	/* SNES: 0x0f3b40 */
#define FILEPOS_CHRDATA			0x62acac	/* SNES: 0x2d7ca0 */
#define FILEPOS_CHRCMDNAME		0x60301e	/* SNES: 0x2ce300 */
#define FILEPOS_CHRLEARN1		0x62aafc	/* SNES: 0x2ce3c0 */
#define FILEPOS_CHRLEARN2		0x62abb8	/* SNES: 0x26f490 */
#define FILEPOS_BLITZDATA		0x62ab3c	/* SNES: 0x047a40 */
#define NUM_CHR					64
#define NUM_CHRNAME				NUM_CHR
#define NUM_CHRJOBNAME			18			/* SNES: NUM_CHR */
#define NUM_CHRCMDNAME			32
#define NUM_CHRLEARN1			2
#define NUM_CHRLEARN2			2
#define SIZEOF_CHREXP			2
#define SIZEOF_CHRHP			1
#define SIZEOF_CHRMP			1
#define SIZEOF_CHRNAME			6
#define SIZEOF_CHRJOBNAME		7
#define SIZEOF_CHRDATA			22
#define SIZEOF_CHRCMDNAME		6
#define SIZEOF_CHRLEARN1		16
#define SIZEOF_CHRLEARN2		8
#define MSGID_CHRNAME			282		/* short */
#define MSGID_CHRJOBNAME		649		/* short */
#define MSGID_CHRCMDNAME		346		/* short */		/* long: 5581 */

static void chr_setname(void);
static void chr_setjobname(void);
static void chr_setcmdname(void);

static char tbl_chrname[NUM_CHRNAME][256];			/* 饯̾Ǽơ֥ */
static char tbl_chrjobname[NUM_CHRJOBNAME][256];	/* ֤̾Ǽơ֥ */
static char tbl_cmdname[NUM_CHRCMDNAME][256];		/* Ʈޥɤ̾Ǽơ֥ */
static int flag_setchrname    = 0;					/* tbl_chrname˥ǡǼƤ뤫ɤ */
static int flag_setchrjobname = 0;					/* tbl_chrjobname˥ǡǼƤ뤫ɤ */
static int flag_setcmdname    = 0;					/* tbl_cmdname˥ǡǼƤ뤫ɤ */

const char *chr_getname(int index)
{
	/* å */
	if (index < 0 || NUM_CHRNAME <= index)
	{
		assert(0);
		return "";
	}

	if (!flag_setchrname)
		chr_setname();

	return tbl_chrname[index];
}

const char *chr_getjobname(int index)
{
	/* å */
	if (index < 0 || NUM_CHRJOBNAME <= index)
	{
		assert(0);
		return "";
	}

	if (!flag_setchrjobname)
		chr_setjobname();

	return tbl_chrjobname[index];
}

const char *chr_getcmdname(int index)
{
	/* å */
	if (index < 0 || NUM_CHRCMDNAME <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setcmdname)
		chr_setcmdname();

	return tbl_cmdname[index];
}

static void chr_setname(void)
{
	int i;

	/* 饯̾ǡ */
	for (i = 0; i < NUM_CHRNAME; i++)
		*tbl_chrname[i] = '\0';

	/* 饯̾ǡɤ߹ */
	for (i = 0; i < NUM_CHRNAME; i++)
		dec_msg(tbl_chrname[i], sizeof(tbl_chrname[i]), MSGID_CHRNAME + i, FLAG_KANA | FLAG_SHORT);

	flag_setchrname = 1;
}

static void chr_setjobname(void)
{
	int i;

	/* 饯Υ̾ǡ */
	for (i = 0; i < NUM_CHRJOBNAME; i++)
		*tbl_chrjobname[i] = '\0';

	/* 饯Υ̾ǡɤ߹ */
	for (i = 0; i < NUM_CHRJOBNAME; i++)
		dec_msg(tbl_chrjobname[i], sizeof(tbl_chrjobname[i]), MSGID_CHRJOBNAME + i, FLAG_KANA | FLAG_SHORT);

	flag_setchrjobname = 1;
}

static void chr_setcmdname(void)
{
	int i;

	/* ƮޥɤΥǡ */
	for (i = 0; i < NUM_CHRCMDNAME; i++)
		*tbl_cmdname[i] = '\0';

	/* ƮޥɤΥǡɤ߹ */
	for (i = 0; i < NUM_CHRCMDNAME; i++)
		dec_msg(tbl_cmdname[i], sizeof(tbl_cmdname[i]), MSGID_CHRCMDNAME + i, FLAG_KANA | FLAG_SHORT);

	flag_setcmdname = 1;
}

chrdata_t *chr_readdata(void)
{
	int i, j;
	chrdata_t *buf_chrdata = NULL;
	chrlvup_t tbl_lvup[NUM_CHRLV];
	const int tbl_level_cng[4] = {0, 2, 5, -3};

	/* 饯ǡǼХåեݤ */
	if ((buf_chrdata = malloc(sizeof(chrdata_t) * NUM_CHR)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ٥뤴ȤδܥҥåȥݥȤɤ߹ */
	tbl_lvup[0].hit_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRHP];

		file_read(buf, FILEPOS_CHRHP + SIZEOF_CHRHP * (i - 1), SIZEOF_CHRHP);
		tbl_lvup[i].hit_point = tbl_lvup[i - 1].hit_point + buf[0];
	}

	/* ٥뤴ȤδܥޥåݥȤɤ߹ */
	tbl_lvup[0].magic_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRMP];

		file_read(buf, FILEPOS_CHRMP + SIZEOF_CHRMP * (i - 1), SIZEOF_CHRMP);
		tbl_lvup[i].magic_point = tbl_lvup[i - 1].magic_point + buf[0];
	}

	/* 饯ǡɤ߹ */
	for (i = 0; i < NUM_CHR; i++)
	{
		unsigned char buf[SIZEOF_CHRDATA];

		file_read(buf, FILEPOS_CHRDATA + SIZEOF_CHRDATA * i, SIZEOF_CHRDATA);

		/* ͷϥǡ */
		buf_chrdata[i].strength            = buf[ 6];
		buf_chrdata[i].speed               = buf[ 7];
		buf_chrdata[i].stamina             = buf[ 8];
		buf_chrdata[i].magic_power         = buf[ 9];
		buf_chrdata[i].attack_power        = buf[10];
		buf_chrdata[i].defense_power       = buf[11];
		buf_chrdata[i].evade_rate          = buf[13];
		buf_chrdata[i].magic_defense_power = buf[12];
		buf_chrdata[i].magic_evade_rate    = buf[14];
		buf_chrdata[i].level_cng           = tbl_level_cng[(buf[21] & 0x0c) >> 2];

		buf_chrdata[i].level[0].hit_point   = buf[ 0];
		buf_chrdata[i].level[0].magic_point = buf[ 1];

		/* ե饰ϥǡ */
		buf_chrdata[i].flag_level = ((buf[21] & 0x10) == 0);

		/* ƥǡ */
		for (j = 0; j < 6; j++)
			buf_chrdata[i].equipment_no[j] = buf[j + 15];

		/* ޥɥǡ */
		for (j = 0; j < 4; j++)
			buf_chrdata[i].cmd_no[j] = buf[j + 2];

		/* ٥뤴ȤΥҥåȥݥȡޥåݥ */
		for (j = 1; j < NUM_CHRLV; j++)
		{
			buf_chrdata[i].level[j].hit_point   = tbl_lvup[j].hit_point   + buf_chrdata[i].level[0].hit_point;
			buf_chrdata[i].level[j].magic_point = tbl_lvup[j].magic_point + buf_chrdata[i].level[0].magic_point;
		}
	}

	return buf_chrdata;
}

chrlvup_t *chr_readlvup(void)
{
	int i;
	chrlvup_t *buf_chrlvup = NULL;

	/* 饯ĹǡǼХåեݤ */
	if ((buf_chrlvup = malloc(sizeof(chrdata_t) * NUM_CHRLV)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ٥뤴Ȥηиͥǡɤ߹ */
	buf_chrlvup[0].experience_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHREXP];

		file_read(buf, FILEPOS_CHREXP + SIZEOF_CHREXP * (i - 1), SIZEOF_CHREXP);
		buf_chrlvup[i].experience_point = buf_chrlvup[i - 1].experience_point + ((buf[1] << 8) | buf[0]) * 8;
	}

	/* ٥뤴ȤδܥҥåȥݥȤɤ߹ */
	buf_chrlvup[0].hit_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRHP];

		file_read(buf, FILEPOS_CHRHP + SIZEOF_CHRHP * (i - 1), SIZEOF_CHRHP);
		buf_chrlvup[i].hit_point = buf_chrlvup[i - 1].hit_point + buf[0];
	}

	/* ٥뤴ȤδܥޥåݥȤɤ߹ */
	buf_chrlvup[0].magic_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRMP];

		file_read(buf, FILEPOS_CHRMP + SIZEOF_CHRMP * (i - 1), SIZEOF_CHRMP);
		buf_chrlvup[i].magic_point = buf_chrlvup[i - 1].magic_point + buf[0];
	}

	return buf_chrlvup;
}

int chr_writeparam(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i;
	FILE *fp = NULL;
	const int tbl_jobname_no[34] =
	{
		/* 0x000x0f */
		0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,

		/* 0x100x11 */
		16, 16,

		/* 0x120x1f */
		10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,

		/* 0x200x21 */
		17, 17,
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯ǡ(ͷ)Ϥե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯ǡ(ͷ)Ϥ */
	fprintf(fp, "No.      ̾           "
				"          "
				"HP MP "
				"     ®   "
				"           "
				"LV LVƶ\n");
	for (i = 0; i < NUM_CHR; i++)
	{
		fprintf(fp, "%03d(%02X): %-12s : %-14s,  %5d, %5d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %+5d,  %2s\n", i, i,
				chr_getname(i),
				(i < (int)(sizeof(tbl_jobname_no) / sizeof(tbl_jobname_no[0]))) ? chr_getjobname(tbl_jobname_no[i]) : "",
				ptr_chrdata[i].level[0].hit_point,
				ptr_chrdata[i].level[0].magic_point,
				ptr_chrdata[i].strength,
				ptr_chrdata[i].speed,
				ptr_chrdata[i].stamina,
				ptr_chrdata[i].magic_power,
				ptr_chrdata[i].attack_power,
				ptr_chrdata[i].defense_power,
				ptr_chrdata[i].evade_rate,
				ptr_chrdata[i].magic_defense_power,
				ptr_chrdata[i].magic_evade_rate,
				ptr_chrdata[i].level_cng,
				str_getflag(ptr_chrdata[i].flag_level));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writeequip(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯νƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯νƥǡϤ */
	fprintf(fp, "No.      ̾           "
				"              "
				"              "
				"Ƭ                "
				"                "
				"꡼1     "
				"꡼2\n");
	for (i = 0; i < NUM_CHR; i++)
	{
		const char *ptr_name[6];

		for (j = 0; j < 6; j++)
			ptr_name[j] = (ptr_chrdata[i].equipment_no[j] != 0xff) ? item_getname(ptr_chrdata[i].equipment_no[j]) : str_nothing;

		fprintf(fp, "%03d(%02X): %-12s : ", i, i, chr_getname(i));
		for (j = 0; j < 6; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmd(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯ΥޥɥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯ΥޥɥǡϤ */
	fprintf(fp, "No.      ̾           "
				"ޥ1         "
				"ޥ2         "
				"ޥ3         "
				"ޥ4\n");
	for (i = 0; i < NUM_CHR; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
			ptr_name[j] = (ptr_chrdata[i].cmd_no[j] != 0xff) ? chr_getcmdname(ptr_chrdata[i].cmd_no[j]) : str_nothing;

		fprintf(fp, "%03d(%02X): %-12s : ", i, i, chr_getname(i));
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writelvup(const char *ptr_filename, const chrlvup_t *ptr_chrlvup)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrlvup == NULL)
	{
		assert(0);
		return 0;
	}

	/* 饯ĹǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 饯ĹǡϤ */
	fprintf(fp, "LV       EXP   NEXT HP 徺HP MP 徺MP\n");
	for (i = 0; i < NUM_CHRLV; i++)
	{
		fprintf(fp, "%2d: %8ld, %5ld, %5d, %5d, %5d, %5d\n", i + 1,
				ptr_chrlvup[i].experience_point,
				(i >= NUM_CHRLV - 1) ? 0 : ptr_chrlvup[i + 1].experience_point - ptr_chrlvup[i].experience_point,
				ptr_chrlvup[i].hit_point,
				(i == 0) ? 0 : ptr_chrlvup[i].hit_point - ptr_chrlvup[i - 1].hit_point,
				ptr_chrlvup[i].magic_point,
				(i == 0) ? 0 : ptr_chrlvup[i].magic_point - ptr_chrlvup[i - 1].magic_point);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
