/*
 *	ff6define.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "ff6define.h"

#define STR_NOTHING		""
#define STR_FLAG_TRUE	""
#define STR_FLAG_FALSE	""

static const char *str_element[NUM_ELEMENT] =
{
	"",
	"ɹ",
	"",
	"",
	"",
	"",
	"",
	"",
};

static const char *str_state[NUM_STATE] =
{
	"",	/* Ű */
	"",	/* ӡ */
	"",	/*  */
	"?1",	/* Ƴޡ */
	"Ʃ",	/* Ʃ */
	"",	/* å */
	"",	/* в */
	"",	/* Ʈǽ */

	"",	/*  */
	"",	/* λ */
	"ʬ",	/* ʬ */
	"",	/*  */
	"",	/* С */
	"",	/*  */
	"",	/* å */
	"̲",	/* ̲ */

	"",	/* Ǥ */
	"",	/* ꥸ */
	"",	/*  */
	"®",	/* إ */
	"",	/* ȥå */
	"",	/*  */
	"ʪ",	/* ץƥ */
	"ȿ",	/* ե쥯 */

	"˽",	/* Ф */
	"",	/* ɹŤ/ */
	"",	/* 쥤 */
	"",	/* ȥ */
	"",	/* ˡ򾧤 */
	"æ",	/* Υæ */
	"˸",	/* 󥿡ץ */
	"",	/* ӥƥ */
};

const char *str_getflag(int flag)
{
	return (flag != 0x00) ? STR_FLAG_TRUE : STR_FLAG_FALSE;
}

const char *str_getelem(elem_t elem, int index)
{
	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* å */
	if (index < 0 || NUM_ELEMENT <= index)
	{
		assert(0);
		return "";
	}

	return ((elem & (1U << index)) != 0x00) ? str_element[index] : "";
}

void str_printelem(FILE *stream, elem_t elem, elem_t elem_mask)
{
	int i;

	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* å */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_ELEMENT; i++)
	{
		if ((elem_mask & (1U << i)) != 0x00)
			fprintf(stream, "%-2s", ((elem & (1U << i)) != 0x00) ? str_element[i] : STR_NOTHING);
	}
}

const char *str_getstate(state_t state, int index)
{
	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* å */
	if (index < 0 || NUM_STATE <= index)
	{
		assert(0);
		return "";
	}

	return ((state & (1U << index)) != 0x00) ? str_state[index] : "";
}

void str_printstate(FILE *stream, state_t state, state_t state_mask)
{
	int i, count = 0;

	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* å */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_STATE; i++)
	{
		if ((i % 8) == 0)
		{
			if (count > 0 && state_mask > (1U << i))
				fprintf(stream, " ");
			count = 0;
		}

		if ((state_mask & (1U << i)) != 0x00)
		{
			fprintf(stream, "%-2s", ((state & (1U << i)) != 0x00) ? str_state[i] : STR_NOTHING);
			count++;
		}
	}
}
