/*
 *	ff6item.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff6decode.h"
#include "ff6monster.h"
#include "ff6item.h"

#define FILEPOS_ITEMNAME		0x79a00e	/* SNES: 0x26f700 */
#define FILEPOS_ITEMDATA		0x62cc64	/* SNES: 0x185000 */
#define FILEPOS_ITEMCOL			0x62a618	/* SNES: 0x10f200 */
#define NUM_ITEM				271			/* SNES: 256 */
#define SIZEOF_ITEMNAME			9
#define SIZEOF_ITEMDATA			30
#define SIZEOF_ITEMCOL			4
#define MSGID_ITEMNAME			378		/* short */		/* long: 4558 */

static void item_setname(void);

static char tbl_itemname[NUM_ITEM][256];	/* ƥ̾Ǽơ֥ */
static int flag_setitemname = 0;			/* tbl_itemname˥ǡǼƤ뤫ɤ */

const char *item_getname(int index)
{
	/* å */
	if (index < 0 || NUM_ITEM <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setitemname)
		item_setname();

	return tbl_itemname[index];
}

static void item_setname(void)
{
	int i;

	/* ƥ̾ǡ */
	for (i = 0; i < NUM_ITEM; i++)
		*tbl_itemname[i] = '\0';

	/* ƥ̾ǡɤ߹ */
	for (i = 0; i < NUM_ITEM; i++)
		dec_msg(tbl_itemname[i], sizeof(tbl_itemname[i]), MSGID_ITEMNAME + i, FLAG_KANA | FLAG_SHORT);

	flag_setitemname = 1;
}

itemdata_t *item_readdata(void)
{
	int i;
	itemdata_t *buf_itemdata = NULL;

	/* ƥγƼǡǼХåեݤ */
	if ((buf_itemdata = malloc(sizeof(itemdata_t) * NUM_ITEM)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ƥγƼǡɤ߹ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMDATA];

		file_read(buf, FILEPOS_ITEMDATA + SIZEOF_ITEMDATA * i, SIZEOF_ITEMDATA);

		/* ͷϥǡ */
		buf_itemdata[i].type                 = buf[ 0] & 0x0f;
		buf_itemdata[i].effect_type          = buf[27];
		buf_itemdata[i].price                = (buf[29] << 8) | buf[28];
		buf_itemdata[i].attack_defense_power = buf[20];
		buf_itemdata[i].evade_rate           = ((buf[26] & 0x0f) >> 0) * 10;
		buf_itemdata[i].magic_evade_rate     = ((buf[26] & 0xf0) >> 4) * 10;

		/* ɸ/̿Ψǡ */
		switch (buf_itemdata[i].type)
		{
		case ITEM_TYPE_TOOL:
			buf_itemdata[i].magic_defense_power = buf_itemdata[i].hit_rate = buf[21];
			break;

		case ITEM_TYPE_WEAPON:
			buf_itemdata[i].magic_defense_power = 0x00;
			buf_itemdata[i].hit_rate            = buf[21];
			break;

		default:
			buf_itemdata[i].magic_defense_power = buf[21];
			buf_itemdata[i].hit_rate            = 0x00;
			break;
		}

		/* 饯οͷϥǡ */
		buf_itemdata[i].increase_strength    = (((buf[16] & 0x08) == 0x00) ? 1 : -1) * ((buf[16] & 0x07) >> 0);
		buf_itemdata[i].increase_speed       = (((buf[16] & 0x80) == 0x00) ? 1 : -1) * ((buf[16] & 0x70) >> 4);
		buf_itemdata[i].increase_stamina     = (((buf[17] & 0x08) == 0x00) ? 1 : -1) * ((buf[17] & 0x07) >> 0);
		buf_itemdata[i].increase_magic_power = (((buf[17] & 0x80) == 0x00) ? 1 : -1) * ((buf[17] & 0x70) >> 4);

		/* ̥ǡ */
		buf_itemdata[i].effect_field  = buf[ 2];
		buf_itemdata[i].effect_status = (buf[11] << 8) | buf[ 9];
		buf_itemdata[i].effect_battle = bits_make32(buf[10], buf[12], buf[13], 0x00);
	
		/* Ѹ̥ǡ */
		switch (buf_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
		case ITEM_TYPE_ARMOR:
		case ITEM_TYPE_SHIELD:
		case ITEM_TYPE_HAT:
		case ITEM_TYPE_RELIC:
			buf_itemdata[i].effect_no        = buf[18] & 0x3f;
			buf_itemdata[i].learn_magic_no   = buf[ 4];
			buf_itemdata[i].learn_magic_rate = buf[ 3];
			break;

		default:
			buf_itemdata[i].effect_no        = 0x00;
			buf_itemdata[i].learn_magic_no   = 0x00;
			buf_itemdata[i].learn_magic_rate = 0x00;
			break;
		}

		/* °ϥǡ */
		buf_itemdata[i].elem_half   = 0x00;
		buf_itemdata[i].elem_none   = 0x00;
		buf_itemdata[i].elem_absorb = 0x00;
		buf_itemdata[i].elem_double = 0x00;
		switch (buf_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			buf_itemdata[i].elem_half  = buf[15];
			break;

		case ITEM_TYPE_ARMOR:
		case ITEM_TYPE_SHIELD:
		case ITEM_TYPE_HAT:
		case ITEM_TYPE_RELIC:
			buf_itemdata[i].elem_half   = buf[15];
			buf_itemdata[i].elem_none   = buf[22];
			buf_itemdata[i].elem_absorb = buf[23];
			buf_itemdata[i].elem_double = buf[24];
			break;

		default:
			break;
		}

		/* Ѳϥǡ */
		buf_itemdata[i].state_defense = 0x00;
		buf_itemdata[i].state_apriori = 0x00;
		buf_itemdata[i].state_effect  = 0x00;
		switch (buf_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
		case ITEM_TYPE_ARMOR:
		case ITEM_TYPE_SHIELD:
		case ITEM_TYPE_HAT:
		case ITEM_TYPE_RELIC:
			buf_itemdata[i].state_defense = bits_make32(buf[ 6], buf[ 7], 0x00, 0x00);
			buf_itemdata[i].state_apriori = bits_make32(0x00, buf[25], buf[ 8], 0x00);
			break;

		case ITEM_TYPE_ITEM:
			buf_itemdata[i].state_effect  = bits_make32(buf[21], buf[22], buf[23], buf[24]);
			break;

		default:
			break;
		}

		/* ե饰ϥǡ */
		buf_itemdata[i].bits_equip_chr         = (buf[ 2] << 8) | buf[ 1];
		buf_itemdata[i].flag_throw             = ((buf[ 0] & 0x10) != 0x00);
		buf_itemdata[i].flag_using_effect      = ((buf[ 0] & 0x20) != 0x00);	/* buf[18] & 0x80 */
		buf_itemdata[i].flag_additional_effect = ((buf[18] & 0x40) != 0x00);
		buf_itemdata[i].flag_runic             = ((buf[19] & 0x80) != 0x00);
		buf_itemdata[i].flag_2handed           = ((buf[19] & 0x40) != 0x00);
		buf_itemdata[i].flag_backrow           = ((buf[19] & 0x20) != 0x00);
		buf_itemdata[i].flag_swordtech         = ((buf[19] & 0x02) != 0x00);
		buf_itemdata[i].flag_range_all         = (buf[14] == 0x04);	/* 0x04̰ */
		buf_itemdata[i].direction              = ((buf[14] & 0x40) != 0x00) ? 1 : 0;
		buf_itemdata[i].range                  = ((buf[14] & 0x08) != 0x00) ? 1 : 0;
		buf_itemdata[i].change_direction       = ((buf[14] & 0x02) == 0x00 && (buf[14] & 0x01) != 0x00) ? 1 : 0;
		buf_itemdata[i].change_range           = ((buf[14] & 0x20) != 0x00 && (buf[14] & 0x01) != 0x00) ? 1 : 0;
	}

	/* ꥢƥǡɤ߹ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMCOL];

		file_read(buf, FILEPOS_ITEMCOL + SIZEOF_ITEMCOL * i, SIZEOF_ITEMCOL);

		if (i >= 256)
		{
			buf_itemdata[i].col_monster_no = ((buf[1] & 0xf0) << 4) | buf[0];
			buf_itemdata[i].col_item_no    = ((buf[1] & 0x0f) << 8) | buf[2];
		}
		else
		{
			buf_itemdata[i].col_monster_no = buf[0];
			buf_itemdata[i].col_item_no    = buf[2];
		}
		buf_itemdata[i].col_flag_erase = (buf[3] != 0x00);
	}

	return buf_itemdata;
}

int item_writeparam(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥοͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥοͷϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"   "
				" / Ψ  Ψ   Hit%% "
				"     ®  \n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :  %5ld, %5d, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price,
				ptr_itemdata[i].attack_defense_power,
				ptr_itemdata[i].evade_rate,
				ptr_itemdata[i].magic_defense_power,
				ptr_itemdata[i].magic_evade_rate,
				ptr_itemdata[i].hit_rate,
				ptr_itemdata[i].increase_strength,
				ptr_itemdata[i].increase_speed,
				ptr_itemdata[i].increase_stamina,
				ptr_itemdata[i].increase_magic_power);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeelem(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥ°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥ°ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"Ⱦ/°     "
				"̵°          "
				"ۼ°          "
				"°\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			/* ° */
			str_printelem(fp, ptr_itemdata[i].elem_half, ELEM_ALL);
			break;

		case ITEM_TYPE_ARMOR:
		case ITEM_TYPE_SHIELD:
		case ITEM_TYPE_HAT:
		case ITEM_TYPE_RELIC:
			/* Ⱦ° */
			str_printelem(fp, ptr_itemdata[i].elem_half, ELEM_ALL);
			fprintf(fp, "  ");

			/* ̵° */
			str_printelem(fp, ptr_itemdata[i].elem_none, ELEM_ALL);
			fprintf(fp, "  ");

			/* ۼ° */
			str_printelem(fp, ptr_itemdata[i].elem_absorb, ELEM_ALL);
			fprintf(fp, "  ");

			/* ° */
			str_printelem(fp, ptr_itemdata[i].elem_double, ELEM_ALL);
			break;

		default:
			break;
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writestate(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥΥե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥΥե饰ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"Ѳ1    2                "
				"Ѳ2    3\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			/* Ѳؤ */
			str_printstate(fp, ptr_itemdata[i].state_defense, STATE_ALL & ~STATE_ASSISTMAG & ~STATE_OTHER);
			fprintf(fp, " ");

			/* ǽ餫ŬѤѲ */
			fprintf(fp, "                 ");
			str_printstate(fp, ptr_itemdata[i].state_apriori, STATE_ASSISTMAG);
			break;

		case ITEM_TYPE_ARMOR:
		case ITEM_TYPE_SHIELD:
		case ITEM_TYPE_HAT:
		case ITEM_TYPE_RELIC:
			/* Ѳؤ */
			str_printstate(fp, ptr_itemdata[i].state_defense, STATE_ALL & ~STATE_ASSISTMAG & ~STATE_OTHER);
			fprintf(fp, " ");

			/* ǽ餫ŬѤѲ */
			str_printstate(fp, ptr_itemdata[i].state_apriori, STATE_ALL & ~STATE_PERMANENT & ~STATE_OTHER);
			break;

		case ITEM_TYPE_ITEM:
			str_printstate(fp, ptr_itemdata[i].state_effect, STATE_ALL);
			break;

		default:
			break;
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeflag(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "..";
	const char *str_direction[2][2] =
	{
		{"̣",	"̣/Ũ"},
		{"Ũ",		"Ũ/̣"},
	};
	const char *str_range[2][2] =
	{
		{"ñ",	"ñ/"},
		{"",	"/ñ"},
	};
	const char *str_range_all = "Ụ̃";
	const char *str_chrmain[] =
	{
		"Ti",	/* ƥ */
		"Lo",	/* å */
		"Ca",	/*  */
		"Sh",	/* ɥ */
		"Ed",	/* ɥ */
		"Ma",	/* ޥå */
		"Ce",	/* ꥹ */
		"St",	/* ȥ饴 */
		"Re",	/*  */
		"Se",	/* åĥ */
		"Mo",	/* ⥰ */
		"Ga",	/*  */
		"Go",	/*  */
		"Um",	/* ޥ */
		"Im",	/* å */
		"Me",	/*  */
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥΥե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥΥե饰ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"ꤲ  ξ  ɬ  "
				"ǽ饯                               "
				"ϰ\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :", i, i, item_getname(i));

		/* Ƽե饰 */
		fprintf(fp, "  %2s   %2s   %2s   %2s   %2s  ",
				str_getflag(ptr_itemdata[i].flag_throw),
				str_getflag(ptr_itemdata[i].flag_backrow),
				str_getflag(ptr_itemdata[i].flag_2handed),
				str_getflag(ptr_itemdata[i].flag_runic),
				str_getflag(ptr_itemdata[i].flag_swordtech));

		/* ǽ饯 */
		for (j = 0; j < 16; j++)
			fprintf(fp, " %-2s", ((ptr_itemdata[i].bits_equip_chr & (1U << (j))) != 0x00) ? str_chrmain[j] : str_nothing);
		fprintf(fp, "  ");

		/* ϰ */
		if (ptr_itemdata[i].flag_range_all)
			fprintf(fp, "%-17s", str_range_all);
		else
		{
			fprintf(fp, "%-7s %-9s",
					str_direction[ptr_itemdata[i].direction][ptr_itemdata[i].change_direction],
					str_range[ptr_itemdata[i].range][ptr_itemdata[i].change_range]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeeffect(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥθ̥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥθ̥ǡϤ */
	fprintf(fp, "No.       ̾               "
				" ɲ  "
				"ü          "
				"ˡ          "
				"Ψ\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		const char *ptr_name_effect = str_nothing;
		const char *ptr_name_learn  = str_nothing;

		fprintf(fp, "%03d(%03X): %-16s :  %2s", i, i, item_getname(i),
					str_getflag(ptr_itemdata[i].flag_using_effect));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
		case ITEM_TYPE_ARMOR:
		case ITEM_TYPE_SHIELD:
		case ITEM_TYPE_HAT:
		case ITEM_TYPE_RELIC:
			if (ptr_itemdata[i].flag_using_effect || ptr_itemdata[i].flag_additional_effect)
				ptr_name_effect = mag_getname(ptr_itemdata[i].effect_no);
			if (ptr_itemdata[i].learn_magic_rate > 0)
				ptr_name_learn = mag_getname(ptr_itemdata[i].learn_magic_no);

			fprintf(fp, "   %2s   %-16s, %-16s,  %5d",
					str_getflag(ptr_itemdata[i].flag_additional_effect),
					ptr_name_effect,
					ptr_name_learn,
					ptr_itemdata[i].learn_magic_rate);
			break;

		default:
			break;
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writecol(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ƥθ̥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ƥθ̥ǡϤ */
	fprintf(fp, "No.       ̾               "
				"ꥢƥ      󥹥\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : %-16s, %-16s\n", i, i, item_getname(i),
				item_getname(ptr_itemdata[i].col_item_no),
				mon_getname(ptr_itemdata[i].col_monster_no));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
