/*
 *	ff6magic.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff6decode.h"
#include "ff6magic.h"

#define FILEPOS_MAGNAME1		0x79bccc	/* SNES: 0x2d7400 */
#define FILEPOS_MAGDATA			0x62bc34	/* SNES: 0x046ac0 */
#define FILEPOS_MAGNAME2		0x79c55e
#define FILEPOS_ESPNAME			0x79c570
#define FILEPOS_ESPDATA			0x62A480	/* SNES: 0x26f567 */
#define NUM_MAG_SHORT			54
#define NUM_MAG_MIDDLE			27
#define NUM_MAG_LONG			175
#define NUM_MAG_ESP1			27
#define NUM_MAG_ESP2			8
#define NUM_MAG					(NUM_MAG_SHORT + NUM_MAG_MIDDLE + NUM_MAG_LONG + NUM_MAG_ESP1 + NUM_MAG_ESP2)	/* = 291 */
#define NUM_MAGNAME2			3
#define NUM_ESP					31			/* SNES: 27 */
#define NUM_ESPNAME2			4
#define NUM_ESPNAME1			(NUM_ESP - NUM_ESPNAME2)
#define SIZEOF_MAGNAME_SHORT	5
#define SIZEOF_MAGNAME_MIDDLE	8
#define SIZEOF_MAGNAME_LONG		8
#define SIZEOF_MAGNAME_ESP		9
#define SIZEOF_ESPNAME			8
#define SIZEOF_ESPDATA			11
#define SIZEOF_MAGDATA			14
#define REGION_SHORT_BEGIN		0
#define REGION_SHORT_END		(REGION_SHORT_BEGIN + NUM_MAG_SHORT - 1)	/* =  53 */
#define REGION_MIDDLE_BEGIN		(REGION_SHORT_END + 1)						/* =  54 */
#define REGION_MIDDLE_END		(REGION_MIDDLE_BEGIN + NUM_MAG_MIDDLE - 1)	/* =  80 */
#define REGION_LONG_BEGIN		(REGION_MIDDLE_END + 1)						/* =  81 */
#define REGION_LONG_END			(REGION_LONG_BEGIN + NUM_MAG_LONG - 1)		/* = 255 */
#define REGION_ESP_BEGIN		(REGION_LONG_END + 1)						/* = 256 */
#define REGION_ESP_END			(REGION_ESP_BEGIN + NUM_MAG_ESP1 - 1)		/* = 282 */
#define MSGID_MAGNAME1			5101	/* long */		/* short: 667 */
#define MSGID_MAGNAME2			5392	/* long */		/* short: 958 */
#define MSGID_ESPNAME1			721		/* short */
#define MSGID_ESPNAME2			2651	/* short */

static void mag_setname(void);
static void mag_setespname(void);

static char tbl_magname[NUM_MAG][256];	/* ˡ̾Ǽơ֥ */
static char tbl_espname[NUM_ESP][256];	/* ä̾Ǽơ֥ */
static int flag_setmagname = 0;			/* tbl_magname˥ǡǼƤ뤫ɤ */
static int flag_setespname = 0;			/* tbl_espname˥ǡǼƤ뤫ɤ */

const char *mag_getname(int index)
{
	/* å */
	if (index < 0 || NUM_MAG <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmagname)
		mag_setname();

	return tbl_magname[index];
}

const char *mag_getespname(int index)
{
	/* å */
	if (index < 0 || NUM_ESP <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setespname)
		mag_setespname();

	return tbl_espname[index];
}

static void mag_setname(void)
{
	int i;

	/* ˡ̾ǡ */
	for (i = 0; i < NUM_MAG; i++)
		*tbl_magname[i] = '\0';

	/* ˡ̾ǡɤ߹ */
	for (i = 0; i < NUM_MAG; i++)
	{
		dec_msg(tbl_magname[i], sizeof(tbl_magname[i]),
				((i >= NUM_MAG - NUM_MAG_ESP2) ? MSGID_MAGNAME2 + NUM_MAG_ESP2 - NUM_MAG : MSGID_MAGNAME1) + i,
				FLAG_KANJI | FLAG_LONG);
	}

	flag_setmagname = 1;
}

static void mag_setespname(void)
{
	int i;

	/* ˡ̾ǡ */
	for (i = 0; i < NUM_ESP; i++)
		*tbl_espname[i] = '\0';

	/* ä̾ǡɤ߹ */
	for (i = 0; i < NUM_ESP; i++)
	{
		dec_msg(tbl_espname[i], sizeof(tbl_espname[i]),
				((i >= NUM_ESPNAME1) ? MSGID_ESPNAME2 - NUM_ESPNAME1 : MSGID_ESPNAME1) + i,
				FLAG_KANA | FLAG_SHORT);
	}

	flag_setespname = 1;
}

magdata_t *mag_readdata(void)
{
	int i;
	magdata_t *buf_magdata = NULL;

	/* ˡγƼǡǼХåեݤ */
	if ((buf_magdata = malloc(sizeof(magdata_t) * NUM_MAG)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ˡγƼǡɤ߹ */
	for (i = 0; i < NUM_MAG; i++)
	{
		unsigned char buf[SIZEOF_MAGDATA];

		file_read(buf, FILEPOS_MAGDATA + SIZEOF_MAGDATA * i, SIZEOF_MAGDATA);

		/* ͷϥǡ */
		buf_magdata[i].effect_type  = buf[9];
		buf_magdata[i].magic_point  = buf[5];

		/* °Ѳϥǡ */
		buf_magdata[i].elem_effect  = buf[1];
		buf_magdata[i].state_effect = bits_make32(buf[10], buf[11], buf[12], buf[13]);

		/* ե饰ϥǡ */
		buf_magdata[i].flag_runic            = ((buf[3] & 0x08) != 0x00);
		buf_magdata[i].flag_reflect          = ((buf[3] & 0x02) == 0x00);
		buf_magdata[i].flag_cure             = ((buf[4] & 0x01) != 0x00);
		buf_magdata[i].flag_absorb           = ((buf[4] & 0x02) != 0x00);
		buf_magdata[i].flag_ignore_defense   = ((buf[2] & 0x20) != 0x00);
		buf_magdata[i].flag_ignore_evade     = ((buf[4] & 0x20) != 0x00);
		buf_magdata[i].flag_kill             = ((buf[2] & 0x02) != 0x00);
		buf_magdata[i].flag_physical         = ((buf[2] & 0x01) != 0x00);
		buf_magdata[i].flag_state_reset      = ((buf[4] & 0x04) != 0x00);
		buf_magdata[i].flag_state_reverse    = ((buf[4] & 0x08) != 0x00);
		buf_magdata[i].flag_ratio            = ((buf[4] & 0x80) != 0x00);
		buf_magdata[i].flag_level            = ((buf[4] & 0x40) != 0x00);
		buf_magdata[i].flag_range_all        = (buf[0] == 0x04);	/* 0x04̰ */
		buf_magdata[i].flag_roulette         = ((buf[0] & 0x80) != 0x00);
		buf_magdata[i].direction             = ((buf[0] & 0x40) != 0x00) ? 1 : 0;
		buf_magdata[i].range                 = ((buf[0] & 0x08) != 0x00) ? 1 : 0;
		buf_magdata[i].change_direction      = ((buf[0] & 0x02) == 0x00 && (buf[0] & 0x01) != 0x00) ? 1 : 0;
		buf_magdata[i].change_range          = ((buf[0] & 0x20) != 0x00 && (buf[0] & 0x01) != 0x00) ? 1 : 0;

		/* buf[6]buf[8]β */
		buf_magdata[i].effect_power = 0;
		buf_magdata[i].effect_rate  = 0;
		buf_magdata[i].effect_level = 0;
		buf_magdata[i].hit_rate     = 0;

		if (buf_magdata[i].flag_ratio)
			buf_magdata[i].effect_rate  = buf[6];
		else
			buf_magdata[i].effect_power = buf[6];

		if (buf_magdata[i].flag_level)
			buf_magdata[i].effect_level = buf[8];
		else
			buf_magdata[i].hit_rate     = buf[8];

	}

	return buf_magdata;
}

espdata_t *mag_readesp(void)
{
	int i, j;
	espdata_t *buf_espdata = NULL;

	/* äγƼǡǼХåեݤ */
	if ((buf_espdata = malloc(sizeof(espdata_t) * NUM_MAG)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* äγƼǡɤ߹ */
	for (i = 0; i < NUM_ESP; i++)
	{
		unsigned char buf[SIZEOF_ESPDATA];

		file_read(buf, FILEPOS_ESPDATA + SIZEOF_ESPDATA * i, SIZEOF_ESPDATA);

		/* ͷϥǡ */
		buf_espdata[i].lvup_bonus = buf[10];

		for (j = 0; j < 5; j++)
		{
			buf_espdata[i].learn_magic_rate[j] = buf[j * 2];
			buf_espdata[i].learn_magic_no[j]   = buf[j * 2 + 1];
		}
	}

	return buf_espdata;
}

int mag_writeparam(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡοͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡοͷϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"MP    Ψ   Hit%% ٥   ü\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, (ptr_magdata[i].effect_rate != 0) ?
					"%03d(%03X): %-18s :  %5d, %5d,%3d/16, %5d, %5d, %5X\n" :
					"%03d(%03X): %-18s :  %5d, %5d, %5d, %5d, %5d, %5X\n",
				i, i, mag_getname(i),
				ptr_magdata[i].magic_point,
				ptr_magdata[i].effect_power,
				ptr_magdata[i].effect_rate,
				ptr_magdata[i].hit_rate,
				ptr_magdata[i].effect_level,
				ptr_magdata[i].effect_type);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeelem(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡ°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡ°ϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"°\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s : ", i, i, mag_getname(i));

		/* ° */
		str_printelem(fp, ptr_magdata[i].elem_effect, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writestate(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_state[2][2] =
	{
		{"Ѳ",	""},
		{"ȿž",	"ȿž"},
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡξѲϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				"  "
				"Ѳ1        2                3                4\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%03X): %-18s : %4s  ", i, i, mag_getname(i),
				str_state[(ptr_magdata[i].flag_state_reverse) ? 1 : 0][(ptr_magdata[i].flag_state_reset) ? 1 : 0]);

		/* Ѳ */
		str_printstate(fp, ptr_magdata[i].state_effect, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeflag(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_type[2] = {"ʪ", "ˡ"};
	const char *str_direction[2][2] =
	{
		{"̣",	"̣/Ũ"},
		{"Ũ",		"Ũ/̣"},
	};
	const char *str_range[2][2] =
	{
		{"ñ",	"ñ/"},
		{"",	"/ñ"},
	};
	const char *str_range_all = "Ụ̃";
	const char *str_roulette  = "롼å";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ˡΥե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ˡΥե饰ϥǡϤ */
	fprintf(fp, "No.       ̾                 "
				" ȿ ̵ ̵ Ψ  "
				"  ϰ\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		/* Ƽե饰 */
		fprintf(fp, "%03d(%03X): %-18s :  %2s   %2s   %2s   %2s   %2s   %4s  ", i, i, mag_getname(i),
				str_getflag(ptr_magdata[i].flag_runic),
				str_getflag(ptr_magdata[i].flag_reflect),
				str_getflag(ptr_magdata[i].flag_ignore_defense),
				str_getflag(ptr_magdata[i].flag_ignore_evade),
				str_getflag(ptr_magdata[i].flag_ratio),
				str_type[(ptr_magdata[i].flag_physical) ? 0 : 1]);

		/* ϰ */
		if (ptr_magdata[i].flag_range_all)
			fprintf(fp, "%-17s", str_range_all);
		else if (ptr_magdata[i].flag_roulette)
			fprintf(fp, "%-17s", str_roulette);
		else
		{
			fprintf(fp, "%-7s %-9s",
					str_direction[ptr_magdata[i].direction][ptr_magdata[i].change_direction],
					str_range[ptr_magdata[i].range][ptr_magdata[i].change_range]);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeesp(const char *ptr_filename, const espdata_t *ptr_espdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing_8  = "--------";
	const char *str_nothing_20 = "--------------------";
	const char *str_lvup_bonus[17] =
	{
		"HP  +10%", "HP  +30%", "HP  +50%",
		"MP  +10%", "MP  +30%", "MP  +50%",
		"HP +100%", "",         "",
		"    +1", "    +2",
		"®    +1", "®    +2",
		"  +1", "  +2",
		"  +1", "  +2",
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_espdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* äγƼǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* äγƼǡϤ */
	fprintf(fp, "No.      ̾               "
				"ܡʥ   ˡ\n");
	for (i = 0; i < NUM_ESP; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : %-8s,  ", i, i,
				mag_getespname(i),
				(ptr_espdata[i].lvup_bonus < 17) ? str_lvup_bonus[ptr_espdata[i].lvup_bonus] : str_nothing_8);

		for (j = 0; j < 5; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");

			if (ptr_espdata[i].learn_magic_no[j] == 0xff)
				fprintf(fp, "%-20s", str_nothing_20);
			else
			{
				fprintf(fp, "%-16s%3d%%",
						mag_getname(ptr_espdata[i].learn_magic_no[j]),
						ptr_espdata[i].learn_magic_rate[j]);
			}
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

