/*
 *	ff6map.c
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff6decode.h"
#include "ff6map.h"

#define FILEPOS_MAPDATA				0x6e1880	/* SNES: 0x2d8f00 */
#define FILEPOS_MAP_MONENC			0x712870	/* SNES: 0x0f5600 */
#define NUM_MAPSUBNAME				75
#define SIZEOF_MAPDATA				33
#define SIZEOF_MAP_MONENC			1
#define SIZEOF_MAPPOS_OFFSET		2
#define MSGID_MAPNAME				2034	/* short */		/* long: 5497 */
#define MSGID_MAPSUBNAME			5475	/* long */		/* short: 966 */

static void map_setname(void);
static void map_setsubname(void);

static char tbl_mapname[NUM_MAP][256];				/* マップの名前が格納されるテーブル */
static char tbl_mapsubname[NUM_MAPSUBNAME][256];	/* マップの補助用の名前が格納されるテーブル */
static int flag_setmapname = 0;						/* tbl_mapnameにデータが格納されているかどうか */
static int flag_setmapsubname = 0;					/* tbl_mapsubnameにデータが格納されているかどうか */

const char *map_getname(int index)
{
	/* 引数をチェックする */
	if (index < 0 || NUM_MAP <= index)
	{
		assert(0);
		return "";
	}

	/* テーブルに値が設定されていない場合は、値を設定する */
	if (!flag_setmapname)
		map_setname();

	return tbl_mapname[index];
}

const char *map_getsubname(int index)
{
	/* 引数をチェックする */
	if (index < 0 || NUM_MAPSUBNAME <= index)
	{
		assert(0);
		return "";
	}

	/* テーブルに値が設定されていない場合は、値を設定する */
	if (!flag_setmapsubname)
		map_setsubname();

	return tbl_mapsubname[index];
}

static void map_setname(void)
{
	int i;
	char *ptr;

	/* マップの名前データを初期化する */
	for (i = 0; i < NUM_MAP; i++)
		*tbl_mapname[i] = '\0';

	/* マップの名前データを読み込む */
	for (i = 0; i < NUM_MAP; i++)
	{
		dec_msg(tbl_mapname[i], sizeof(tbl_mapname[i]), MSGID_MAPNAME + i, FLAG_KANA | FLAG_SHORT);

		/* 改行を記号に置き換える */
		while ((ptr = strchr(tbl_mapname[i], '\n')) != NULL)
			*ptr = ' ';
	}

	flag_setmapname = 1;
}

static void map_setsubname(void)
{
	int i;

	/* マップの補助用の名前データを初期化する */
	for (i = 0; i < NUM_MAPSUBNAME; i++)
		*tbl_mapsubname[i] = '\0';

	/* マップの補助用の名前データを読み込む */
	for (i = 1; i < NUM_MAPSUBNAME; i++)
		dec_msg(tbl_mapsubname[i], sizeof(tbl_mapsubname[i]), MSGID_MAPSUBNAME + i - 1, FLAG_KANJI | FLAG_LONG);

	flag_setmapsubname = 1;
}

mapdata_t *map_readdata(void)
{
	int i;
	mapdata_t *buf_mapdata = NULL;

	/* マップデータを格納するバッファを確保する */
	if ((buf_mapdata = malloc(sizeof(mapdata_t) * NUM_MAP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* マップデータを読み込む */
	for (i = 0; i < NUM_MAP; i++)
	{
		unsigned char buf[SIZEOF_MAPDATA];

		file_read(buf, FILEPOS_MAPDATA + SIZEOF_MAPDATA * i, SIZEOF_MAPDATA);

		/* 数値系データ */
		buf_mapdata[i].sub_no         = buf[ 0];
		buf_mapdata[i].background_no  = buf[ 2] & 0x7f;
		buf_mapdata[i].music_no       = buf[28];

		/* フラグ系データ */
		buf_mapdata[i].flag_encounter = ((buf[5] & 0x80) != 0x00);
	}

	/* モンスター遭遇データを読み込む */
	for (i = 0; i < NUM_MAP; i++)
	{
		unsigned char buf[SIZEOF_MAP_MONENC];

		file_read(buf, FILEPOS_MAP_MONENC + SIZEOF_MAP_MONENC * i, SIZEOF_MAP_MONENC);
		buf_mapdata[i].enc_no = buf[0];
	}

	return buf_mapdata;
}
