/*
 *	ff6monster.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 *		Copyright (C) 2008 Inu
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "errmes.h"
#include "file.h"
#include "ff6decode.h"
#include "ff6character.h"
#include "ff6monster.h"

#define FILEPOS_MONNAME			0x722c94	/* SNES: 0x0fc800 */
#define FILEPOS_MONDATA			0x7379a6	/* SNES: 0x0f0000 */
#define FILEPOS_MONITEM			0x71bb9c	/* SNES: 0x0f3000 */
#define FILEPOS_MONRAGNA		0x62b23c	/* SNES: 0x047f40 */
#define FILEPOS_MONCTRL			0x726b86	/* SNES: 0x0f3d00 */
#define FILEPOS_MONSKETCH		0x5f78ee	/* SNES: 0x0f4300 */
#define FILEPOS_MONRAGE			0x5fce18	/* SNES: 0x0f4600 */
#define FILEPOS_MONACT_ADDR		0x727fd8	/* SNES: 0x0f8400 */
#define FILEPOS_MONACT			0x728326	/* SNES: 0x0f8700 */
#define FILEPOS_MONTECHNAME		0x72ca18	/* SNES: 0x0fd400 */
#define FILEPOS_MONTECHWPN		0x71c238	/* SNES: 0x0f37c0 */
#define FILEPOS_MONMES_ADDR		0x53b4cc	/* SNES: 0x0fe000 */
#define FILEPOS_MONMES			0x7fe400	/* SNES: 0x0fe400 */
#define FILEPOS_MONPARTY		0x61c33c	/* SNES: 0x0f6200 */
#define FILEPOS_MONLEARNGAIN	0x79f400	/* SNES: 0x10fe00 */
#define FILEPOS_MONPARTY2		0x61b928	/* SNES: 0x0f5900 */
#define FILEPOS_EVENTMES_ADDR	0x79933a	/* SNES: 0x10fc00 */
#define FILEPOS_EVENTMES		0x67d200	/* SNES: 0x10d200 */
#define FILEPOS_EVENTDATA_ADDR	0x679800	/* SNES: 0x109800 */
#define FILEPOS_EVENTDATA		0x679842	/* SNES: 0x109842 */
#define FILEPOS_MONENCRND		0x711190	/* SNES: 0x0f4800 */
#define FILEPOS_MONENCFIXED		0x7ff88e	/* SNES: 0x0f5000 */
#define FILEPOS_MONMAP			0x712a70	/* SNES: 0x0f5400 */
#define NUM_MON					423			/* SNES: 384 */
#define NUM_MONTECH				NUM_MON
#define NUM_MONMES				160			/* SNES: 142 */
#define NUM_MONRAGNA			26
#define NUM_MONPARTY			645			/* SNES: 576 */
#define NUM_MONENCRND			256
#define NUM_MONENCFIXED			256
#define NUM_MONENC				(NUM_MONENCRND + NUM_MONENCFIXED)	/* = 512 */
#define NUM_MONMAP				2
#define NUM_MONMAPAREA			8
#define NUM_EVENTMES			256
#define NUM_EVENTDATA			33
#define SIZEOF_MONNAME			8
#define SIZEOF_MONDATA			32
#define SIZEOF_MONITEM			4
#define SIZEOF_MONRAGNA			4
#define SIZEOF_MONCTRL			4
#define SIZEOF_MONSKETCH		2
#define SIZEOF_MONRAGE			2
#define SIZEOF_MONTECHNAME		8
#define SIZEOF_MONTECHWPN		1
#define SIZEOF_MONLEARNGAIN		1
#define SIZEOF_MONPARTY			15
#define SIZEOF_MONPARTY2		4
#define SIZEOF_MONENCRND		8
#define SIZEOF_MONENCFIXED		4
#define SIZEOF_MONMAP			4
#define SIZEOF_OFFSETADDR		2
#define SIZEOF_MONACT_TBL		(0x72c9e8 - FILEPOS_MONACT)			/* = 18114 */
#define MSGID_MONNAME			1039	/* short */		/* long: 5670 */
#define MSGID_MONTECHNAME		3262	/* long */
#define MSGID_MONMES			3685	/* long */
#define MSGID_EVENTMES			4014	/* long */

static const char *mon_gettechname(int index);
static void mon_setname(void);
static void mon_settechname(void);
static void mon_setmes(void);
static void mon_seteventdata(void);
static const char *mon_gettarget(int index);
static const char *mon_getmes(int index);
static const char *mon_geteventmes(int index);
static int mon_geteventdata(int index1, int index2);
static unsigned char *mon_printactcnd(FILE *stream, unsigned char *buf_cnd);
static void mon_printact(FILE *stream, int index, int monster_no);

static char tbl_monname[NUM_MON][256];			/* 󥹥̾Ǽơ֥ */
static char tbl_montechname[NUM_MONTECH][256];	/* ֤Ȥפ̾Ǽơ֥ */
static char tbl_monmes[NUM_MONMES][256];		/* 󥹥ΥåǼơ֥ */
static char tbl_eventmes[NUM_EVENTMES][256];	/* ٥ȤΥåǼơ֥ */
static int tbl_eventdata[NUM_EVENTDATA][256];	/* ٥ȥǡǼơ֥ */
static int flag_setmonname      = 0;			/* tbl_monname˥ǡǼƤ뤫ɤ */
static int flag_setmontechname  = 0;			/* tbl_montechname˥ǡǼƤ뤫ɤ */
static int flag_setmoneventmes  = 0;			/* tbl_monmes, tbl_eventmes˥ǡǼƤ뤫ɤ */
static int flag_setmoneventdata = 0;			/* tbl_eventdata˥ǡǼƤ뤫ɤ */

static const int tbl_monevent_skip_byte[21] = {2, 2, 1, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 1, 1, 1, 1, 48, 2, 2};
static const int tbl_monact_skip_byte[16] = {4, 2, 4, 3, 4, 4, 4, 2, 3, 4, 4, 3, 4, 1, 1, 1};

const char *mon_getname(int index)
{
	/* å */
	if (index < 0 || NUM_MON <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmonname)
		mon_setname();

	return tbl_monname[index];
}

static const char *mon_gettechname(int index)
{
	/* å */
	if (index < 0 || NUM_MONTECH <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmontechname)
		mon_settechname();

	return tbl_montechname[index];
}

static const char *mon_getmes(int index)
{
	/* å */
	if (index < 0 || NUM_MONMES <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmoneventmes)
		mon_setmes();

	return tbl_monmes[index];
}

static const char *mon_geteventmes(int index)
{
	/* å */
	if (index < 0 || NUM_EVENTMES <= index)
	{
		assert(0);
		return "";
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmoneventmes)
		mon_setmes();

	return tbl_eventmes[index];
}

static int mon_geteventdata(int index1, int index2)
{
	/* å */
	if (index1 < 0 || NUM_EVENTDATA <= index1 || index2 < 0 || 64 <= index2)
	{
		assert(0);
		return -1;
	}

	/* ơ֥ͤꤵƤʤϡͤꤹ */
	if (!flag_setmoneventdata)
		mon_seteventdata();

	return tbl_eventdata[index1][index2];
}

static const char *mon_gettarget(int index)
{
	const char *ptr = NULL;
	static const char *str_target[] =
	{
		"󥹥1", "󥹥2", "󥹥3", "󥹥4",
		"󥹥5", "󥹥6", "󥹥ʬ", "ʬʳΥ󥹥",
		"󥹥", "ʬʳΥ󥹥ï", "󥹥ï", "ƮǽΥ饯",
		"ƮǽΥ饯ï", "ƮǽΥ󥹥", "ƮǽΥ󥹥ï", "ե쥯֤Υ饯",
		"ե쥯֤Υ饯ï", "ե쥯֤Υ󥹥", "ե쥯֤Υ󥹥ï", "饯",
		"饯ï", "Ǹ˼ʬ˹⤷饯ï", "Ụ̃", "ɸ",
		"1ܤΥ饯", "2ܤΥ饯", "3ܤΥ饯", "4ܤΥ饯",
		"ʬʳỤ̃/ï", "åƥ󥰤줿饯",
	};

	assert(sizeof(str_target) / sizeof(str_target[0]) == 0x4e - 0x30);

	if (index < 0x30)
		ptr = chr_getname(index);
	else if (index < 0x4e)
		ptr = str_target[index - 0x30];

	return ptr;
}

static void mon_setname(void)
{
	int i;

	/* 󥹥̾ǡ */
	for (i = 0; i < NUM_MON; i++)
		*tbl_monname[i] = '\0';

	/* 󥹥̾ǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
		dec_msg(tbl_monname[i], sizeof(tbl_monname[i]), MSGID_MONNAME + i, FLAG_KANA | FLAG_SHORT);

	flag_setmonname = 1;
}

static void mon_settechname(void)
{
	int i;

	/* 󥹥Ρ֤Ȥפ̾ǡ */
	for (i = 0; i < NUM_MONTECH; i++)
		*tbl_montechname[i] = '\0';

	/* 󥹥Ρ֤Ȥפ̾ǡɤ߹ */
	for (i = 0; i < NUM_MONTECH; i++)
		dec_msg(tbl_montechname[i], sizeof(tbl_montechname[i]), MSGID_MONTECHNAME + i, FLAG_KANJI | FLAG_LONG);

	flag_setmontechname = 1;
}

static void mon_setmes(void)
{
	int i;
	char *ptr;

	/* åǡ */
	for (i = 0; i < NUM_MONMES; i++)
		*tbl_monmes[i] = '\0';
	for (i = 0; i < NUM_EVENTMES; i++)
		*tbl_eventmes[i] = '\0';

	/* 󥹥Υåǡɤ߹ */
	for (i = 0; i < NUM_MONMES; i++)
	{
		dec_msg(tbl_monmes[i], sizeof(tbl_monmes[i]), MSGID_MONMES + i, FLAG_KANJI | FLAG_LONG);

		/* Ԥ򵭹֤ */
		while ((ptr = strchr(tbl_monmes[i], '\n')) != NULL)
			*ptr = '/';
	}

	/* ٥ȤΥåǡɤ߹ */
	for (i = 0; i < NUM_EVENTMES; i++)
	{
		dec_msg(tbl_eventmes[i], sizeof(tbl_eventmes[i]), MSGID_EVENTMES + i, FLAG_KANJI | FLAG_LONG);

		/* Ԥ򵭹֤ */
		while ((ptr = strchr(tbl_eventmes[i], '\n')) != NULL)
			*ptr = '/';
	}

	flag_setmoneventmes = 1;
}

static void mon_seteventdata(void)
{
	int i;
	const size_t len_offset = 0x9842;
	const size_t bufsize = 0xd100 - len_offset;
	unsigned char *buf = NULL;
	int tbl_offset[NUM_EVENTDATA];

	/* ٥ȥǡ */
	for (i = 0; i < NUM_EVENTDATA; i++)
		memset(tbl_eventdata[i], 0, sizeof(tbl_eventdata[i]));

	/* ٥ȥǡǼƤ륢ɥ쥹(ؤΥեå)ɤ߹ */
	for (i = 0; i < NUM_EVENTDATA; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_EVENTDATA_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}

	/* ٥ȥǡɤ߹ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read(buf, FILEPOS_EVENTDATA, bufsize);

	for (i = 0; i < NUM_EVENTDATA; i++)
	{
		int index = 0;
		int flag_chunk = 0;
		unsigned char *ptr = buf + (ptrdiff_t)(tbl_offset[i] - len_offset);

		while (*ptr != 0xff || flag_chunk)
		{
			/*
			 * XXX: Ĵ
			 * [03...0b xx yy zz] ϻꤷ饯Фư٥Ȥ¹Ԥ
			 * ư٥Ȥ (FILEPOS_EVENTDATA - len_offest + zzyy) ¹Ԥ
			 */

			switch (*ptr)
			{
			case 0x00:	/* åɽ(ɥ) (00 xx) */
			case 0x01:	/* åɽ(ɥ) (01 xx) */
				ptr++;
				tbl_eventdata[i][index] = *ptr;
				index++;
				break;

			case 0x02:	/*  */
				break;

			/* 饯ꤷƥ٥ȼ¹(03 xx yy zz) */
			/*   xx: о */
			/*   yy: ɥ쥹ؤΥեå() */
			/*   zz: ɥ쥹ؤΥեå() */
			case 0x03:
			case 0x04:
			case 0x05:
			case 0x06:
				ptr += 3;
				break;

			/* ٥ȼ¹(07 xx yy zz) */
			/*   xx:  */
			/*   yy: ɥ쥹ؤΥեå() */
			/*   zz: ɥ쥹ؤΥեå() */
			case 0x07:
			case 0x08:
			case 0x09:
			case 0x0a:
			case 0x0b:
			case 0x0c:
				ptr += 3;
				break;

			case 0x0d:	/* и(0d xx yy zz) */
				ptr += 3;
				break;

			case 0x0e:	/* ϢΥ٥ȳ? */
				flag_chunk = 1;
				break;

			case 0x0f:	/* ϢΥ٥Ȥ¹ */
				flag_chunk = 0;
				break;

			case 0x10:	/* ɥõ */
			case 0x11:	/* ɥɽ */
				break;

			case 0x12:	/*  */
				break;

			case 0x13:	/* (13 xx) */
			case 0x14:	/* (14 xx) */
				ptr++;
				break;
			}

			ptr++;
		}

		for (; index < 64; index++)
			tbl_eventdata[i][index] = -1;
	}

	free(buf);
	buf = NULL;

	flag_setmoneventdata = 1;
}

mondata_t *mon_readdata(void)
{
	int i, j;
	mondata_t *buf_mondata = NULL;
	int tbl_ragna[NUM_MONRAGNA][SIZEOF_MONRAGNA];

	/* 󥹥γƼǡǼХåեݤ */
	if ((buf_mondata = malloc(sizeof(mondata_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥νƥǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONITEM];

		file_read(buf, FILEPOS_MONITEM + SIZEOF_MONITEM * i, SIZEOF_MONITEM);
		for (j = 0; j < SIZEOF_MONITEM; j++)
			buf_mondata[i].item_no[j] = buf[j];
	}

	/* 󥹥ΥƥѲǡɤ߹ */
	for (i = 0; i < NUM_MONRAGNA; i++)
	{
		unsigned char buf[SIZEOF_MONRAGNA];

		file_read(buf, FILEPOS_MONRAGNA + SIZEOF_MONRAGNA * i, SIZEOF_MONRAGNA);
		for (j = 0; j < SIZEOF_MONRAGNA; j++)
			tbl_ragna[i][j] = buf[j];
	}

	/* 󥹥Ρ֤Ĥ׹ưǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONCTRL];

		file_read(buf, FILEPOS_MONCTRL + SIZEOF_MONCTRL * i, SIZEOF_MONCTRL);
		for (j = 0; j < SIZEOF_MONCTRL; j++)
			buf_mondata[i].control_no[j] = buf[j];
	}

	/* 󥹥Ρ֥å׹ưǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONSKETCH];

		file_read(buf, FILEPOS_MONSKETCH + SIZEOF_MONSKETCH * i, SIZEOF_MONSKETCH);
		for (j = 0; j < SIZEOF_MONSKETCH; j++)
			buf_mondata[i].sketch_no[j] = buf[j];
	}

	/* 󥹥Ρ֤Ф׹ưǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i >= 256)
		{
			for (j = 0; j < SIZEOF_MONRAGE; j++)
				buf_mondata[i].rage_no[j] = 0xff;
		}
		else
		{
			unsigned char buf[SIZEOF_MONRAGE];

			file_read(buf, FILEPOS_MONRAGE + SIZEOF_MONRAGE * i, SIZEOF_MONRAGE);
			for (j = 0; j < SIZEOF_MONRAGE; j++)
				buf_mondata[i].rage_no[j] = buf[j];
		}
	}

	/* 󥹥γƼǡɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONDATA];

		file_read(buf, FILEPOS_MONDATA + SIZEOF_MONDATA * i, SIZEOF_MONDATA);

		/* ͷϥǡ */
		buf_mondata[i].level               = buf[16];
		buf_mondata[i].hit_point           = (buf[ 9] << 8) | buf[ 8];
		buf_mondata[i].magic_point         = (buf[11] << 8) | buf[10];
		buf_mondata[i].experience_point    = (buf[13] << 8) | buf[12];
		buf_mondata[i].gil                 = (buf[15] << 8) | buf[14];
		buf_mondata[i].speed               = buf[ 0];
		buf_mondata[i].attack_power        = buf[ 1];
		buf_mondata[i].defense_power       = buf[ 5];
		buf_mondata[i].evade_rate          = buf[ 3];
		buf_mondata[i].magic_attack_power  = buf[ 7];
		buf_mondata[i].magic_defense_power = buf[ 6];
		buf_mondata[i].magic_evade_rate    = buf[ 4];
		buf_mondata[i].hit_rate            = buf[ 2];
		buf_mondata[i].item_metamorph      = (buf[17] & 0xe0) >> 5;
		buf_mondata[i].weapon_no           = buf[26];

		/* ѥ᡼黻ФǤǡ */
		if ((buf_mondata[i].stamina = buf_mondata[i].hit_point / 512 + 16) > 40)
			buf_mondata[i].stamina  = 40;

		/* ü쵻ǡ */
		buf_mondata[i].tech_flag_ignore_evade = ((buf[31] & 0x80) != 0x00);
		buf_mondata[i].tech_flag_damage       = ((buf[31] & 0x40) == 0x00);
		buf_mondata[i].tech_type              = buf[31] & 0x3f;

		/* °ϥǡ */
		buf_mondata[i].elem_none   = buf[24];
		buf_mondata[i].elem_absorb = buf[23];
		buf_mondata[i].elem_double = buf[25];

		/* Ѳϥǡ */
		buf_mondata[i].state_defense = bits_make32(buf[20], buf[21], buf[22], 0x00);
		buf_mondata[i].state_apriori = bits_make32(buf[27], buf[28], buf[29], buf[30]);

		/* ²ϥǡ */
		buf_mondata[i].flag_undead     = ((buf[18] & 0x80) != 0x00);
		buf_mondata[i].flag_imp        = ((buf[18] & 0x40) != 0x00);
		buf_mondata[i].flag_human      = ((buf[18] & 0x10) != 0x00);

		/* ե饰ϥǡ */
		buf_mondata[i].flag_erase_name = ((buf[18] & 0x04) != 0x00);
		buf_mondata[i].flag_mp0        = ((buf[18] & 0x01) != 0x00);
		buf_mondata[i].flag_escape     = ((buf[19] & 0x08) == 0x00);
		buf_mondata[i].flag_escape_h   = ((buf[19] & 0x01) != 0x00);
		buf_mondata[i].flag_suplex     = ((buf[19] & 0x04) == 0x00);
		buf_mondata[i].flag_control    = ((buf[19] & 0x80) == 0x00);
		buf_mondata[i].flag_scan       = ((buf[19] & 0x10) == 0x00);
		buf_mondata[i].flag_sketch     = ((buf[19] & 0x20) == 0x00);
		buf_mondata[i].flag_initiative = ((buf[19] & 0x02) != 0x00);
		/* 18: unused: 0x20, 0x08, 0x02 */
		/* 19: unknown: 0x40, 0x08 */

		/* ƥѲǡ */
		for (j = 0; j < 4; j++)
			buf_mondata[i].ragna_item_no[j] = tbl_ragna[buf[17] & 0x1f][j];
	}

	/* ü쵻λṲ̄ɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONTECHWPN];

		file_read(buf, FILEPOS_MONTECHWPN + SIZEOF_MONTECHWPN * i, SIZEOF_MONTECHWPN);
		buf_mondata[i].tech_weapon_no = buf[0];
	}

	return buf_mondata;
}

monact_t *mon_readact(void)
{
	int i, j;
	monact_t *buf_monact = NULL;
	int tbl_offset[NUM_MON];
	unsigned char tbl_monact[SIZEOF_MONACT_TBL];

	/* 󥹥ιưѥǡǼХåեݤ */
	if ((buf_monact = malloc(sizeof(monact_t) * NUM_MON)) == NULL)
		return NULL;
	for (i = 0; i < NUM_MON; i++)
	{
		buf_monact[i].ptn_num = 0;
		for (j = 0; j < (int)(sizeof(buf_monact[i].ptn) / sizeof(buf_monact[i].ptn[0])); j++)
		{
			buf_monact[i].ptn[j].action_size = 0;
			buf_monact[i].ptn[j].action = NULL;
		}
	}

	/* 󥹥ιưѥǡǼƤ륢ɥ쥹(ؤΥեå)ɤ߹ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_MONACT_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}

	/* 󥹥ιưѥǡɤ߹ */
	file_read(tbl_monact, FILEPOS_MONACT, SIZEOF_MONACT_TBL);

	for (i = 0; i < NUM_MON; i++)
	{
		int ptn_num = 0;
		int separator_num = 0;
		unsigned char *ptr_buf = tbl_monact + tbl_offset[i];
		unsigned char *ptr_mark = ptr_buf;

		while (separator_num < 2)
		{
			if (*ptr_buf < 0xf0)
				ptr_buf++;
			else
			{
				if (*ptr_buf == 0xfe || *ptr_buf == 0xff)
				{
					size_t len_read = ptr_buf - ptr_mark;

					if (len_read > 0)
					{
						if ((buf_monact[i].ptn[ptn_num].action = malloc(len_read)) == NULL)
							err_printf(STR_FATAL_ERROR_IN_MALLOC);
						memmove(buf_monact[i].ptn[ptn_num].action, ptr_mark, len_read);
						buf_monact[i].ptn[ptn_num].action_size = len_read;
						ptn_num++;
					}

					if (*ptr_buf == 0xff)
						separator_num++;

					ptr_mark = ptr_buf + 1;
				}

				ptr_buf += tbl_monact_skip_byte[*ptr_buf - 0xf0];
			}
		}

		buf_monact[i].ptn_num = ptn_num;
	}

	return buf_monact;
}

monparty_t *mon_readparty(void)
{
	int i, j;
	monparty_t *buf_monparty = NULL;

	/* 󥹥ΥѡƥǡǼХåեݤ */
	if ((buf_monparty = malloc(sizeof(monparty_t) * NUM_MONPARTY)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥ΥѡƥγƼǡɤ߹ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY];

		file_read(buf, FILEPOS_MONPARTY + SIZEOF_MONPARTY * i, SIZEOF_MONPARTY);
		buf_monparty[i].bits_exist = buf[1];

		for (j = 0; j < 6; j++)
		{
			buf_monparty[i].monster_no[j] = buf[j + 2];
			if ((buf[14] & (1U << j)) != 0x00)
				buf_monparty[i].monster_no[j] += 0x100;
		}
	}

	/* 󥹥ѡƥˡͥǡɤ߹ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		if (i >= 512)
			buf_monparty[i].magic_learning_gain = 0;
		else
		{
			unsigned char buf[SIZEOF_MONLEARNGAIN];

			file_read(buf, FILEPOS_MONLEARNGAIN + SIZEOF_MONLEARNGAIN * i, SIZEOF_MONLEARNGAIN);
			buf_monparty[i].magic_learning_gain = buf[0];
		}
	}

	/* 󥹥ѡƥΥե饰ϥǡɤ߹ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY2];
		int flag_beastfield = 0;

		file_read(buf, FILEPOS_MONPARTY2 + SIZEOF_MONPARTY2 * i, SIZEOF_MONPARTY2);

		buf_monparty[i].bits_vfx        = buf[0] & 0x0f;
		buf_monparty[i].flag_normal     = ((buf[0] & 0x10) == 0x00);
		buf_monparty[i].flag_backattack = ((buf[0] & 0x20) == 0x00);
		buf_monparty[i].flag_pincer     = ((buf[0] & 0x40) == 0x00);
		buf_monparty[i].flag_sideattack = ((buf[0] & 0x80) == 0x00);

		if ((buf[3] & 0x02) != 0x00)
			flag_beastfield = 0;
		else if (i >= 512)
			flag_beastfield = 0;
		else
		{
			for (j = 0; j < 6; j++)
			{
				if (buf_monparty[i].monster_no[j] < 0x100)
				{
					flag_beastfield = 1;
					break;
				}
			}
		}

		buf_monparty[i].flag_beastfield = flag_beastfield;
	}

	return buf_monparty;
}

monenc_t *mon_readenc(void)
{
	int i, j;
	monenc_t *buf_monenc = NULL;

	/* 󥹥ơ֥ǡǼХåեݤ */
	if ((buf_monenc = malloc(sizeof(monenc_t) * NUM_MONENC)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	for (i = 0; i < NUM_MONENC; i++)
	{
		for (j = 0; j < 4; j++)
			buf_monenc[i].party_no[j] = 0;
	}

	/* ̾󥹥Υ󥹥ơ֥ǡɤ߹ */
	for (i = 0; i < NUM_MONENCRND; i++)
	{
		unsigned char buf[SIZEOF_MONENCRND];

		file_read(buf, FILEPOS_MONENCRND + SIZEOF_MONENCRND * i, SIZEOF_MONENCRND);
		for (j = 0; j < 4; j++)
			buf_monenc[i].party_no[j] = ((buf[j * 2 + 1] & 0x7f) << 8) | buf[j * 2];
	}

	/* 󥹥Υ󥹥ơ֥ǡɤ߹ */
	for (i = NUM_MONENCRND; i < NUM_MONENC; i++)
	{
		unsigned char buf[SIZEOF_MONENCFIXED];

		file_read(buf, FILEPOS_MONENCFIXED + SIZEOF_MONENCFIXED * (i - NUM_MONENCRND), SIZEOF_MONENCFIXED);
		for (j = 0; j < 2; j++)
			buf_monenc[i].party_no[j] = ((buf[j * 2 + 1] & 0x7f) << 8) | buf[j * 2];
	}

	return buf_monenc;
}

monmap_t *mon_readmap(void)
{
	int i, j, x, y;
	long filepos;
	monmap_t *buf_monmap = NULL;

	/* 󥹥ޥåץǡǼХåեݤ */
	if ((buf_monmap = malloc(sizeof(monmap_t) * NUM_MONMAP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󥹥ޥåץǡɤ߹ */
	filepos = FILEPOS_MONMAP;
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < NUM_MONMAPAREA; y++)
		{
			for (x = 0; x < NUM_MONMAPAREA; x++)
			{
				unsigned char buf[SIZEOF_MONMAP];

				filepos += file_read(buf, filepos, SIZEOF_MONMAP);
				for (j = 0; j < 4; j++)
					buf_monmap[i].area[x][y].enc_no[j] = buf[j];
			}
		}
	}

	return buf_monmap;
}

int mon_writeparam(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥οͷϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥οͷϥǡϤ */
	fprintf(fp, "No.       ̾                "
				" LV     HP     MP    EXP    "
				"  ®          ⹶   Hit%%\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :  %3d, %5ld, %5ld, %5ld, %5ld, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d\n", i, i,
				mon_getname(i),
				ptr_mondata[i].level,
				ptr_mondata[i].hit_point,
				ptr_mondata[i].magic_point,
				ptr_mondata[i].experience_point,
				ptr_mondata[i].gil,
				ptr_mondata[i].speed,
				ptr_mondata[i].attack_power,
				ptr_mondata[i].defense_power,
				ptr_mondata[i].evade_rate,
				ptr_mondata[i].magic_attack_power,
				ptr_mondata[i].magic_defense_power,
				ptr_mondata[i].magic_evade_rate,
				ptr_mondata[i].hit_rate);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeelem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥°ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥°ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"̵°          "
				"ۼ°          "
				"°\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));

		/* ̵° */
		str_printelem(fp, ptr_mondata[i].elem_none, ELEM_ALL);
		fprintf(fp, "  ");

		/* ۼ° */
		str_printelem(fp, ptr_mondata[i].elem_absorb, ELEM_ALL);
		fprintf(fp, "  ");

		/* ° */
		str_printelem(fp, ptr_mondata[i].elem_double, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writestate(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ξѲϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ξѲϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"Ѳ1    2                3                 "
				"Ѳ1    2                3                4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));

		/* Ѳؤ */
		str_printstate(fp, ptr_mondata[i].state_defense, STATE_ALL & ~STATE_OTHER);
		fprintf(fp, "  ");

		/* ǽ餫ŬѤѲ */
		str_printstate(fp, ptr_mondata[i].state_apriori, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeflag(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;
	const char *str_escape_h = "";
	const char *str_montype[] = 
	{
		"ǥå",
		"ʹ",
		"å",
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Υե饰ϥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Υե饰ϥǡϤ */
	fprintf(fp, "No.       ̾               "
				"ƨ MP    ²\n");
	for (i = 0; i < NUM_MON; i++)
	{
		/* Ƽե饰 */
		fprintf(fp, "%03d(%03X): %-16s :  %2s   %2s   %2s   %2s   %2s", i, i, mon_getname(i),
				(ptr_mondata[i].flag_escape && ptr_mondata[i].flag_escape_h) ?
					str_escape_h :
					str_getflag(ptr_mondata[i].flag_escape),
				str_getflag(ptr_mondata[i].flag_mp0),
				str_getflag(ptr_mondata[i].flag_control),
				str_getflag(ptr_mondata[i].flag_sketch),
				str_getflag(ptr_mondata[i].flag_initiative));

		/* ² */
		if (ptr_mondata[i].flag_undead || ptr_mondata[i].flag_human || ptr_mondata[i].flag_imp)
		{
			fprintf(fp, " ");

			if (ptr_mondata[i].flag_undead)
				fprintf(fp, " %s", str_montype[0]);
			if (ptr_mondata[i].flag_human)
				fprintf(fp, " %s", str_montype[1]);
			if (ptr_mondata[i].flag_imp)
				fprintf(fp, " %s", str_montype[2]);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeitem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥νƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥νƥǡϤ */
	fprintf(fp, "No.       ̾               "
				"(Ψ)      "
				"(Ψ)      "
				"Ȥ(Ψ)    "
				"Ȥ(Ψ)\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
			ptr_name[j] = (ptr_mondata[i].item_no[j] != 0xff) ? item_getname(ptr_mondata[i].item_no[j]) : str_nothing;

		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeragna(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ΥƥѲǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ΥƥѲǡϤ */
	fprintf(fp, "No.       ̾               "
				"   "
				"ƥ1         "
				"ƥ2         "
				"ƥ3         "
				"ƥ4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
			ptr_name[j] = (ptr_mondata[i].ragna_item_no[j] != 0xff) ? item_getname(ptr_mondata[i].ragna_item_no[j]) : str_nothing;

		fprintf(fp, "%03d(%03X): %-16s :  %3d,  ", i, i,
			mon_getname(i),
			ptr_mondata[i].item_metamorph);
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writectrl(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "------------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Ρ֤Ĥ׹ưǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Ρ֤Ĥ׹ưǡϤ */
	fprintf(fp, "No.       ̾               "
				"  "
				"Ĥ1           "
				"Ĥ2           "
				"Ĥ3           "
				"Ĥ4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
		{
			if (ptr_mondata[i].control_no[j] == 0xef)		/* 0xef = "Ȥ" */
				ptr_name[j] = mon_gettechname(i);
			else if (ptr_mondata[i].control_no[j] == 0xff)	/* 0xff = ̤ */
				ptr_name[j] = str_nothing;
			else
				ptr_name[j] = mag_getname(ptr_mondata[i].control_no[j]);
		}

		fprintf(fp, "%03d(%03X): %-16s :  %2s   ", i, i, mon_getname(i),
				str_getflag(ptr_mondata[i].flag_control));
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-18s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writerage(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Ρ֤Ф׹ưǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Ρ֤Ф׹ưǡϤ */
	fprintf(fp, "No.       ̾               "
				"Ф1         "
				"Ф2\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[2];

		for (j = 0; j < 2; j++)
		{
			if (ptr_mondata[i].rage_no[j] == 0xef)			/* 0xef = "Ȥ" */
				ptr_name[j] = mon_gettechname(i);
			else if (ptr_mondata[i].rage_no[j] == 0xff)		/* 0xff = ̤ */
				ptr_name[j] = str_nothing;
			else
				ptr_name[j] = mag_getname(ptr_mondata[i].rage_no[j]);
		}

		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));
		for (j = 0; j < 2; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writesketch(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Ρ֥å׹ưǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Ρ֥å׹ưǡϤ */
	fprintf(fp, "No.       ̾               "
				"  "
				"å1         "
				"å2\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[2];

		for (j = 0; j < 2; j++)
		{
			if (ptr_mondata[i].sketch_no[j] == 0xef)		/* 0xef = "Ȥ" */
				ptr_name[j] = mon_gettechname(i);
			else if (ptr_mondata[i].sketch_no[j] == 0xff)	/* 0xff = ̤ */
				ptr_name[j] = str_nothing;
			else
				ptr_name[j] = mag_getname(ptr_mondata[i].sketch_no[j]);
		}

		fprintf(fp, "%03d(%03X): %-16s :  %2s   ", i, i, mon_getname(i),
				str_getflag(ptr_mondata[i].flag_sketch));
		for (j = 0; j < 2; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writetech(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;
	const char *str_nothing = "----------";
	const char *str_tech_effect[3] = {
		"HPۼ",
		"MPۼ",
		": ȿ",
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ü쵻ΥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ü쵻ΥǡϤ */
	fprintf(fp, "No.       "
				"̾               "
				"Ȥ           "
				"᡼%%    Ѳ   ̵\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_state = str_nothing;
		int damage_multiplier = 0;

		if (ptr_mondata[i].tech_type <= 0x1f)
			ptr_state = str_getstate(~0UL, ptr_mondata[i].tech_type & 0x1f);
		else if (ptr_mondata[i].tech_type <= 0x2f)
			damage_multiplier = (ptr_mondata[i].tech_type & 0x0f) + 1;
		else if (ptr_mondata[i].tech_type <= 0x32)
			ptr_state = str_tech_effect[ptr_mondata[i].tech_type & 0x0f];

		fprintf(fp, "%03d(%03X): %-16s : %-18s,    %5d, %10s,   %2s\n", i, i,
				mon_getname(i),
				mon_gettechname(i),
				(ptr_mondata[i].tech_flag_damage) ? 100 + 50 * damage_multiplier : 0,
				ptr_state,
				str_getflag(ptr_mondata[i].tech_flag_ignore_evade));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemisc(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥Τ¾ΥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥Τ¾ΥǡϤ */
	fprintf(fp, "No.       "
				"̾               "
				"̾ﹶ\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : %-18s\n", i, i,
				mon_getname(i),
				item_getname(ptr_mondata[i].weapon_no));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeact(const char *ptr_filename, const monact_t *ptr_monact, const mondata_t *ptr_mondata)
{
	int i, j, k;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monact == NULL || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ιưѥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ιưѥǡϤ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X): %s]\n", i, i, mon_getname(i));

		for (j = 0; j < ptr_monact[i].ptn_num; j++)
		{
			int turn = 1;
			int turn_old = 0;
			int letter_num = 0;
			unsigned char *buf = NULL;
			unsigned char *ptr_buf = NULL;
			const size_t action_size = ptr_monact[i].ptn[j].action_size;
			const size_t bufsize = action_size + 16;

			/* ѤΰХåեݤ */
			if ((buf = malloc(bufsize)) == NULL)
				err_printf(STR_FATAL_ERROR_IN_MALLOC);
			memset(buf, 0x00, bufsize);
			memmove(buf, ptr_monact[i].ptn[j].action, action_size);

			/* ŬѾ */
			ptr_buf = buf;
			ptr_buf = mon_printactcnd(fp, ptr_buf);
			fputc('\n', fp);

			/* Τ¸ߤƹưѥ¸ߤʤ */
			if (ptr_buf - buf >= (ptrdiff_t)action_size)
			{
				fprintf(fp, "%d: ", turn);
				mon_printact(fp, 0xfe, i);
				fputc('\n', fp);
			}

			/* ưѥ */
			while (ptr_buf - buf < (ptrdiff_t)action_size)
			{
				/* в */
				if (*ptr_buf == 0xfd)
				{
					turn++;
					ptr_buf++;
					continue;
				}

				/* ɽ */
				if (turn != turn_old)
					letter_num = fprintf(fp, "%d: ", turn);
				else
					fprintf(fp, "%*s", letter_num, "");
				turn_old = turn;

				/* ưѥǡβ */
				if (*ptr_buf < 0xf0)
				{
					mon_printact(fp, *ptr_buf, i);
					ptr_buf++;
				}
				else if (*ptr_buf == 0xfc)
				{
					ptr_buf = mon_printactcnd(fp, ptr_buf);
				}
				else
				{
					if (ptr_buf[0] == 0xf0)
					{
						/* ư(f0 xx yy zz) */
						/*   xx: ˡֹ1 */
						/*   yy: ˡֹ2 */
						/*   zz: ˡֹ3 */
						for (k = 0; k < 3; k++)
						{
							if (k != 0)
								fprintf(fp, ", ");
							mon_printact(fp, ptr_buf[k + 1], i);
						}
					}
					else if (ptr_buf[0] == 0xf1)
					{
						/* ɸ(f1 xx) */
						/*   xx: о */
						fprintf(fp, "[ɸ: %s]", mon_gettarget(ptr_buf[1]));
					}
					else if (ptr_buf[0] == 0xf2)
					{
						/* 󥹥ѡƥѹ(f2 xx yy zz) */
						/*   xx: и? */
						/*   yy: 󥹥ѡƥֹ(8ӥå) */
						/*   zz: 󥹥ѡƥֹ(7ӥå) */
						int party_no = ((ptr_buf[3] & 0x7f) << 8) | ptr_buf[2];

						fprintf(fp, "[̥󥹥ѡƥо: %d(%03Xh)%s]", party_no, party_no,
								((ptr_buf[3] & 0x80) == 0x00) ? " (HPݻ)" : "");
					}
					else if (ptr_buf[0] == 0xf3)
					{
						/* åɽ(f3 xx yy) */
						/*   xx: åֹ(8ӥå) */
						/*   yy: åֹ(8ӥå) */
						fprintf(fp, "[åɽ: %s]", mon_getmes((ptr_buf[2] << 8) | ptr_buf[1]));
					}
					else if (ptr_buf[0] == 0xf4)
					{
						/* ̣ѥޥɤΥ¹(f4 xx yy zz) */
						/*   xx: ޥɤֹ1 */
						/*   yy: ޥɤֹ2 */
						/*   zz: ޥɤֹ3 */
						for (k = 0; k < 3; k++)
						{
							if (k != 0)
								fprintf(fp, ", ");

							if (ptr_buf[k + 1] != 0xfe)
								fprintf(fp, "%s", chr_getcmdname(ptr_buf[k + 1]));
							else
								mon_printact(fp, ptr_buf[k + 1], i);
						}
					}
					else if (ptr_buf[0] == 0xf5)
					{
						/* и(f5 xx yy zz) */
						/*   xx: Ū */
						/*   yy: μ */
						/*   zz: о(ӥå) */
						const char *str_operation[] = {"и()", "Ʈǽ", "и", "õ()", "õ"};

						if (ptr_buf[2] < sizeof(str_operation) / sizeof(str_operation[0]))
						{
							fprintf(fp, "[и: ");

							if (ptr_buf[3] == 0x00)
								/* 󥹥ʬ */
								fputs(mon_gettarget(0x36), fp);
							else if (ptr_buf[3] == 0xff)
								/* 󥹥 */
								fputs(mon_gettarget(0x38), fp);
							else
							{
								int index;
								char str_no[256] = "";
								char *ptr = str_no;

								for (index = 0; index < 8; index++)
								{
									if ((ptr_buf[3] & (1U << index)) != 0x00)
										ptr += sprintf(ptr, "%s%d", (ptr > str_no) ? ", " : "", index + 1);
								}

								fprintf(fp, "%sܤΥ󥹥", str_no);
							}

							fprintf(fp, "%s] [и: %02Xh]",
									str_operation[ptr_buf[2]],
									ptr_buf[1]);
						}
					}
					else if (ptr_buf[0] == 0xf6)
					{
						/* ƥ(f6 xx yy zz) */
						/*   xx: ե饰(0: ƥѤ, 1: ƥꤲ) */
						/*   yy: ƥֹ1 */
						/*   zz: ƥֹ2 */
						fprintf(fp, "[ƥ%s] ", (ptr_buf[1] == 0x00) ? "" : "ꤲ");
						for (k = 0; k < 2; k++)
						{
							if (k != 0)
								fprintf(fp, ", ");
							fprintf(fp, "%s", item_getname(ptr_buf[k + 2]));
						}
					}
					else if (ptr_buf[0] == 0xf7)
					{
						/* ٥ȯ(f7 xx) */
						/*   xx: ٥Ȥֹ */
						int index;

						fprintf(fp, "[٥ȯ: %d(%02Xh)]", ptr_buf[1], ptr_buf[1]);
						for (k = 0; k < 64 && (index = mon_geteventdata(ptr_buf[1], k)) >= 0; k++)
						{
							fprintf(fp, "\n%*s[åɽ: ", letter_num, "");
							fputs(mon_geteventmes(index), fp);
							fprintf(fp, "]");
						}
					}
					else if (ptr_buf[0] == 0xf8)
					{
						/* ѿ(f8 xx yy) */
						/*   xx: ⡼ֹ */
						fprintf(fp, "[ѿ%d: %s%d]",
								ptr_buf[1] & 0x1f,
								((ptr_buf[2] & 0x80) == 0x00) ? "" : (((ptr_buf[2] & 0x40) == 0x00) ? "+" : "-"),
								ptr_buf[2] & 0x3f);
					}
					else if (ptr_buf[0] == 0xf9)
					{
						/* ӥåѿ(f9 xx yy zz) */
						/*   xx: ӥåμ(0: ȿž, 1: å, 2: ꥢ) */
						/*   yy: ӥåֹ */
						/*   zz: ӥåȥѥ */
						const char *str_operation[] = {"ȿž", "å", "ꥢ"};

						if (ptr_buf[1] < sizeof(str_operation) / sizeof(str_operation[0]))
						{
							fprintf(fp, "[ӥåѿ%d%s: bit%d]",
									ptr_buf[2],
									str_operation[ptr_buf[1]],
									ptr_buf[3]);
						}
					}
					else if (ptr_buf[0] == 0xfa)
					{
						/* 󥹥и(fa xx yy zz) */
						/*   xx: μ */
						/*   yy: о(ӥå) */
						/*   zz:  */
						const char *str_operation[16] = {
							"֤ȯ",
							"1",	/* ٤ */
							"1",	/* ٤ */
							"1",	/* ® */
							"1",	/* ® */
							"饯¦˰ư",
							"饯¦˰ư",
							"3",
							"3",
							"̲",
							"饹ȥХȥѤα",
							"Ĺȯ",
							"ûȯ",
							"Ʈλα",
							"0Eh",
							"0Fh",
						};

						if (ptr_buf[1] < sizeof(str_operation) / sizeof(str_operation[0]))
						{
							fprintf(fp, "[󥹥и: ");

							if (ptr_buf[2] != 0x00)
							{
								int index;
								char str_no[256] = "";
								char *ptr = str_no;

								for (index = 0; index < 8; index++)
								{
									if ((ptr_buf[2] & (1U << index)) != 0x00)
										ptr += sprintf(ptr, "%s%d", (ptr > str_no) ? ", " : "", index + 1);
								}

								fprintf(fp, "%sܤΥ󥹥", str_no);
							}

							fprintf(fp, "%s", str_operation[ptr_buf[1]]);
							if (ptr_buf[1] == 0x09)
								fprintf(fp, ": %d(%02Xh)", ptr_buf[3], ptr_buf[3]);

							fprintf(fp, "]");
						}
					}
					else if (ptr_buf[0] == 0xfb)
					{
						/* Ƽ(fb xx yy) */
						/*   xx: ̤μ */
						/*   yy: о */
						switch (ptr_buf[1])
						{
						case 0x00:
							fprintf(fp, "[󥹥и֤]");
							break;
						case 0x01:
						case 0x05:
							fprintf(fp, "[%s̵Ũ%s]",
									mon_gettarget(ptr_buf[2]),
									(ptr_buf[1] == 0x01) ? "" : "");
							break;
						case 0x02:
						case 0x09:
							fprintf(fp, "[Ʈλ]");
							break;
						case 0x03:
							fprintf(fp, "[ѡƥС˲]");
							break;
						case 0x04:
							fprintf(fp, "[Ʈϸηв֤]");
							break;
						case 0x06:
						case 0x07:
							fprintf(fp, "[%sɸ%sˤ]",
									mon_gettarget(ptr_buf[2]),
									(ptr_buf[1] == 0x06) ? "ǽ" : "ǽ");
							break;
						case 0x0b:
						case 0x0c:
							fprintf(fp, "[%s: \"%s\"]",
									(ptr_buf[1] == 0x0b) ? "" : "",
									str_getstate(~0UL, ptr_buf[2]));
							break;
						case 0x0d:
							/* 󥹥ʬ */
							fprintf(fp, "[%sΥæ֤ꤹ]",
									mon_gettarget(0x36));
							break;
						default:
							fprintf(fp, "[%02X %02X %02X]", ptr_buf[0], ptr_buf[1], ptr_buf[2]);
							break;
						}
					}
					else
					{
						fprintf(fp, "[");
						for (k = 0; k < tbl_monact_skip_byte[*ptr_buf - 0xf0]; k++)
						{
							if (k != 0)
								fprintf(fp, " ");
							fprintf(fp, "%02X", ptr_buf[k]);
						}
						fprintf(fp, "]");
					}

					ptr_buf += tbl_monact_skip_byte[*ptr_buf - 0xf0];
				}

				fputc('\n', fp);
			}

			free(buf);
			buf = NULL;
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemes(const char *ptr_filename)
{
	int i;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0')
	{
		assert(0);
		return 0;
	}

	/* 󥹥ΥåǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ΥåǡϤ */
	fprintf(fp, "No.      å\n");
	for (i = 0; i < NUM_MONMES; i++)
	{
		fprintf(fp, "%03d(%02X): ", i, i);
		fputs(mon_getmes(i), fp);
		fputc('\n', fp);
	}

	/* ٥ȤΥåǡϤ */
	fprintf(fp, "No.      å\n");
	for (i = 0; i < NUM_EVENTMES; i++)
	{
		fprintf(fp, "%03d(%02X): ", i, i);
		fputs(mon_geteventmes(i), fp);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeevent(const char *ptr_filename)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0')
	{
		assert(0);
		return 0;
	}

	/* ٥ȤΥåǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ٥ȤΥåǡϤ */
	for (i = 0; i < NUM_EVENTDATA; i++)
	{
		int index;

		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X)]\n", i, i);
		for (j = 0; j < 64 && (index = mon_geteventdata(i, j)) >= 0; j++)
		{
			fputs(mon_geteventmes(index), fp);
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeparty(const char *ptr_filename, const monparty_t *ptr_monparty)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ΥѡƥǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ΥѡƥǡϤ */
	fprintf(fp, "No.       "
				"  "
				"̾ ظ  ξ¦  "
				"и   "
				"󥹥1         "
				"󥹥2         "
				"󥹥3         "
				"󥹥4         "
				"󥹥5         "
				"󥹥6\n");
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		fprintf(fp, "%03d(%03X):  %3d,  %2s   %2s   %2s   %2s   ", i, i,
				ptr_monparty[i].magic_learning_gain,
				str_getflag(ptr_monparty[i].flag_normal),
				str_getflag(ptr_monparty[i].flag_backattack),
				str_getflag(ptr_monparty[i].flag_pincer),
				str_getflag(ptr_monparty[i].flag_sideattack));

		for (j = 0; j < 6; j++)
			fprintf(fp, "%-2s", str_getflag((ptr_monparty[i].bits_exist & (1U << j)) != 0x00));

		fprintf(fp, "  ");

		mon_printparty(fp, &ptr_monparty[i]);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeenc(const char *ptr_filename, const monenc_t *ptr_monenc, const monparty_t *ptr_monparty)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monenc == NULL || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ơ֥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ơ֥Ϥ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		const int len_monsters = (i < NUM_MONENCRND) ? 4 : 2;

		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X)]\n", i, i);
		for (j = 0; j < len_monsters; j++)
		{
			const int monparty_no = ptr_monenc[i].party_no[j];

			fprintf(fp, "%02d: %03d(%03X): ", j + 1, monparty_no, monparty_no);

			mon_printparty(fp, &ptr_monparty[monparty_no]);
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemap(const char *ptr_filename, const monmap_t *ptr_monmap)
{
	int i, j, x, y;
	FILE *fp = NULL;
	const char *str_name[NUM_MONMAP][4] =
	{
		{"()",	"()",		"()",		"(ʿ)"},
		{"()",	"()",		"()",		"(ʿ)"},
	};

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monmap == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ޥåץǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ޥåפϤ */
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (j = 0; j < 4; j++)
		{
			if (i != 0 || j != 0)
				fputc('\n', fp);

			fprintf(fp, "[No.%1d-%1d: %s]\n", i, j, str_name[i][j]);
			for (y = 0; y < NUM_MONMAPAREA; y++)
			{
				for (x = 0; x < NUM_MONMAPAREA; x++)
				{
					if (x != 0)
						fprintf(fp, " ");

					fprintf(fp, "%03d", ptr_monmap[i].area[x][y].enc_no[j]);
				}
				fputc('\n', fp);
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writelocation(const char *ptr_filename, const mapdata_t *ptr_mapdata,
	const monenc_t *ptr_monenc, const monparty_t *ptr_monparty)
{
	int i, j;
	FILE *fp = NULL;

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mapdata == NULL ||
		ptr_monenc == NULL || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󥹥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󥹥ǡϤ */
	for (i = 0; i < NUM_MAP; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X): %s", i, i, map_getname(i));

		if (ptr_mapdata[i].sub_no != 0)
			fprintf(fp, " / %s", map_getsubname(ptr_mapdata[i].sub_no));

		fprintf(fp, "]\n");

		/* 󥹥ơ֥Ϥ */
		if (ptr_mapdata[i].flag_encounter)
		{
			for (j = 0; j < 4; j++)
			{
				const int monparty_no = ptr_monenc[ptr_mapdata[i].enc_no].party_no[j];

				fprintf(fp, "%02d: %03d(%03X): ", j + 1, monparty_no, monparty_no);

				mon_printparty(fp, &ptr_monparty[monparty_no]);
				fputc('\n', fp);
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeveldt(const char *ptr_filename, const monparty_t *ptr_monparty)
{
	int i;
	FILE *fp = NULL;
	const char *str_nothing = "...";

	/* å */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* åΥ󥹥ǡϤե򳫤 */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* åΥ󥹥ǡϤ */
	fprintf(fp, "No.  A   B   C   D   E   E   F   G \n");
	for (i = 0; i < ((NUM_MONPARTY < 512) ? NUM_MONPARTY : 512); i++)
	{
		if (i % 8 == 0)
		{
			if (i != 0)
				fputc('\n', fp);
			fprintf(fp, "%02d: ", i / 8 + 1);
		}
		else
			fprintf(fp, " ");

		if (ptr_monparty[i].flag_beastfield)
			fprintf(fp, "%03d", i);
		else
			fprintf(fp, "%-3s", str_nothing);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

static unsigned char *mon_printactcnd(FILE *stream, unsigned char *buf_cnd)
{
	int condition_num = 0;
	unsigned char *ptr_cnd = buf_cnd;

	/* å */
	if (stream == NULL || buf_cnd == NULL)
	{
		assert(0);
		return buf_cnd;
	}

	while (*ptr_cnd == 0xfc)
	{
		if (condition_num == 0)
			fprintf(stream, ": ");
		else
			fprintf(stream, " & ");

		if (ptr_cnd[1] == 0x01)
		{
			/* ΥޥɤФ륫󥿡(fc 01 xx yy) */
			/*   xx: ޥɤֹ1 */
			/*   yy: ޥɤֹ2 */
			fprintf(stream, "[\"%s\" ", chr_getcmdname(ptr_cnd[2]));
			if (ptr_cnd[2] != ptr_cnd[3])
				fprintf(stream, " \"%s\" ", chr_getcmdname(ptr_cnd[3]));
			fprintf(stream, "]");
		}
		else if (ptr_cnd[1] == 0x02)
		{
			/* ˡФ륫󥿡(fc 02 xx yy) */
			/*   xx: ˡֹ1 */
			/*   yy: ˡֹ2 */
			fprintf(stream, "[ˡ \"%s\" ", mag_getname(ptr_cnd[2]));
			if (ptr_cnd[2] != ptr_cnd[3])
				fprintf(stream, " \"%s\" ", mag_getname(ptr_cnd[3]));
			fprintf(stream, "]");
		}
		else if (ptr_cnd[1] == 0x03)
		{
			/* ΥƥФ륫󥿡(fc 03 xx yy) */
			/*   xx: ƥֹ1 */
			/*   yy: ƥֹ2 */
			fprintf(stream, "[ƥ \"%s\" ", item_getname(ptr_cnd[2]));
			if (ptr_cnd[2] != ptr_cnd[3])
				fprintf(stream, " \"%s\" ", item_getname(ptr_cnd[3]));
			fprintf(stream, "]");
		}
		else if (ptr_cnd[1] == 0x04)
		{
			/* °Ф륫󥿡(fc 04 xx 00) */
			/*   xx: ° */
			int index;
			char str_element[256] = "";
			char *ptr = str_element;
			const elem_t element = ptr_cnd[2];

			for (index = 0; index < NUM_ELEMENT; index++)
				ptr += sprintf(ptr, "%s", str_getelem(element, index));

			fprintf(stream, "[%s°ι]", str_element);
		}
		else if (ptr_cnd[1] == 0x05)
		{
			/* ˥᡼Ƥ˥󥿡(fc 05 00 00) */
			fprintf(stream, "[᡼/]");
		}
		else if (ptr_cnd[1] == 0x06)
		{
			/* ҥåȥݥȤͰʲ(fc 06 xx yy) */
			/*   xx: о */
			/*   yy: ҥåȥݥȤ128ǳä */
			fprintf(stream, "[%sHP%ldʲǤ]",
					mon_gettarget(ptr_cnd[2]),
					(long)ptr_cnd[3] * 128);
		}
		else if (ptr_cnd[1] == 0x07)
		{
			/* ޥåݥȤͰʲ(fc 07 xx yy) */
			/*   xx: о */
			/*   yy: ޥåݥ */
			fprintf(stream, "[%sMP%ldʲǤ]",
					mon_gettarget(ptr_cnd[2]),
					(long)ptr_cnd[3]);
		}
		else if (ptr_cnd[1] == 0x08 || ptr_cnd[1] == 0x09)
		{
			/* ֤ξ֤Ȱפ(fc 08 xx yy) */
			/* ֤ξ֤ǤϤʤ(fc 09 xx yy) */
			/*   xx: о */
			/*   yy: ֤Υǥå */
			fprintf(stream, "[%sξ֤ \"%s\" %s]",
					mon_gettarget(ptr_cnd[2]),
					str_getstate(~0UL, ptr_cnd[3]),
					(ptr_cnd[1] == 0x08) ? "" : "ʤ");
		}
		else if (ptr_cnd[1] == 0x0b || ptr_cnd[1] == 0x16)
		{
			/* 󥹥иְ֤ʾǤ(fb 0b xx 00) */
			/* Ʈϸηвְ֤ʾǤ(fb 16 xx 00) */
			/*   xx: ñ̻ */
			fprintf(stream, "[%s%dñְ̻ʾǤ]",
					(ptr_cnd[1] == 0x0b) ? "󥹥и" : "Ʈϸηв",
					ptr_cnd[2]);
		}
		else if (ptr_cnd[1] == 0x0c || ptr_cnd[1] == 0x0d)
		{
			/* ѿ̤Ǥ(fc 0c xx yy) */
			/* ѿͰʾǤ(fc 0d xx yy) */
			/*   xx: ɤѿͤѤ뤫 */
			/*   yy: ѿ */
			fprintf(stream, "[ѿ%d%d%sǤ]",
					ptr_cnd[2] & 0x1f,
					ptr_cnd[3],
					(ptr_cnd[1] == 0x0c) ? "̤" : "ʾ");
		}
		else if (ptr_cnd[1] == 0x0e || ptr_cnd[1] == 0x0f)
		{
			/* ٥뤬̤Ǥ(fc 0e xx yy) */
			/* ٥뤬ͰʾǤ(fc 0f xx yy) */
			/*   xx: о */
			/*   yy: ٥ */
			fprintf(stream, "[%sΥ٥뤬%d%sǤ]",
					mon_gettarget(ptr_cnd[2]),
					ptr_cnd[3],
					(ptr_cnd[1] == 0x0e) ? "̤" : "ʾ");
		}
		else if (ptr_cnd[1] == 0x10)
		{
			/* ʬʳΥ󥹥ʤ(fc 10 00 00) */
			fprintf(stream, "[ʬʳΥ󥹥ʤ]");
		}
		else if (ptr_cnd[1] == 0x11 || ptr_cnd[1] == 0x12 || ptr_cnd[1] == 0x19)
		{
			/* ֤˥󥹥¸ߤ(fc 11 xx 00) */
			/* ֤˥󥹥¸ߤʤ(fc 12 xx 00) */
			/* ʬȤ֤¸ߤ(fc 19 xx 00) */
			/*   xx: (ӥå) */
			int index;
			char str_no[256] = "";
			char *ptr = str_no;

			for (index = 0; index < 8; index++)
			{
				if ((ptr_cnd[2] & (1U << index)) != 0x00)
					ptr += sprintf(ptr, "%s%d", (ptr > str_no) ? ", " : "", index + 1);
			}

			switch (ptr_cnd[1])
			{
			case 0x11:
				fprintf(stream, "[%sܤΥ󥹥иƤ]", str_no);
				break;
			case 0x12:
				if (ptr_cnd[2] != 0x00)
					fprintf(stream, "[%sܤΥ󥹥ʤ]", str_no);
				else
					fprintf(stream, "[եʥ륢å]");
				break;
			case 0x19:
				fprintf(stream, "[ʬȤ%sܤΰ֤ˤ]", str_no);
				break;
			}
		}
		else if (ptr_cnd[1] == 0x13)
		{
			/* ̣/ŨοͿ(fc 13 xx yy) */
			/*   xx: ե饰(0: ̣οͿyyͰʾ夤뤫, 1: Ũοyyΰʲ) */
			/*   yy: ̣/ŨοͿ */
			if (ptr_cnd[2] == 0x00)
				fprintf(stream, "[̣饯%dͰʾ¸ߤ]", ptr_cnd[3]);
			else
				fprintf(stream, "[󥹥ο%dΰʲǤ]", ptr_cnd[3]);
		}
		else if (ptr_cnd[1] == 0x14 || ptr_cnd[1] == 0x15)
		{
			fprintf(stream, "[ӥåѿ%dbit%d%sƤ]",
					ptr_cnd[2],
					ptr_cnd[3],
					(ptr_cnd[1] == 0x14) ? "å" : "ꥢ");
		}
		else if (ptr_cnd[1] == 0x17)
		{
			/* ɸ¸ߤ(fc 17 xx 00) */
			/*   xx: о */
			fprintf(stream, "[ɸ \"%s\" ¸ߤ]", mon_gettarget(ptr_cnd[2]));
		}
		else if (ptr_cnd[1] == 0x18)
		{
			/* ѡƥ˲ääƤʤ(fc 18 00 00) */
			fprintf(stream, "[ѡƥ˲ääƤʤ]");
		}
		else if (ptr_cnd[1] == 0x1a)
		{
			/* °Ȱפ(fc 1a xx yy) */
			/*   xx: о */
			/*   yy: ° */
			int index;

			fprintf(stream, "[%sμ \"",
					mon_gettarget(ptr_cnd[2]));
			for (index = 0; index < NUM_ELEMENT; index++)
				fprintf(stream, "%s", str_getelem(ptr_cnd[3], index));
			fprintf(stream, "\" Ǥ]");
		}
		else if (ptr_cnd[1] == 0x1b)
		{
			/* 󥹥ѡƥֹ椬ͤȰפ(fc 1b xx yy) */
			/*   xx: 󥹥ѡƥֹ(̥Х) */
			/*   yy: 󥹥ѡƥֹ(̥Х) */
			int party_no = (ptr_cnd[3] << 8) | ptr_cnd[2];

			fprintf(stream, "[󥹥ѡƥֹ椬%d(%03Xh)Ǥ]", party_no, party_no);
		}
		else if (ptr_cnd[1] == 0x1c)
		{
			/* ˿(fc 1c 00 00) */
			fprintf(stream, "[˹ư/ȿ⤹]");
		}
		else
		{
			fprintf(stream, "[%02X %02X %02X %02X]", ptr_cnd[0], ptr_cnd[1], ptr_cnd[2], ptr_cnd[3]);
		}

		condition_num++;
		ptr_cnd += 4;
	}

	if (condition_num == 0)
		fprintf(stream, ": [̾]");

	return ptr_cnd;
}

static void mon_printact(FILE *stream, int index, int monster_no)
{
	/* å */
	if (stream == NULL || index < 0 || index >= 0xff || monster_no < 0 || monster_no >= NUM_MON)
	{
		assert(0);
		return;
	}

	switch (index)
	{
	case 0xef:		/* Ȥ */
		fprintf(stream, "%s", mon_gettechname(monster_no));
		break;
	case 0xfe:
		fprintf(stream, "(⤷ʤ)");
		break;
	default:
		fprintf(stream, "%s", mag_getname(index));
		break;
	}
}

void mon_printparty(FILE *stream, const monparty_t *ptr_monparty)
{
	int i;
	const char *str_nothing = "----------------";

	/* å */
	if (stream == NULL || ptr_monparty == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < 6; i++)
	{
		const char *ptr_name = (ptr_monparty->monster_no[i] < NUM_MON) ? mon_getname(ptr_monparty->monster_no[i]) : str_nothing;
		const int flag_absent = (ptr_monparty->monster_no[i] < NUM_MON && (ptr_monparty->bits_exist & (1U << i)) == 0x00);

		if (i != 0)
			fprintf(stream, ", ");

		fprintf(stream, "%s%-16s%s",
				(flag_absent) ? "(" : " ",
				ptr_name,
				(flag_absent) ? ")" : " ");
	}
}

void mon_freeact(monact_t *ptr_monact)
{
	int i, j;

	/* å */
	if (ptr_monact == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_MON; i++)
	{
		for (j = 0; j < ptr_monact[i].ptn_num; j++)
		{
			if (ptr_monact[i].ptn[j].action != NULL)
			{
				free(ptr_monact[i].ptn[j].action);
				ptr_monact[i].ptn[j].action = NULL;
			}
		}
	}
}
